// ***************************************************************************************
// File: http.cpp
// Description: http server and client
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;

map<string,int> RedirectionExcludeList;
NvMutex ProtectRedirectionExcludeList;

void NvHttpClientList::Add(void *Item)
{
	NV_LOCKOBJECT(this);
	m_vClients.push_back(Item);
	NV_UNLOCKOBJECT(this);
}

void NvHttpClientList::Remove(void *Item)
{
	NV_LOCKOBJECT(this);
	m_vClients.remove(Item);
	NV_UNLOCKOBJECT(this);
}

void* NvHttpClientList::RemoveLast()
{
	void *Item = NULL;
	NV_LOCKOBJECT(this);
	if(!m_vClients.empty())
	{
		Item = m_vClients.back();
		m_vClients.pop_back();
	}
	NV_UNLOCKOBJECT(this);
	return Item;
}

bool NvHttpClientList::IsEmpty()
{
	bool result = false;
	NV_LOCKOBJECT(this);
	if(m_vClients.empty())
		result = true;
	NV_UNLOCKOBJECT(this);
	return result;
}

// Kills all the connections actually used by a session
void NvHttpClientList::Kill(NvSession *Session)
{
	std::vector<NvHttpClient *> ClientsToKill;		// client list to kill
	NV_LOCKOBJECT(this);
	list<void *>::iterator pos;
	for(pos = m_vClients.begin(); pos != m_vClients.end(); ++pos)
	{
		NvHttpClient *Client = (NvHttpClient *)(*pos);
		if(Client != NULL)
			if(Client->GetSession() == Session)
				ClientsToKill.push_back(Client);
	}
	int i;
	for(i=0; i<(int)(ClientsToKill.size()); i++)
	{
		NvHttpClient *Client = ClientsToKill[i];
		m_vClients.remove(Client);
		Client->KillRequested = true;
		Client->Kill();
	}
	NV_UNLOCKOBJECT(this);
}

int NvHttpClientList::Size()
{
	int result = 0;
	NV_LOCKOBJECT(this);
	result = (int)(m_vClients.size());
	NV_UNLOCKOBJECT(this);
	return result;
}



NvHttpServer::NvHttpServer()
{
	b_Initialized = false;
	b_UseSSL = false;
	b_LocalOnly = false;
	b_SendHttpServerHeader = true;
	b_IpV6 = false;

}


// Gets the root directory
string NvHttpServer::GetRootDir()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = sRootDir;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Gets the default page
string NvHttpServer::GetDefaultPage()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = sDefaultPage;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// HTTP server main thread function
int NvHttpServer::Main()
{
//	int ClientSid;
	int NumClientSid;
	while(1)
	{
		// we first listen the socket
		NvHttpClientBuilder *Builder = NVNEW NvHttpClientBuilder;
		Builder->b_UseSSL = b_UseSSL;
		Builder->b_LocalOnly = b_LocalOnly;
		Builder->b_SendHttpServerHeader = b_SendHttpServerHeader;
		sSocket.bSslConnection = b_UseSSL;
		sSocket.bIpv6 = b_IpV6;
		NumClientSid = sSocket.Listen(Port, 2, &(Builder->m_vPending), &(Builder->m_vPendingSSL), &(Builder->m_vPendingPeerAddr));
		if(NumClientSid == -1)
		{
			// Error occured when listening for a client connection
			// We leave the process with error
			delete Builder;
			sSocket.Close();		// close the server socket in order to reopen it
			continue;
		}

		// we leave if requested
		if(AskToTerminate.IsSignaled())
		{
			int i;
			for(i=0; i<(int)(Builder->m_vPending.size()); i++)
			{
				int ClientSid = Builder->m_vPending[i];
				if(b_UseSSL)
				{
					SSL* clientssl = (SSL*)(Builder->m_vPendingSSL[i]);
					SSL_shutdown (clientssl);
				}
				shutdown(ClientSid, 2);
#ifdef WIN32
				closesocket(ClientSid);
#else
				close(ClientSid);
#endif
			}
			delete Builder;
			break;
		}

		if(NumClientSid == 0)
		{
			delete Builder;
			continue;
		}

		Builder->HttpServer = this;
		Builder->ClientTimeOut = ClientTimeOut;
		Builder->MaxMemBody = MaxMemBody;
		Builder->Run(false, true);

	}
	return 0;
}


// called to stop the server
void NvHttpServer::Stop()
{
//	SendToLog("Request to stop the server");
	
	// we just the AskToTerminate event to signaled
	AskToTerminate.Signal();
}


// called when an exception occurs (see class NvTask)
void NvHttpServer::OnException(string sInfo)
{
//	SendToLog("** Exception occured on the http server **");

	GetMain()->IncNumExceptions();
	GetMain()->Logs.Write("Exceptions", "Exception on Http server", NULL, NV_LOG_ERROR);

	// We must signal the AskToTerminate event for all client threads
	AskToTerminate.Signal();
}


bool NvHttpServer::InitSSL(const char* CertificateFile, const char* PrivateKeyFile, bool SSLV3Only, bool AskClientCertificate, const char* ClientCAFile)
{
	if(!b_UseSSL)
		return true;
	if(CertificateFile == NULL)
		return false;
	if(PrivateKeyFile == NULL)
		return false;
	sSocket.bSslConnection = true;
	if(CertificateFile != NULL)
		sSocket.CertificateFile = CertificateFile;
	sSocket.PrivateKeyFile = PrivateKeyFile;
	sSocket.bSSLV3Only = SSLV3Only;
	sSocket.bAskClientCertificate = AskClientCertificate;
	if(ClientCAFile != NULL)
		sSocket.ClientCAFile = ClientCAFile;
	if(!sSocket.InitServerSSL())
		return false;
	return true;
}


// called before the Main function (see class NvTask)
bool NvHttpServer::OnInit()
{
	NvTask::OnInit();	// call the base class

	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_HTTP_SERVER);

	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type HTTP SERVER", NV_LOG_INFO, 1, sThreadId, "system");
	
	AskToTerminate.Reset();		// we reset the asktoterminate event in case it is set

	sSocket.bSslConnection = b_UseSSL;
	if(!sSocket.InitServerSSL())
	{
		if(!m_bSameThread)
			GetMain()->Logs.Write("Threads", "Error init thread (init SSL)", "Type HTTP SERVER", NV_LOG_ERROR, 1, sThreadId, "system");
		return false;
	}

	sSocket.bIpv6 = b_IpV6;
	b_Initialized = true;

	// Implement your code here
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type HTTP SERVER", NV_LOG_INFO, 1, sThreadId, "system");

	return true;
}


// called after the Main function (see class NvTask)
void NvHttpServer::OnCleanup()
{

	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type HTTP SERVER", NV_LOG_INFO, 1, sThreadId, "system");

	if(b_Initialized)
	{
		// we have to wait for all clients to close themselves
		// and them to hardly close the rest

		// wait a maximum of MaxStopWaitTime seconds for the client list size is 0
	//	SendToLog("Wait for all http client threads to terminate");
		time_t initial_time = time(NULL);
		while((int)(time(NULL) - initial_time) < MaxStopWaitTime)
		{
			if(Clients.IsEmpty())
				break;	// the client size is empty meaning that all the clients have beeen closed properly

			// let other threads do their work (in milliseconds)
			SLEEP(500);
		}
		
		// now, we hardly kill the unterminated threads
		if(!Clients.IsEmpty())
		{
	//		SendToLog("Some http client threads have not been closed properly");
			while(1)
			{
				// we remove and get a client from the list
				NvHttpClient *Client = (NvHttpClient *)(Clients.RemoveLast());
				
				// If no more clients, we break;
				if(Client == NULL)
					break;
				
				// otherwise, we kill the client thread
				Client->KillRequested = true;
				Client->Kill();
			}
			// We let again time for threads to terminate
			SLEEP(2000);
		}
	}
	
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type HTTP SERVER", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}





NvHttpClientBuilder::NvHttpClientBuilder()
{
	b_Initialized = false;
	b_UseSSL = false;
	b_LocalOnly = false;
	b_SendHttpServerHeader = true;
}



// HTTP client builder main thread function
int NvHttpClientBuilder::Main()
{
	int ClientSid;
	int i;
	for(i=0; i<(int)(m_vPending.size()); i++)
	{
		ClientSid = m_vPending[i]; 
		if(ClientSid != 0)
		{
			NvHttpClient *Client = NVNEW NvHttpClient;
			if(b_UseSSL)
			{
				Client->sSocket.ssl = (SSL*)(m_vPendingSSL[i]);
				Client->sSocket.bSslConnection = true;
			}
			Client->b_LocalOnly = b_LocalOnly;
			Client->b_SendHttpServerHeader = b_SendHttpServerHeader;
			Client->sSocket.Sid = ClientSid;
			Client->Server = HttpServer;
			Client->iTimeOut = ClientTimeOut;
			Client->iMaxMemBody = MaxMemBody;
			Client->IpAddress = m_vPendingPeerAddr[i];
			// We copy the alias table to each client for performance reason
			Client->InputMessage.Aliases.m_bCaseSensitive = true;
			GetMain()->GetAliasTable(&(Client->InputMessage.Aliases));
			if(GetMain()->GetDebug())
			{
				if(GetMain()->m_sDebugMode.find("http=") != string::npos)
					Client->bDebugSocket = true;
				if(GetMain()->m_sDebugMode.find("sso=") != string::npos)
					Client->bDebugSso = true;
			}			
			// we run the client thread
			Client->Run(false, true);
		}
	}
	return 0;
}


// called when an exception occurs (see class NvTask)
void NvHttpClientBuilder::OnException(string sInfo)
{
	GetMain()->IncNumExceptions();
	GetMain()->Logs.Write("Exceptions", "Exception on Http client builder", NULL, NV_LOG_ERROR);
	// Do nothing special	
//	throw this;
}


// called before the Main function (see class NvTask)
bool NvHttpClientBuilder::OnInit()
{
	NvTask::OnInit();	// call the base class

	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_HTTP_CLIENT_BUILDER);

	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type HTTP CLIENT BUILDER", NV_LOG_INFO, 1, sThreadId, "system");

	// Do nothing special	
	b_Initialized = true;
	
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type HTTP CLIENT BUILDER", NV_LOG_INFO, 1, sThreadId, "system");

	return true;
}


// called after the Main function (see class NvTask)
void NvHttpClientBuilder::OnCleanup()
{
	// Do nothing special	
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type HTTP CLIENT BUILDER", NV_LOG_INFO, 1, sThreadId, "system");

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type HTTP CLIENT BUILDER", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	
	NvTask::OnCleanup();	// call the base class
}



// Sets the actual session
void NvHttpClient::SetSession(NvSession *Session)
{
	NV_LOCKOBJECT(this);
	m_sSession = Session;
	NV_UNLOCKOBJECT(this);
}

	
// Gets the actual session
NvSession* NvHttpClient::GetSession()
{
	NvSession *RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sSession;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}

	



NvHttpClient::NvHttpClient()
{
	b_Initialized = false;
	b_LocalOnly = false;
	b_SendHttpServerHeader = true;
	bDebugSocket = false;
	bDebugSso = false;
	bRedirected = false;
}

// Client thread main function
int NvHttpClient::Main()
{
	if(b_LocalOnly)
	{
		if((IpAddress != "127.0.0.1") && (IpAddress != "::1"))
		{
			// We must check the ip address
			// if the sender ip adress is empty we do not accept
			if(!IpAddress.length())
				return 0;

			
			vector<string> hosts;
			MyGetHostByName(GetMain()->GetMachineName().c_str(), hosts);
			bool Match = false;
			int i = 0;
			for(i=0;i<(int)(hosts.size());i++)
			{
				string MachineIpAdress = hosts[i];
				CleanString(MachineIpAdress);
				if(IpAddress == MachineIpAdress)
				{
					Match = true;
					break;
				}
			}

/*			struct hostent *hp;
			hp = gethostbyname(GetMain()->GetMachineName().c_str());
			if(hp == NULL)
				return 0;
			struct in_addr addr;
			bool Match = false;
			int i = 0;
			while (hp->h_addr_list[i] != 0)
			{
				addr.s_addr = *(u_long *) hp->h_addr_list[i++];
				string MachineIpAdress = inet_ntoa(addr);
				CleanString(MachineIpAdress);
				if(IpAddress == MachineIpAdress)
				{
					Match = true;
					break;
				}
			}
*/
			if(!Match)
				return 0;
		}
	}

	string TempWorkDir = GetMain()->GetNirvaDir();
	AddDirSuffix(TempWorkDir);
	TempWorkDir += "Work";
	string LogFile = "";
	if(bDebugSocket)
	{
		LogFile = GetMain()->GetNirvaDir();
		AddDirSuffix(LogFile);
		LogFile += "Bin";
		AddDirSuffix(LogFile);
		char Buf[50];
		sprintf(Buf, "socket_%d.log", sSocket.Sid);
		LogFile += Buf;
	}
	while(1)
	{
		// we check the time out if requested
		if(iTimeOut)
		{
			if((int)(time(NULL) - tLastCommandTime) > iTimeOut)
			{
//				SendToLog("** Http client thread time out **", this);
				break;		// yes, we have reach the time-out condition
			}
		}
		
		// we leave if requested
		if(Server->AskToTerminate.IsSignaled())
			break;
		
		// by default we do not stop the connection
		bool bContinue = true;
		
		// we try to read some input for a maximum of 1 second
		int Result = -1;
		try
		{
			sSocket.LogFile = LogFile;
			Result = InputMessage.Read(1, true, iMaxMemBody, TempWorkDir.c_str());
		}
		catch(...)
		{
			if(bDebugSocket)
			{
				sSocket.WriteLog("\n** Http client thread exception occured (Read message)**");
				sSocket.LogFile = "";
				sSocket.CloseLog();
			}
			GetMain()->IncNumExceptions();
			GetMain()->Logs.Write("Exceptions", "Http client thread exception occured (Read message)", NULL, NV_LOG_ERROR);
			printf("** Http client thread exception occured (Read message)**\n");
			Result = -1;
		}
		if(Result < 0)	// error
		{
			if(bDebugSocket)
			{
				sSocket.LogFile = "";
				sSocket.CloseLog();
			}
			break;
		}
		if(bDebugSocket)
		{
			sSocket.LogFile = "";
			sSocket.CloseLog();
		}
		if(Result == 1)	// something has been read
		{
			try
			{
				// we leave if requested
				if(Server->AskToTerminate.IsSignaled())
					break;

				// we clear the output message
				OutputMessage.Clear();
			}
			catch(...)
			{
				GetMain()->IncNumExceptions();
				GetMain()->Logs.Write("Exceptions", "Http client thread exception occured (Before ProcessMessage)", NULL, NV_LOG_ERROR);
				printf("** Http client thread exception occured (Before ProcessMessage)**\n");
				bContinue = false;
			}

			// Check the X-Forwarded-For header for real client IP (if a proxy or load balancing is used)
			string XForwardedFor = InputMessage.GetHeader("X-FORWARDED-FOR");
			CleanString(XForwardedFor);
			if(XForwardedFor.length())
			{
				string::size_type index = XForwardedFor.find(',');
				if(index != string::npos)
				{
					XForwardedFor = XForwardedFor.substr(0,index);
					CleanString(XForwardedFor);
				}
				if(XForwardedFor.length() && (XForwardedFor != "127.0.0.1") && (XForwardedFor != "::1"))
					IpAddress = XForwardedFor;
			}

			// we process the message and stop the connection if requested
			try
			{
				bContinue = ProcessMessage();
			}
			catch(...)
			{
				GetMain()->IncNumExceptions();
				GetMain()->Logs.Write("Exceptions", "Http client thread exception occured (ProcessMessage)", NULL, NV_LOG_ERROR);
				printf("** Http client thread exception occured (ProcessMessage)**\n");
				bContinue = false;
			}

			try
			{
				// we send the connection: close header to signal it to the client
				if(!bContinue)
					OutputMessage.AddHeader("Connection", "close");

				InputMessage.Clear();


				// We send the output to client
				sSocket.LogFile = LogFile;
				Result = OutputMessage.Write();
				if(bDebugSocket)
				{
					sSocket.LogFile = "";
					sSocket.CloseLog();
				}
				if(Result != 1)
					break;		// we leave connection

				// we clear the output message
	//			InputMessage.Clear();
				OutputMessage.Clear();

				tLastCommandTime = time(NULL);
			}
			catch(...)
			{
				if(bDebugSocket)
				{	
					sSocket.LogFile = "";
					sSocket.CloseLog();
				}
				GetMain()->IncNumExceptions();
				GetMain()->Logs.Write("Exceptions", "Http client thread exception occured (After ProcessMessage)", NULL, NV_LOG_ERROR);
				printf("** Http client thread exception occured (After ProcessMessage)**\n");
				bContinue = false;
			}
			if(!bContinue)
				break;
		}

		// we leave if requested
		if(Server->AskToTerminate.IsSignaled())
			break;

		// we stop if requested
		if(!bContinue)
			break;
	
	}
	return 0;
}


// called when an exception occurs (see class NvTask)
void NvHttpClient::OnException(string sInfo)
{
	GetMain()->IncNumExceptions();
	GetMain()->Logs.Write("Exceptions", "Http client thread exception occured", sInfo.c_str(), NV_LOG_ERROR);
//	SendToLog("** Http client thread exception occured **", this);
	printf("** Http client thread exception occured **\n");
	printf("** Exception info: (%s) **\n", sInfo.c_str());
//	throw this;
}


// called before the Main function (see class NvTask)
bool NvHttpClient::OnInit()
{
	NvTask::OnInit();	// call the base class
	
	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_HTTP_CLIENT);

	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type HTTP CLIENT", NV_LOG_INFO, 1, sThreadId, "system");

	SetSession(NULL);
	RemoveSessionOnCleanup = false;
	KillRequested = false;

	// Initialize the random generator for this thread
#ifdef WIN32
	int ThreadHandle = GetCurrentThreadId();						// Thread handle for window
#else
	pthread_t ThreadHandle = pthread_self();					// Thread handle for unix
#endif

	unsigned int randinit = (unsigned int)ThreadHandle;
	randinit *= (unsigned int)time(NULL);
	srand(randinit);
	
	tLastCommandTime = time(NULL);

	InputMessage.sSocket = &sSocket;
	OutputMessage.sSocket = &sSocket;

	Server->Clients.Add(this);

	b_Initialized = true;

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type HTTP CLIENT", NV_LOG_INFO, 1, sThreadId, "system");

	return true;
}


// called after the Main function (see class NvTask)
void NvHttpClient::OnCleanup()
{
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type HTTP CLIENT", NV_LOG_INFO, 1, sThreadId, "system");
	if(b_Initialized)
	{
	
		// Cleanup the actual command
		Command.Clear();

		// Free the session if not made
		NvSession *Session = GetSession();
		if(Session != NULL)
		{
			if(Session->ActualCommand != NULL)
			{
				// The session is in a command so it's probably a crash
				if(Session->ActualCommand->m_sService != "SYSTEM")
					GetMain()->Services.FreeService(Session->ActualCommand->m_sService.c_str());
				Session->ActualCommand = NULL;
				Session->OriginalCommand = NULL;
				Session->InContainer = NULL;
				Session->OutContainer = NULL;
				Session->SetActualCommand("");
				Session->SetOriginalCommand("");
				NV_UNLOCKOBJECT(Session);
			}

			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

			// We remove the session if requested
			if(RemoveSessionOnCleanup)
				GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		}

		SetSession(NULL);
		RemoveSessionOnCleanup = false;

		// close the socket
		sSocket.Close();

		// remove this client from the server client list
		if(!KillRequested)
			Server->Clients.Remove(this);

		// Detach thread from java interface if not yet made
		if(!m_bSameThread)
			GetMain()->JavaInterface.DetachJavaThread();
	}

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type HTTP CLIENT", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}


// this function processes the InputMessage message and sends the result to
// OutputMessage
// The return value is generally true except if there is a request from
// the client to close the connection (at this time the result is false) 
// We can also force a deconnection by sending a Connection: close header in the output message
bool NvHttpClient::ProcessMessage()
{
	NvData *Data = NVNEW NvData;
	OutputMessage.vDatas.push_back(Data);

	// by default, we have no content in the body
	OutputMessage.AddHeader("Content-Length", "0");

	// reset the error to default value (OK)
	SetHttpError();

	// first check the keep connection request
	
	bool Ret = true;
	
	// if we have a Connection: close header
	// we must close the connection after leaving
	// so the return value will be false
	string sClosed = InputMessage.GetHeader("CONNECTION");
	MakeUpper(sClosed);
	if(sClosed.find("CLOSE") != string::npos)
		Ret = false;
//	if(InputMessage.GetHeader("CONNECTION") == "close")
//		Ret = false;

	// for HTTP 1.0 request, we close the connection if we have not the CONNECTION:Keep-Alive header
	if(InputMessage.sHttpVersion == "HTTP/1.0")
	{
		string ConnectionInHeader = InputMessage.GetHeader("CONNECTION");
		CleanString(ConnectionInHeader);
		MakeUpper(ConnectionInHeader);
		if(ConnectionInHeader != "KEEP-ALIVE")
			Ret = false;
	}

	if(!InputMessage.sHttpVersion.length())
	{
		SetHttpError(HTTP_ERROR_BADREQUEST, "bad request syntax");
		Ret = false;
		return Ret;
	}

	// now we test the host header (only for HTTP 1.1)
	if(InputMessage.sHttpVersion == "HTTP/1.1")
	{
		// we must have the Host: header for an HTTP 1.1 request
		if(!InputMessage.HeaderExist("HOST"))
		{
			SetHttpError(HTTP_ERROR_BADREQUEST, "HTTP 1.1 requests must include the Host: header");
			return Ret;
		}
	}

	// get current time and date
	time_t CurrentTime;
	time(&CurrentTime);
	struct tm tCurrentTime;
	char sTime[200];
	gmtime_r(&CurrentTime, &tCurrentTime);
//	strftime(sTime, sizeof(sTime), "%a, %d %b %Y %H:%M:%S GMT", &tCurrentTime);
	MyGetTime(sTime, &tCurrentTime, true);

	// and set it as the Date header
	OutputMessage.AddHeader("Date", sTime);

	// Set the server name header
	string sServer = HTTP_SERVER_STRING;
//	sServer += " version ";
//	sServer += GetMain()->m_sVersion;
	if(b_SendHttpServerHeader)
		OutputMessage.AddHeader("Server", sServer);

	// Test if the method requested is supported
	if(InputMessage.iMethod == HTTP_METHOD_NONE)
	{
		SetHttpError(HTTP_ERROR_NOTIMPLEMENTED, "The method is not implemented in server");
		Ret = false;
		return Ret;
	}

	// Test if the service requested is for us
	// An url can be /path/service?command&parameters
	// or /path/command?parameters or /path/file
	try
	{
		string sServiceNameHere = InputMessage.sService;
		MakeUpper(sServiceNameHere);
		string sCommandNameHere = InputMessage.sCommand;
		MakeUpper(sCommandNameHere);
		bool IsStaticView = false;			// Used for the nirva framework
		// Check if we have /path/command?parameters (NVS is ommitted and the command is used as a service)
		bool bCommandInService = false;
		if(sServiceNameHere == "COMMAND")
			bCommandInService = true;
		else if(sServiceNameHere == "XMLCOMMAND")
			bCommandInService = true;
		else if(sServiceNameHere == "SOAPCOMMAND")
			bCommandInService = true;
		else if(sServiceNameHere == "WEBS")
			bCommandInService = true;
		else if(sServiceNameHere == "WSDL")
			bCommandInService = true;
		else if(((sServiceNameHere.find(".NVFM") != string::npos) && (sServiceNameHere.rfind(".NVFM") == (sServiceNameHere.length() - 5))) || ((sCommandNameHere.find(".NVFM") != string::npos) && (sCommandNameHere.rfind(".NVFM") == (sCommandNameHere.length() - 5))))
		{
			// This is for the internal framework
			// ex: http://localhost:1081/nv_app_zut/myview.nvfm&P1=a&P2=b
			// We must check if this is a static or dynamic view request

			// Get the application name
			string AppName = "";
			if(InputMessage.ParameterExist("NV_APPLICATION"))
				AppName = InputMessage.GetParameter("NV_APPLICATION");
			else
			{
				// If the NV_APPLICATION parameter is not given, we try to get it from
				// the URL
				AppName = InputMessage.sPath;
				MakeUpper(AppName);
				string::size_type index = AppName.find("NV_APP_");
				if(index != string::npos)
				{
					AppName = AppName.substr(index+7);
					if(((index = AppName.find('/')) != string::npos) || ((index = AppName.find('\\')) != string::npos))
						AppName = AppName.substr(0, index);
				}
				else
					AppName = "";
			}
			CleanString(AppName);
			MakeUpper(AppName);
			if(!AppName.length())
				AppName = "NVDEF";

			// Check if the view is static or dynamic
			NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
			if((Application != NULL) && Application->IsEnable())
			{
				if(sServiceNameHere.length())
					IsStaticView = Application->NidgetFramework.IsStaticView(sServiceNameHere.c_str(), GetMain()->GetDebug());
				else
					IsStaticView = Application->NidgetFramework.IsStaticView(sCommandNameHere.c_str(), GetMain()->GetDebug());
				GetMain()->Applications.FreeApplication(AppName.c_str());
			}

			InputMessage.mParameters["NV_HREND"] = "NVFM";
			if(sServiceNameHere.length())
				InputMessage.mParameters["NV_HREND_PAGE"] = sServiceNameHere;
			else
				InputMessage.mParameters["NV_HREND_PAGE"] = sCommandNameHere;

			if(IsStaticView)
				sServiceNameHere = "";
			else
			{
				sServiceNameHere = "NVS";
				sCommandNameHere = "COMMAND";
			}
		}
		else if(sServiceNameHere.length() && (sServiceNameHere.find('.') != string::npos))
		{
			// The service is given and is directly pointing to a rendering page
			// ex: http://localhost:1081/nv_app_zut/tot/titi/login.jsp?P1=a&P2=b
			// We override the NV_HREND and NV_HREND_PAGE parameters
			string hrend_page = InputMessage.sPath;
			hrend_page += InputMessage.sService;
			string hrend_pageHere = hrend_page;
			MakeUpper(hrend_pageHere);
			string::size_type index = hrend_pageHere.find("/NV_APP_");
			if(index == 0)
			{
				hrend_page = hrend_page.substr(1);
				index = hrend_page.find("/");
				if(index != string::npos)
				{
					hrend_page = hrend_page.substr(index+1);
					string hrend = InputMessage.sService;
					MakeUpper(hrend);
					index = hrend.find('.');
					hrend = hrend.substr(index+1);
					if(hrend == "XSL")
						hrend = "XSLT";
					//PM TODO: we must have a mapping from extension to the rendering service name
					bool bValidRendererService = false;
					if(hrend == "XSLT")
						bValidRendererService = true;
					else
					{
						if(GetMain()->Services.IsRenderer(hrend.c_str()))
							bValidRendererService = true;
					}
					if(bValidRendererService)
					{
						InputMessage.mParameters["NV_HREND"] = hrend;
						InputMessage.mParameters["NV_HREND_PAGE"] = hrend_page;
						sServiceNameHere = "NVS";
						sCommandNameHere = "COMMAND";
					}
				}
			}
		}
		if(bCommandInService)
		{
			sCommandNameHere = sServiceNameHere;
			sServiceNameHere = "NVS";
		}
		if(sServiceNameHere.find("NVS") == 0)	// NIRVA specific orders
		{
			if(sCommandNameHere == "PRIVATE")
			{
				// The command comes from a browser client
				ProcessPrivateCommand();
			}
			else
			{
				if(sCommandNameHere == "NIRVA")
				{
					// The command comes from a NIRVA client (nvc)
					string sNvSourceHere = InputMessage.GetHeader("NV_SOURCE");
					if(sNvSourceHere == HTTP_CLIENT_STRING)
						DispatchNirvaCommand();
					else
						SetHttpError(HTTP_ERROR_BADREQUEST, "Unknown command");
				}
				else if(sCommandNameHere == "COMMAND")
				{
					// The command comes from a browser client
					BrowserFunction();
				}
				else if(sCommandNameHere == "XMLCOMMAND")
				{
					// The command comes from a browser client
					BrowserFunction(NV_HTTP_XML);
				}
				else if(sCommandNameHere == "SOAPCOMMAND")
				{
					// The command comes from a browser client
					BrowserFunction(NV_HTTP_SOAP);
				}
				else if(sCommandNameHere == "WEBS")
				{
					// The command comes from a browser client
					BrowserFunction(NV_HTTP_WEBS);
				}
				else if(sCommandNameHere == "WSDL")
				{
					// The command comes from a browser client
					GetWsdl();
				}
			}
		}
		else
		{
			if(IsStaticView)
			{
				if(!GetStaticView())
					SetHttpError(HTTP_ERROR_NOTFOUND, "Cannot get URL");
			}
			else
			{
				if(!SendFile())
					SetHttpError(HTTP_ERROR_NOTFOUND, "Cannot get URL");
			}
		}
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Http client thread exception occured (ProcessMessage)", NULL, NV_LOG_ERROR);
		SetHttpError(HTTP_ERROR_NOTFOUND, "Cannot get URL");
	}


	char buffer[20];
	NvInt64 BodySize = OutputMessage.GetWriteBodySize();
	sprintf(buffer, "%llu", BodySize);
	OutputMessage.AddHeader("Content-Length", buffer);
	
	// See if we force the http connection to close
	string sClose = OutputMessage.GetHeader("Connection");
	CleanString(sClose);
	MakeUpper(sClose);
	if(sClose == "CLOSE")
		Ret = false;

//	if(Ret)
//		OutputMessage.AddHeader("Connection", "Keep-Alive");

	// In case of redirection, we'll close this connection
	if(bRedirected)
		Ret = false;

	return Ret;
}

void NvHttpClient::SetHttpError(string HttpError, string HttpErrorDesc, bool ClearData)
{
	string sError = "HTTP/1.1 ";
	sError += HttpError;
	if(HttpErrorDesc.length())
	{
		sError += " - ";
		sError += HttpErrorDesc;
	}
	// we must clear all the output in case of error
	// Except for SOAP message
	if(HttpError != HTTP_ERROR_OK)
	{
		if(ClearData)
		{
			OutputMessage.Clear();
			NvData *Data = NVNEW NvData;
			OutputMessage.vDatas.push_back(Data);

			// We have no content in the body
			OutputMessage.AddHeader("Content-Length", "0");
		}

		// get current time and date
		time_t CurrentTime;
		time(&CurrentTime);
		struct tm tCurrentTime;
		gmtime_r(&CurrentTime, &tCurrentTime);
		char sTime[200];

//		strftime(sTime, sizeof(sTime), "%a, %d %b %Y %H:%M:%S GMT", &tCurrentTime);
		MyGetTime(sTime, &tCurrentTime, true);

		// and set it as the Date header
		OutputMessage.AddHeader("Date", sTime);

		// Set the server name header
		if(b_SendHttpServerHeader)
			OutputMessage.AddHeader("Server", HTTP_SERVER_STRING);
	}

	OutputMessage.vDatas[0]->sInfo = sError;
}


void NvHttpClient::SetHttpErrorBrowserCommand(int iOutputType, int XmlCoding, bool HideSessionInfo)
{
	// Allways set the session id (blank by default)
	OutputMessage.AddHeader("Nv_Session_Id", "");

	string sEncoding = "";
	if(Command.iXmlInputType)
		sEncoding = Command.m_sInputCharset;
	if(Command.Parameters.KeyExist("NV_XML_ENCODING"))
		sEncoding = Command.Parameters.GetValue("NV_XML_ENCODING");
	CleanString(sEncoding);
	MakeUpper(sEncoding);
	if((sEncoding.length()) && (sEncoding != "ISO-8859-1") && (sEncoding != "UTF-8"))
		sEncoding = "";
	bool bUnicode = false;
	if(!sEncoding.length())
	{
		if(GetMain()->IsUnicode())
			sEncoding = "UTF-8";
		else
			sEncoding = "ISO-8859-1";
	}
	if(sEncoding == "UTF-8")
		bUnicode = true;

	string AppName = "";
	NvSession* SessionHere = GetSession();
	if(SessionHere != NULL)
	{
		if(!HideSessionInfo)
			OutputMessage.AddHeader("Nv_Session_Id", SessionHere->m_sIdent);
		AppName = SessionHere->GetAppName();
	}
	if(!AppName.length())
	{
		if(InputMessage.ParameterExist("NV_APPLICATION"))
			AppName = InputMessage.GetParameter("NV_APPLICATION");
		else
		{
			// If the NV_APPLICATION parameter is not given, we try to get it from
			// the URL
			AppName = InputMessage.sPath;
			MakeUpper(AppName);
			string::size_type index = AppName.find("NV_APP_");
			if(index != string::npos)
			{
				AppName = AppName.substr(index+7);
				if(((index = AppName.find('/')) != string::npos) || ((index = AppName.find('\\')) != string::npos))
					AppName = AppName.substr(0, index);
			}
			else
				AppName = "";
		}
	}
	CleanString(AppName);
	MakeUpper(AppName);
	if(!AppName.length() && (iOutputType == NV_HTTP_WEBS))
	{
		// for web service, the application name is in the InputMessage.sPath /appname/webservice;
		AppName = InputMessage.sPath;
		if(AppName.length() && AppName[0] == '/')
			AppName = AppName.substr(1);
		string::size_type index = AppName.find('/');
		if(index != string::npos)
			AppName = AppName.substr(0,index);
		CleanString(AppName);
		MakeUpper(AppName);
	}

	if(!AppName.length())
		AppName = "NVDEF";
	
	string ErrorPage = "";
	string sHrend = InputMessage.GetParameter("NV_HREND");
	if(sHrend == "NVFM")
	{
		// This is a framework view request
		SetHttpErrorBrowserCommandFramework(&Command, AppName, sEncoding, bUnicode, XmlCoding, HideSessionInfo);
		return;
	}

	if(iOutputType == NV_HTTP_XML)
		SetHttpErrorBrowserCommandXml(ErrorPage, AppName, sEncoding, bUnicode, XmlCoding, HideSessionInfo);
	else if(iOutputType == NV_HTTP_SOAP)
		SetHttpErrorBrowserCommandSoap(ErrorPage, AppName, sEncoding, bUnicode, XmlCoding, HideSessionInfo);
	else if(iOutputType == NV_HTTP_WEBS)
		SetHttpErrorBrowserCommandSoap(ErrorPage, AppName, sEncoding, bUnicode, XmlCoding, HideSessionInfo);
	else
		SetHttpErrorBrowserCommandHtml(ErrorPage, AppName, sEncoding, bUnicode, XmlCoding, HideSessionInfo);

	OutputMessage.AddHeader("Cache-Control", "no-cache");
	OutputMessage.AddHeader("Pragma", "no-cache");
	OutputMessage.AddHeader("Expires", "Wed, 26 Feb 1997 08:21:57 GMT");

	OutputMessage.vDatas[0]->FreeData();
	OutputMessage.vDatas[0]->Data = (char *)malloc(ErrorPage.length());
	if(OutputMessage.vDatas[0]->Data != NULL)
	{
		OutputMessage.vDatas[0]->DataSize = (unsigned long)(ErrorPage.length());
		memcpy(OutputMessage.vDatas[0]->Data, ErrorPage.c_str(), ErrorPage.length());
	}
}


void NvHttpClient::SetHttpErrorBrowserCommandXml(string& ErrorPage, string AppName, string sEncoding, bool bUnicode, int XmlCoding, bool HideSessionInfo)
{
	NvXml XmlObjectForCoding;
	if(Command.Parameters.KeyExist("NV_XML_STRICT"))
	{
		string sStrict = Command.Parameters.GetValue("NV_XML_STRICT");
		CleanString(sStrict, true);
		MakeUpper(sStrict);
		if(sStrict == "YES")
			XmlObjectForCoding.bStrict = true;
	}

	// Write a simple XML data
	string NsPrefix = "";
	string NsHref = "";
	string NameSpaceHere = Command.m_sXmlNameSpace;
	if(NameSpaceHere.length())
	{
		string::size_type index;
		index = NameSpaceHere.find('=');
		if(index != string::npos)
		{
			if(index)
				NsPrefix = NameSpaceHere.substr(0, index);
			if(index < (NameSpaceHere.length()-1))
				NsHref = NameSpaceHere.substr(index+1);
		}
	}
	ErrorPage = "<?xml version=\"1.0\" encoding=\"";
	ErrorPage += sEncoding;
	ErrorPage += "\" ?>\n";
	ReserveString(ErrorPage, 1000);
	ErrorPage += "<";

	if(NsHref.length() && NsPrefix.length())
	{
		ErrorPage += NsPrefix;
		ErrorPage += ":";
	}
	if((XmlCoding == XML_MODEL_SIMPLE) || (XmlCoding == XML_MODEL_COMPACT) || (XmlCoding == XML_MODEL_TINY))
		ErrorPage += "nvfault";
	else
		ErrorPage += "NIRVA";
	if(NsHref.length())
	{
		ErrorPage += " xmlns";
		if(NsPrefix.length())
		{
			ErrorPage += ":";
			ErrorPage += NsPrefix;
		}
		ErrorPage += "=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(NsHref, bUnicode);
		ErrorPage += "\"";
	}
	if(NsPrefix.length())
		NsPrefix += ":";

	NvSession* SessionHere = GetSession();
	if((SessionHere != NULL) && (!HideSessionInfo))
	{
		ErrorPage += " session=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->m_sIdent, bUnicode);
		ErrorPage += "\" application=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(AppName, bUnicode);
		ErrorPage += "\" user=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->GetUserName(), bUnicode);
		ErrorPage += "\"";
	}
	else
	{
		ErrorPage += " application=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(AppName, bUnicode);
		ErrorPage += "\"";
	}
	ErrorPage += " language=\"";
	ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.m_sLanguage, bUnicode);
	ErrorPage += "\"";
	ErrorPage += ">\n";
	if((XmlCoding == XML_MODEL_SIMPLE) || (XmlCoding == XML_MODEL_COMPACT) || (XmlCoding == XML_MODEL_TINY))
	{
		ErrorPage += "\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "code>";
		char Buf[50];
		sprintf(Buf, "%d", Command.Error.m_iErrorCode);
		string sValue = Buf;
		ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "code>\n";
		ErrorPage += "\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "service>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "service>\n";
		ErrorPage += "\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "class>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "class>\n";
		ErrorPage += "\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "description>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "description>\n";
		ErrorPage += "\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "info>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "info>\n";
	}
	else
	{
		ErrorPage += "\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERROR>\n";
		ErrorPage += "\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCODE>";
		char Buf[50];
		sprintf(Buf, "%d", Command.Error.m_iErrorCode);
		string sValue = Buf;
		ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCODE>\n";
		ErrorPage += "\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORSERVICE>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORSERVICE>\n";
		ErrorPage += "\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCLASS>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCLASS>\n";
		ErrorPage += "\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORDESC>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORDESC>\n";
		ErrorPage += "\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORINFO>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORINFO>\n";
		ErrorPage += "\t</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERROR>\n";
	}
	ErrorPage += "</";
	ErrorPage += NsPrefix;
	if((XmlCoding == XML_MODEL_SIMPLE) || (XmlCoding == XML_MODEL_COMPACT) || (XmlCoding == XML_MODEL_TINY))
		ErrorPage += "nvfault>\n";
	else
		ErrorPage += "NIRVA>\n";
	OutputMessage.AddHeader("Content-Type", "text/xml" );
	string ErrorXsl = InputMessage.GetParameter("NV_XML_XSL_ERROR");
	CleanString(ErrorXsl);
	string XslFile = "";
	if(ErrorXsl.length())
	{
		// Get the real file name
		if(NvSession::GetXslFile(NULL, ErrorXsl, AppName))
			XslFile = ErrorXsl;
	}
	if(XslFile.length())
	{
		if(GetNoCaseRealFileName(XslFile))
		{
			NvXml XmlParser;
			unsigned long OutputDataSize = (unsigned long)(ErrorPage.length());
			char *OutputData = (char*)malloc(OutputDataSize+10);
			if(OutputData != NULL)
			{
				strcpy(OutputData, ErrorPage.c_str());
				if(XmlParser.XsltTransform(&OutputData, &OutputDataSize, XslFile.c_str()))
				{
					char *NewOutputData = (char*)malloc(OutputDataSize+10);
					if(NewOutputData != NULL)
					{
						memcpy(NewOutputData, OutputData, OutputDataSize);
						NewOutputData[OutputDataSize] = '\0';
						ErrorPage = NewOutputData;
						free(NewOutputData);
					}
				}
				free(OutputData);
			}
		}
	}
}


void NvHttpClient::SetHttpErrorBrowserCommandSoap(string& ErrorPage, string AppName, string sEncoding, bool bUnicode, int XmlCoding, bool HideSessionInfo)
{

	NvXml XmlObjectForCoding;
	if(Command.Parameters.KeyExist("NV_XML_STRICT"))
	{
		string sStrict = Command.Parameters.GetValue("NV_XML_STRICT");
		CleanString(sStrict, true);
		MakeUpper(sStrict);
		if(sStrict == "YES")
			XmlObjectForCoding.bStrict = true;
	}

	// Write a simple SOAP XML data
	string NsPrefix = "";
	string NsHref = "";
	string NameSpaceHere = "nvs=http://www.nirva-systems.com/2004-03-29/nirva-xml-fault";
	// Verify if we have a web service name and set the namespace according
	if(Command.bWebs)
	{
		string WebserviceName = Command.Parameters.GetValue("WEBSERVICE");
		CleanString(WebserviceName);
		MakeUpper(WebserviceName);
		if(WebserviceName.length())
		{
			string NsPrefixInWebs = "";
			NvWebs* WebService = GetMain()->Webservices.GetWebservice(WebserviceName.c_str());
			if(WebService != NULL)
			{
				WebserviceName = WebService->GetName();
				NsPrefixInWebs = WebService->GetNsPrefix();
				GetMain()->Webservices.FreeWebservice(WebserviceName.c_str());
			}
			string Host = GetMain()->GetHostWebs();
			MakeLower(WebserviceName);
			string AppNameHere = AppName;
			if(!AppNameHere.length())
				AppNameHere = "nvdef";
			NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
			if(Application != NULL)
			{
				AppNameHere = Application->GetAppNameForFiles();
				GetMain()->Applications.FreeApplication(AppName.c_str());
			}
			MakeLower(AppNameHere);
			NameSpaceHere = "nvs=http://";
			if(NsPrefixInWebs.length())
				NameSpaceHere += NsPrefixInWebs;
			else
				NameSpaceHere += Host;
			NameSpaceHere += "/nirva/";
			NameSpaceHere += AppNameHere;
			NameSpaceHere += "/";
			NameSpaceHere += WebserviceName;
		}
	}

	if(NameSpaceHere.length())
	{
		string::size_type index;
		index = NameSpaceHere.find('=');
		if(index != string::npos)
		{
			if(index)
				NsPrefix = NameSpaceHere.substr(0, index);
			if(index < (NameSpaceHere.length()-1))
				NsHref = NameSpaceHere.substr(index+1);
		}
	}
	
	ErrorPage = "<?xml version=\"1.0\" encoding=\"";
	ErrorPage += sEncoding;
	ErrorPage += "\" ?>\n";
	ReserveString(ErrorPage, 2000);
	ErrorPage += "<env:Envelope xmlns:env=\"http://schemas.xmlsoap.org/soap/envelope/\">\n";
	ErrorPage += "\t<env:Body>\n";
	ErrorPage += "\t<env:Fault>\n";
	ErrorPage += "\t\t<faultcode>env:Server</faultcode>\n";
	ErrorPage += "\t\t<faultstring>";
	ErrorPage += "NIRVA-ERROR:";
	ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
	ErrorPage += ":";
	ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
	ErrorPage += ":";
	char Buf[50];
	sprintf(Buf, "%d", Command.Error.m_iErrorCode);
	string sValue = Buf;
	ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
	ErrorPage += ":";
	ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
	if(Command.Error.m_sErrorInfo.length())
	{
		ErrorPage += ":";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
	}
	ErrorPage += "\t\t</faultstring>\n";
	ErrorPage += "\t\t<faultactor>";
	ErrorPage += Command.m_sHost;
	ErrorPage += "\t\t</faultactor>\n";
	ErrorPage += "\t\t<detail>\n";

	ErrorPage += "\t\t\t<";

	if(NsHref.length() && NsPrefix.length())
	{
		ErrorPage += NsPrefix;
		ErrorPage += ":";
	}
	if((XmlCoding == XML_MODEL_SIMPLE) || (XmlCoding == XML_MODEL_COMPACT) || (XmlCoding == XML_MODEL_TINY))
		ErrorPage += "nvfault";
	else
		ErrorPage += "NVERROR";
	if(NsHref.length())
	{
		ErrorPage += " xmlns";
		if(NsPrefix.length())
		{
			ErrorPage += ":";
			ErrorPage += NsPrefix;
		}
		ErrorPage += "=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(NsHref, bUnicode);
		ErrorPage += "\"";
	}
	if(NsPrefix.length())
		NsPrefix += ":";

	NvSession* SessionHere = GetSession();
	if((SessionHere != NULL) && (!HideSessionInfo))
	{
		ErrorPage += " session=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->m_sIdent, bUnicode);
		ErrorPage += "\" application=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(AppName, bUnicode);
		ErrorPage += "\" user=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->GetUserName(), bUnicode);
		ErrorPage += "\" language=\"";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.m_sLanguage, bUnicode);
		ErrorPage += "\"";
	}
	ErrorPage += ">\n";
	if((XmlCoding == XML_MODEL_SIMPLE) || (XmlCoding == XML_MODEL_COMPACT) || (XmlCoding == XML_MODEL_TINY))
	{
		ErrorPage += "\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "code>";
		sprintf(Buf, "%d", Command.Error.m_iErrorCode);
		sValue = Buf;
		ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "code>\n";
		ErrorPage += "\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "service>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "service>\n";
		ErrorPage += "\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "class>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "class>\n";
		ErrorPage += "\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "description>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "description>\n";
		ErrorPage += "\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "info>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "info>\n";
	}
	else
	{
		ErrorPage += "\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERROR>\n";
		ErrorPage += "\t\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCODE>";
		sprintf(Buf, "%d", Command.Error.m_iErrorCode);
		sValue = Buf;
		ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCODE>\n";
		ErrorPage += "\t\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORSERVICE>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORSERVICE>\n";
		ErrorPage += "\t\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCLASS>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORCLASS>\n";
		ErrorPage += "\t\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORDESC>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORDESC>\n";
		ErrorPage += "\t\t\t\t\t<";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORINFO>";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERRORINFO>\n";
		ErrorPage += "\t\t\t\t</";
		ErrorPage += NsPrefix;
		ErrorPage += "NVERROR>\n";
	}
	ErrorPage += "\t\t\t</";
	ErrorPage += NsPrefix;
	if((XmlCoding == XML_MODEL_SIMPLE) || (XmlCoding == XML_MODEL_COMPACT) || (XmlCoding == XML_MODEL_TINY))
		ErrorPage += "nvfault>\n";
	else
		ErrorPage += "NVERROR>\n";
	ErrorPage += "\t\t</detail>\n";
	ErrorPage += "\t</env:Fault>\n";
	ErrorPage += "\t</env:Body>\n";
	ErrorPage += "</env:Envelope>\n";
	OutputMessage.AddHeader("Content-Type", "text/xml" );
	SetHttpError(HTTP_ERROR_INTERNAL, "NIRVA SOAP server internal error", false);

}


void NvHttpClient::SetHttpErrorBrowserCommandHtml(string& ErrorPage, string AppName, string sEncoding, bool bUnicode, int XmlCoding, bool HideSessionInfo)
{
	// Check if there is an XSL file for error display
	string ErrorXsl = InputMessage.GetParameter("NV_XML_XSL_ERROR");
	if(!ErrorXsl.length())
	{
		// no xsl error explicitly given so we try to see if we have a default one at session level and then in application File directory
		// Check if we have a session level xsl
		NvSession* SessionHere = GetSession();
		if(SessionHere != NULL)
		{
			ErrorXsl = SessionHere->GetErrorXsl();
			if(ErrorXsl.length())
				InputMessage.mParameters["NV_XML_XSL_ERROR"] = ErrorXsl;
		}

		// if nothing at session level check if we have an error.xsl in application File directory
		if(!ErrorXsl.length())
		{
			NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
			if(Application != NULL)
			{
				string XslFile = "";
				XslFile = Application->GetFileDir();
				GetMain()->Applications.FreeApplication(AppName.c_str());
				XslFile += "error.xsl";
				InputMessage.mParameters["NV_XML_XSL_ERROR"] = "error.xsl";
				if(GetNoCaseRealFileName(XslFile))
				{
					InputMessage.mParameters["NV_XML_XSL_ERROR"] = "error.xsl";
					ErrorXsl="error.xsl";
				}
			}
		}
	}
	if(ErrorXsl.length())
	{
		// Always simple coding is used
		SetHttpErrorBrowserCommandXml(ErrorPage, AppName, sEncoding, bUnicode, XML_MODEL_SIMPLE, HideSessionInfo);
		OutputMessage.AddHeader("Content-Type", "text/html" );
		return;
	}

	// No default xsl found for error so we just send an html error message

	// get current time and date
	time_t CurrentTime;
	time(&CurrentTime);
	struct tm tCurrentTime;
	char sTime[200];
	if(GetMain()->GetLocalTime())
	{
		localtime_r(&CurrentTime, &tCurrentTime);
//		strftime(sTime, sizeof(sTime), "%a, %d %b %Y %H:%M:%S", &tCurrentTime);
		MyGetTime(sTime, &tCurrentTime, false);
	}
	else
	{
		gmtime_r(&CurrentTime, &tCurrentTime);
//		strftime(sTime, sizeof(sTime), "%a, %d %b %Y %H:%M:%S GMT", &tCurrentTime);
		MyGetTime(sTime, &tCurrentTime, true);
	}

	char Buf[50];
	sprintf(Buf, "%d", Command.Error.m_iErrorCode);
	// We prepare error information
	string Info = "Error ";
	Info += Command.Error.m_sErrorService;
	Info += ":";
	Info += Command.Error.m_sErrorClass;
	Info += ":";
	Info += Buf;
	Info += " (";
	Info += Command.Error.GetErrorDesc(Command.m_sLanguage.c_str());
	Info += ")";
	if(Command.Error.m_sErrorInfo.length())
	{
		Info += " : ";
		Info += Command.Error.m_sErrorInfo;
	}
	
	// We use an xml encoding here in order to avai < and > in result message that
	// can be eventually interpreted by the browser (ex <script>). Security issue
	NvXml XmlObjectForCoding;
	Info = XmlObjectForCoding.EncodeOutputString(Info, bUnicode);
//	EncodeUrlString(Info, bUnicode, false);

	string DefaultCssPage = "http://";
	string XForwardedHost = InputMessage.GetHeader("X-FORWARDED-HOST");
	if(XForwardedHost.length())
		DefaultCssPage += XForwardedHost;
	else
		DefaultCssPage += InputMessage.GetHeader("HOST");
	if(InputMessage.sPath.length() && (InputMessage.sPath[InputMessage.sPath.length()-1] != '\\'))
		if(InputMessage.sPath.length() && (InputMessage.sPath[InputMessage.sPath.length()-1] != '/'))
			DefaultCssPage += "/";
	DefaultCssPage += "Error/";
	DefaultCssPage += "error.css";

	// and write a simple HTML page
	ErrorPage = "<html>\r\n";
	ErrorPage += "<head>\n";
	ErrorPage += "<link rel=\"stylesheet\" type=\"text/css\" href=\"";
//	ErrorPage += DefaultCssPage;
	ErrorPage += "error.css";
	ErrorPage += "\"/>\r\n";
	ErrorPage += "<title>Nirva error</title>\r\n";
	ErrorPage += "</head>\r\n";
	ErrorPage += "<body>\r\n";
	ErrorPage += "<a href=\"javascript:history.back()\">Back</a>\r\n";
	ErrorPage += "<h3>NIRVA Server";
	ErrorPage += " : ";
	ErrorPage += sTime;
	ErrorPage += "</h3>\r\n";
	ErrorPage += "<h3>NIRVA Server error";
	ErrorPage += " : ";
	ErrorPage += Info;
	ErrorPage += "</h3>\r\n";
	ErrorPage += "</body>\r\n";
	ErrorPage += "</html>\r\n";

	string ContentType = "text/html; charset=";
	ContentType += sEncoding;
	MakeLower(ContentType);
	OutputMessage.AddHeader("Content-Type", ContentType);
}




void NvHttpClient::SetHttpErrorBrowserCommandFramework(NvCommand *CommandForError, string AppName, string sEncoding, bool bUnicode, int XmlCoding, bool HideSessionInfo)
{
	string ErrorPage = "";

	// Check if there is an error view and if not, display a normal html error
	bool bGotErrorView = false;

	NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
	if((Application != NULL) && (Application->IsEnable()) && (Application->GetUseFramework()))
	{
		OutputMessage.vDatas[0]->FreeData();
		NvData *Data = OutputMessage.vDatas[0];
		NvSession* SessionHere = GetSession();
		string SessionIdHere = "";
		if(SessionHere != NULL)
			SessionIdHere = SessionHere->m_sIdent;

		// Get language
		string LanguageHere = InputMessage.GetParameter("NV_LANGUAGE");
		MakeUpper(LanguageHere);
		CleanString(LanguageHere);
		if(!LanguageHere.length())
			if(SessionHere != NULL)
				LanguageHere = SessionHere->m_sLanguage;

		// Get optional error view that eventually overrides the default framework error view
		string ErrorView = InputMessage.GetParameter("NV_HREND_ERROR_PAGE");

		bool RetVal = false;
		try
		{
			NvCommand CommandToExecute;
			CommandToExecute.m_sLanguage = LanguageHere;
			CommandToExecute.Error.SetError(CommandForError->Error.m_sErrorService, CommandForError->Error.m_sErrorClass, CommandForError->Error.m_iErrorCode, CommandForError->Error.m_sErrorInfo);
			RetVal = Application->NidgetFramework.GetView(ErrorView.c_str(), SessionIdHere.c_str(), &CommandToExecute, &(Data->Data), &(Data->DataSize), GetMain()->GetDebug(), true);
		}
		catch(...)
		{
			GetMain()->IncNumExceptions();
			GetMain()->Logs.Write("Exceptions", "Exception occured in get framework view", NULL, NV_LOG_ERROR);
			RetVal = false;
		}
		GetMain()->Applications.FreeApplication(AppName.c_str());
		if(RetVal)
		{
			OutputMessage.AddHeader("Cache-Control", "no-cache");
			OutputMessage.AddHeader("Pragma", "no-cache");
			OutputMessage.AddHeader("Expires", "Wed, 26 Feb 1997 08:21:57 GMT");
			bGotErrorView = true;
		}
	}

	if(!bGotErrorView)
	{
		SetHttpErrorBrowserCommandHtml(ErrorPage, AppName, sEncoding, bUnicode, XmlCoding, HideSessionInfo);
		OutputMessage.AddHeader("Cache-Control", "no-cache");
		OutputMessage.AddHeader("Pragma", "no-cache");
		OutputMessage.AddHeader("Expires", "Wed, 26 Feb 1997 08:21:57 GMT");

		OutputMessage.vDatas[0]->FreeData();
		OutputMessage.vDatas[0]->Data = (char *)malloc(ErrorPage.length());
		if(OutputMessage.vDatas[0]->Data != NULL)
		{
			OutputMessage.vDatas[0]->DataSize = (unsigned long)(ErrorPage.length());
			memcpy(OutputMessage.vDatas[0]->Data, ErrorPage.c_str(), ErrorPage.length());
		}
	}
}


void NvHttpClient::FormatDate(string& DateToFormat)
{
	if(DateToFormat.length() == 8)
	{
		string provistr = DateToFormat.substr(0, 4);
		provistr += "-";
		provistr += DateToFormat.substr(4, 2);
		provistr += "-";
		provistr += DateToFormat.substr(6);
		DateToFormat = provistr;
	}
}

	
void NvHttpClient::FormatTime(string& TimeToFormat)
{
	if(TimeToFormat.length() == 6)
	{
		string provistr = TimeToFormat.substr(0, 2);
		provistr += ":";
		provistr += TimeToFormat.substr(2, 2);
		provistr += ":";
		provistr += TimeToFormat.substr(4);
		TimeToFormat = provistr;
	}
}

	
void NvHttpClient::SetCommandError(int IndexPart)
{
	// We set the part error headers
	
	char sErrorCode[20];
	sprintf(sErrorCode, "%d", Command.Error.m_iErrorCode);
	OutputMessage.AddHeader("Nv_Error_Code", sErrorCode, IndexPart);
	OutputMessage.AddHeader("Nv_Error_Service", Command.Error.m_sErrorService, IndexPart);
	OutputMessage.AddHeader("Nv_Error_Class", Command.Error.m_sErrorClass, IndexPart);
	OutputMessage.AddHeader("Nv_Error_Desc", Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), IndexPart);
	OutputMessage.AddHeader("Nv_Error_Info", Command.Error.m_sErrorInfo, IndexPart);
	
	// We also set the global NvError header to yes if there is one error in one part
	if(Command.Error.m_iErrorCode != NV_OK)
		OutputMessage.AddHeader("Nv_Error", "yes");

}

void NvHttpClient::DispatchNirvaCommand()
{
	RemoveSessionOnCleanup = false;

	// Get the source machine name and user (we do that before any decoding in order to have the same encoding than on the client
	// On client side machine and user names are supposed to be always utf8 encoded
	string SourceMachineNameUtf8 = InputMessage.GetHeader("NV_MACHINE_NAME");
	string SourceMachineUserUtf8 = InputMessage.GetHeader("NV_MACHINE_USER");

	// we have to convert all input to internal server coding
	bool bClientUnicode = false;
	if(InputMessage.GetHeader("NV_CLIENT_UNICODE") == "YES")
		bClientUnicode = true;
	bool bDoConversion = true;
	string sNoConversion = InputMessage.GetHeader("NV_NO_CONVERSION");
	CleanString(sNoConversion);
	MakeUpper(sNoConversion);
	if(sNoConversion == "YES")
		bDoConversion = false;
	if(bDoConversion && (GetMain()->IsUnicode() != bClientUnicode))
	{
		// We must convert all input data
		int i;
		for(i=0; i<(int)(InputMessage.vDatas.size()); i++)
		{
			NvData *Data = InputMessage.vDatas[i];
			std::map<std::string, std::string> mNewHeaders;	
			map<string,string>::iterator pos;
			for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
			{
				string ParameterNameHere = pos->first;
				string ParameterValueHere = pos->second;
				DecodeUrlString(ParameterNameHere, bClientUnicode, false);
				DecodeUrlString(ParameterValueHere, bClientUnicode, false);
				mNewHeaders[ParameterNameHere] = ParameterValueHere;
			}
			Data->mHeaders.clear();
			Data->mHeaders = mNewHeaders;
		}
	}


	// Get the source machine name and user that are supposed to be always utf8
	string SourceMachineName = SourceMachineNameUtf8;
	string SourceMachineUser = SourceMachineUserUtf8;
	DecodeUrlString(SourceMachineName, true, false);
	DecodeUrlString(SourceMachineUser, true, false);
	InputMessage.vDatas[0]->mHeaders["NV_MACHINE_NAME"] = SourceMachineName;
	InputMessage.vDatas[0]->mHeaders["NV_MACHINE_USER"] = SourceMachineUser;

	// Cleanup the actual command
	Command.Clear();
	Command.m_sMachineName = SourceMachineName;
	Command.m_sMachineUser = SourceMachineUser;
	Command.m_sClientAddress = IpAddress;
	Command.bClientUnicode = bClientUnicode;
	Command.m_sHost = InputMessage.GetHeader("X-FORWARDED-HOST");
	if(!Command.m_sHost.length())
		Command.m_sHost = InputMessage.GetHeader("HOST");


	// We prepare the output global headers
	string sContentType = "multipart/form-data; boundary=";
	sContentType += HTTP_SEND_BOUNDARY;

	OutputMessage.AddHeader("Content-Type", sContentType);
	OutputMessage.AddHeader("Nv_Error", "no");

	// The other output headers (Date and Server) are filled from the calling function

	// prepare the first part
	// The first return part always concern information about the session

	NvData *Data = NVNEW NvData;
	OutputMessage.vDatas.push_back(Data);
	string provistr = "\r\n--";
	provistr += HTTP_SEND_BOUNDARY;
	OutputMessage.vDatas[1]->sInfo = provistr;
	
	// The session ID is blank by default
	OutputMessage.AddHeader("Nv_Session_Id", "", 1);
//	OutputMessage.AddHeader("Nv_Service", "SYSTEM", 1);
//	OutputMessage.AddHeader("Nv_Class", "SESSION", 1);
//	OutputMessage.AddHeader("Nv_Command", "GET", 1);

	// No error for now in the part 1
	SetCommandError(1);

	// Get the session ID from the parameters 
	string SessionIdHere = InputMessage.GetHeader("NV_SESSION_ID");
	CleanString(SessionIdHere);
	if(SessionIdHere.length() && (SessionIdHere[0] == '.'))
		SessionIdHere = "";


	// The nolock mode is used for session that call an external program by the SYSTEM EXEC command
	// that itself reconnect to the same session. In this case, the session should not lock the command.

	bool LockSession = true;
	
	// And try to get the session

	NvSession *Session = GetMain()->Sessions.GetSession(SessionIdHere.c_str());
	SetSession(Session);

	// We get the langage because we need it for session login
	string LanguageHere = InputMessage.GetHeader("NV_LANGUAGE");
	MakeUpper(LanguageHere);
	CleanString(LanguageHere);
	if(!LanguageHere.length())
	{
		if(Session != NULL)
			LanguageHere = Session->m_sLanguage;
		else
			LanguageHere = "ENGLISH";
	}
	
	bool bIsCommandAttachConsole = false;
	bool ExplicitClose = false;
	bool bGotLogin = false;
	if(Session == NULL)
	{
		// First test if requested to create a new session if the actual one is expired
		string NewIfExpired = InputMessage.GetHeader("NV_NEW_IF_EXPIRED");
		CleanString(NewIfExpired);
		MakeUpper(NewIfExpired);
		if(NewIfExpired == "YES")
		{
			InputMessage.mParameters.erase("NV_SESSION_ID");
			SessionIdHere = "";
		}

		// The session is not found
		if(SessionIdHere.length())
		{
			// The requested session was not blank so it must exist
			Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_NOTEXIST, SessionIdHere);
		}
		else
		{
			// The requested session is blank (or not given)
			// So we try to create it
			if(GetMain()->CheckOverTime())
			{
				Command.Error.SetError("SYSTEM", "SYSTEM", NV_SYSTEM_OVER_TIME);
			}
			else
			{
				Session = GetMain()->Sessions.GetSession(NULL);
				SetSession(Session);
				if(Session == NULL)
				{
					Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTCREATE);
					int MaxSessions = GetMain()->GetMaxSessions();
					if(MaxSessions != -1)
					{
						if(GetMain()->Sessions.GetNumSessions() >= MaxSessions)
						{
							string AppliNameHere = InputMessage.GetHeader("NV_APPLICATION");
							CleanString(AppliNameHere);
							MakeUpper(AppliNameHere);
							if(!AppliNameHere.length())
								AppliNameHere = "NVDEF";
							string UserHere = InputMessage.GetHeader("NV_USER");
							CleanString(UserHere);
							if(!UserHere.length())
								UserHere = "nvdef";
							MakeLower(UserHere);
							if((AppliNameHere == "NVDEF") && (UserHere == "nvadmin"))
							{
								Session = GetMain()->Sessions.GetSession(NULL, false);
								SetSession(Session);
								if(Session != NULL)
								{
									// This session will only be authorized to list and remove sessions
									Session->m_bOverMaxSession = true;
									Command.Error.SetError();
								}
							}
							else
								Command.Error.SetError("SYSTEM", "SYSTEM", NV_SYSTEM_MAX_SESSIONS);
						}
					}
				}
			}
			if(Session != NULL)
			{
				// The session creation was OK
				// Now, we must login to it
				bool bHidden = false;
				if(InputMessage.GetHeader("NV_HIDDEN") == "YES")
					bHidden = true;
				Session->bHidden = bHidden;
				string AppliNameHere = InputMessage.GetHeader("NV_APPLICATION");
				CleanString(AppliNameHere);
				string UserHere = InputMessage.GetHeader("NV_USER");
				CleanString(UserHere);
				string PasswordHere = InputMessage.GetHeader("NV_PASSWORD");
				CleanString(PasswordHere);
				if(PasswordHere.length() && (PasswordHere.length() < 256))
				{
					// The string must be decrypted
					string EncryptionKey = SourceMachineNameUtf8;
					EncryptionKey += "nvc";
					EncryptionKey += AppliNameHere;
					MakeLower(EncryptionKey);
					int BufferLength = 100 + (4*((int)(PasswordHere.length())));
					char *Buffer = (char *)malloc(BufferLength);
					strcpy(Buffer, PasswordHere.c_str());
					char *Buffer2 = (char *)malloc(4*EncryptionKey.length());
					strcpy(Buffer2, EncryptionKey.c_str());
					DecryptString(Buffer2, Buffer, BufferLength);
					PasswordHere = Buffer;
					free(Buffer);
					free(Buffer2);
				}
				bool bWithNewPassword = false;
				if(InputMessage.HeaderExist("NV_NEW_PASSWORD"))
					bWithNewPassword = true;
				string NewPasswordHere = InputMessage.GetHeader("NV_NEW_PASSWORD");
				CleanString(NewPasswordHere);
				if(NewPasswordHere.length() && (NewPasswordHere.length() < 256))
				{
					// The string must be decrypted
					string EncryptionKey = SourceMachineNameUtf8;
					EncryptionKey += "nvc";
					EncryptionKey += AppliNameHere;
					MakeLower(EncryptionKey);

					int BufferLength = 100 + (4*((int)(NewPasswordHere.length())));
					char *Buffer = (char *)malloc(BufferLength);
					strcpy(Buffer, NewPasswordHere.c_str());
					char *Buffer2 = (char *)malloc(4*EncryptionKey.length());
					strcpy(Buffer2, EncryptionKey.c_str());
					DecryptString(Buffer2, Buffer, BufferLength);
					NewPasswordHere = Buffer;
					free(Buffer);
					free(Buffer2);
				}
				string NewPasswordHereConfirm = InputMessage.GetHeader("NV_NEW_PASSWORD_CONFIRM");
				CleanString(NewPasswordHereConfirm);
				if(NewPasswordHereConfirm.length() && (NewPasswordHereConfirm.length() < 256))
				{
					// The string must be decrypted
					string EncryptionKey = SourceMachineNameUtf8;
					EncryptionKey += "nvc";
					EncryptionKey += AppliNameHere;
					MakeLower(EncryptionKey);

					int BufferLength = 100 + (4*((int)(NewPasswordHereConfirm.length())));
					char *Buffer = (char *)malloc(BufferLength);
					strcpy(Buffer, NewPasswordHereConfirm.c_str());
					char *Buffer2 = (char *)malloc(4*EncryptionKey.length());
					strcpy(Buffer2, EncryptionKey.c_str());
					DecryptString(Buffer2, Buffer, BufferLength);
					NewPasswordHereConfirm = Buffer;
					free(Buffer);
					free(Buffer2);
				}
				bool bWithNewPasswordConfirm = false;
				if(bWithNewPassword)
				{
					if(InputMessage.HeaderExist("NV_NEW_PASSWORD_CONFIRM"))
						bWithNewPasswordConfirm = true;
				}
				string Hash = AppliNameHere + UserHere + PasswordHere;
				Hash = MD5_string(Hash);
				MakeUpper(Hash);
				string sSessionTimeOut = InputMessage.GetHeader("NV_TIMEOUT");
				CleanString(sSessionTimeOut);
				int iSessionTimeOut = atoi(sSessionTimeOut.c_str());
				if(iSessionTimeOut < 0)
					iSessionTimeOut = 0;
				string OpenProc = InputMessage.GetHeader("NV_SESSION_OPEN");
				CleanString(OpenProc);
				if(!OpenProc.length())
					OpenProc = "session_open";
				if(OpenProc == "NV_SESSION_OPEN_NONE")
					OpenProc = "";
				Session->m_sOpenProc = OpenProc;
				string CloseProc = InputMessage.GetHeader("NV_SESSION_CLOSE");
				CleanString(CloseProc);
				if(!CloseProc.length())
					CloseProc = "session_close";
				if(CloseProc == "NV_SESSION_CLOSE_NONE")
					CloseProc = "";
				Session->m_sCloseProc = CloseProc;
				if(!UserHere.length())
					UserHere = "nvdef";

				// We use the hash and hash2 parameters to control
				// if the login request comes from another nirva server and
				// from a transaction (or scheduler)
				string HashHere = InputMessage.GetHeader("NV_CONTROL");
				string Hash2 = HashHere;
				Hash2 += "CTRL:NV_NO_CTRL";
				Hash2 = MD5_string(Hash2);
				MakeUpper(Hash2);

				string HashHere2 = InputMessage.GetHeader("NV_CONTROL2");
				if(HashHere2 == Hash2)
					Session->m_bFromTransaction = true;

				// For the login, we must transform all headers of the first command to parameters
				// in order for the session_open proc to access them
				if(InputMessage.vDatas.size() > 1)
				{
					NvData *Data = InputMessage.vDatas[1];
					map<string,string>::iterator pos;
					for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
					{
						string ParameterNameHere = pos->first;
						if(ParameterNameHere == "CONTENT-TYPE")
							continue;
						if(ParameterNameHere == "NV_SERVICE")
							continue;
						if(ParameterNameHere == "NV_CLASS")
							continue;
						if(ParameterNameHere == "NV_COMMAND")
							continue;
						if(ParameterNameHere == "NV_CMD")
							continue;
						if(ParameterNameHere == "NV_IN_CONTAINER")
							continue;
						if(ParameterNameHere == "NV_OUT_CONTAINER")
							continue;
						if(ParameterNameHere == "NV_LANGUAGE")
							continue;
						string ParameterValueHere = pos->second;
						InputMessage.mParameters[ParameterNameHere] = ParameterValueHere;
					}
				}

				string SSOUser ="";
				string SSODomain ="";
				string SSOErrInfo ="";
				int SSORetVal = SSOAuthenticate(AppliNameHere, UserHere, SSOUser, SSODomain, true, SSOErrInfo);	
				if(SSORetVal <= 0)
				{
					// SSO failed
					// We prepare error information

					// We need user name in session for error message
					// In case of SSO, the UserName is eventually the SSOUser name
					NvApplication *Application = GetMain()->Applications.GetApplication(AppliNameHere.c_str());
					if(Application != NULL)
					{
						if(Application->GetNirvaIsSsoUser())
							UserHere = SSOUser;
						GetMain()->Applications.FreeApplication(AppliNameHere.c_str());
					}
					Session->SetUserName(UserHere.c_str());
					string Info = UserHere;
					Info += " - ";
					Info += AppliNameHere;
					if(SSORetVal == 0 && SSOErrInfo.length())
					{
						Info += " - ";
						Info += SSOErrInfo;
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_AUTHENTICATION_ERROR, Info);
					}
					if(SSORetVal == -4)
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_APPLICATION, Info);

					Session->SetUserName("");
					
					// We free the session
					GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

					// And remove it
					GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
					Session = NULL;
					SetSession(Session);

					if(SSORetVal == -200)
					{
						SetHttpError(HTTP_ERROR_UNAUTHORIZED, "", false);
						return;
					}
				}
				else
				{
					// SSO stuff was ok so we can login
					bool bSSO = false;
					if(SSORetVal == 2)
						bSSO = true;

					// Special managment for nvd be able to attach session to its console
					// If nvd request login we must do the attachment before login
					// in order to trace also the session open procedure
					// So the attachment is done before openiong the session
					if(((IpAddress == "127.0.0.1") || (IpAddress == "::1")) && InputMessage.vDatas.size() > 1)
					{
						string CommandHere = InputMessage.GetHeader("NV_CMD", 1);
						MakeUpper(CommandHere);
						CleanString(CommandHere);
						if(CommandHere == "SESSION:ATTACH_CONSOLE")
							bIsCommandAttachConsole = true;
						if(bIsCommandAttachConsole)
						{
							string  ClientPid = InputMessage.GetHeader("PID", 1);
							Command.Parameters.SetValue("PID", ClientPid);
							int iLevel = 1;
							string sLevel = InputMessage.GetHeader("LEVEL", 1);
							Command.Parameters.SetValue("LEVEL", sLevel);
							CleanString(sLevel);
							if(sLevel.length())
								iLevel = atoi(sLevel.c_str());
							if(iLevel < 1)
								iLevel = 1;
							Command.m_sClientAddress = IpAddress;
							string Type = InputMessage.GetHeader("TYPE", 1);
							MakeUpper(Type);
							CleanString(Type);
							if(Type.length())
							{
								// Get Parameters
								string Application = InputMessage.GetHeader("NV_APPLICATION");
								int iType = ATTACH_REQUEST_TYPE_ANY;
								if(Type == "CLIENT")
									iType = ATTACH_REQUEST_TYPE_CLIENT;
								else if(Type == "WEB")
									iType = ATTACH_REQUEST_TYPE_WEB;
								else if(Type == "WEBS")
									iType = ATTACH_REQUEST_TYPE_WEBS;
								else if(Type == "XML")
									iType = ATTACH_REQUEST_TYPE_XML;
								else if(Type == "SOAP")
									iType = ATTACH_REQUEST_TYPE_SOAP;
								else if(Type == "TRANSACTION")
									iType = ATTACH_REQUEST_TYPE_TRANSACTION;
								else if(Type == "SCHEDULER")
									iType = ATTACH_REQUEST_TYPE_SCHEDULER;
								else if(Type == "NAMED")
									iType = ATTACH_REQUEST_TYPE_NAMED;
								else if(Type == "LISTENER")
									iType = ATTACH_REQUEST_TYPE_LISTENER;
								else if(Type == "THREAD")
									iType = ATTACH_REQUEST_TYPE_THREAD;
								else if(Type == "INTERNAL")
									iType = ATTACH_REQUEST_TYPE_INTERNAL;
								else if(Type == "MQ")
									iType = ATTACH_REQUEST_TYPE_MQ;
								// We explicitly close the session in this case
								// This flag allows to not returning a session close error to nvd
								ExplicitClose = true;
								if(!GetMain()->AttachRequestList.Add(Session, &Command, atoi(ClientPid.c_str()), iType, Application, iLevel))
									Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE);

								// The request is to attach a future session
								// We remove the session because we don't need it
								Session->SetUserName("");
								
								// We free the session
								GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

								// And remove it
								GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
								Session = NULL;
								SetSession(Session);

							}
							else
							{
								// The request is to open a session and to attach it
								if(!Session->SessionAttachConsole(&Command))
								{
									Session->SetUserName("");
									
									// We free the session
									GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

									// And remove it
									GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
									Session = NULL;
									SetSession(Session);
								}
							}
						}
					}
					if(Session != NULL)
					{
						// Attach the session to an nvd console if there is such request
						GetMain()->AttachRequestList.Get(Session, ATTACH_REQUEST_TYPE_CLIENT, AppliNameHere);
						int LoginRetVal = Session->Login(UserHere.c_str(), PasswordHere.c_str(), AppliNameHere.c_str(), SESSION_TYPE_CLIENT, NewPasswordHere.c_str(), NewPasswordHereConfirm.c_str(), bWithNewPassword, bWithNewPasswordConfirm, iSessionTimeOut, Hash.c_str(), LanguageHere.c_str(), &InputMessage.mParameters, &Command, bSSO, SSOUser.c_str(), SSODomain.c_str());
						if(LoginRetVal != 1)
						{
							// Login failed
							// We prepare error information

							// We need user name in session for error message
							// In case of SSO, the UserName is eventually the SSOUser name
							if(bSSO)
							{
								NvApplication *Application = GetMain()->Applications.GetApplication(AppliNameHere.c_str());
								if(Application != NULL)
								{
									if(Application->GetNirvaIsSsoUser())
										UserHere = SSOUser;
									GetMain()->Applications.FreeApplication(AppliNameHere.c_str());
								}
							}
							Session->SetUserName(UserHere.c_str());
							string Info = UserHere;
							Info += " - ";
							Info += AppliNameHere;
							if(LoginRetVal == 0)
								Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTLOGIN, Info);
							if(LoginRetVal == -1)
								Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_USER, Info);
							if(LoginRetVal == -2)
								Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CANNOT_GET_CONTEXT, Info);
							if(LoginRetVal == -3)
								Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD, Info);
							if(LoginRetVal == -4)
								Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_APPLICATION, Info);
							if(LoginRetVal == -5)
								Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CHANGE_PASSWORD, Info);
							if(LoginRetVal == -6)
								Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD_SYNTAX, Info);

							Session->SetUserName("");
							
							// We free the session
							GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

							// And remove it
							GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
							Session = NULL;
							SetSession(Session);
						}
						else
							bGotLogin = true;
					}
				}
//				string UserHereForCheck = UserHere;
//				CleanString(UserHereForCheck, true);
//				MakeLower(UserHereForCheck);
//				if(Session != NULL)
//					if(UserHereForCheck == "nvadmin")
//						Session->m_bAdministrator = true; // Administrator user
			}
		}
	}
	else
	{
		// The session has been found
		// We must verify that the order has been sent by the correct user
		// We do that only if the order doesn't come from the local machine

		if((IpAddress != "127.0.0.1") && (IpAddress != "::1"))
		{
			string HashHere = InputMessage.GetHeader("NV_CONTROL");
			string HashSession = Session->GetHash();
			if(HashHere.length() && HashSession.length())
			{
				if(HashHere != HashSession)
				{
					// Humm... somebodys is trying to do bad things
					Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_IDENTIFICATION_FAILED);

					// We free the session
					GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

					Session = NULL;
					SetSession(Session);
				}
			}
		}		
		if(Session != NULL)
		{
			// Check the nolock mode. It can be activated only if the command comes from the same machine
			// than the server
			string sLockSession = InputMessage.GetHeader("NV_LOCK_SESSION");
			CleanString(sLockSession);
			MakeUpper(sLockSession);
			if((sLockSession == "NO") && (GetMain()->GetMachineName() == SourceMachineNameUtf8) && (GetMain()->GetMachineUser() == SourceMachineUserUtf8))
				LockSession = false;
		}
	}


	// Get the no error flag
	string NoError = InputMessage.GetHeader("NV_NO_ERROR");
	MakeUpper(NoError);
	CleanString(NoError);
	bool bNoError = false;
	if(NoError == "YES")
		bNoError = true;;

	if(bNoError)
		Command.Error.SetError();

	// Set the command error for the first part
	SetCommandError(1);
	
	// Check for the stop on error flag
	string sStopOnErrorHere = InputMessage.GetHeader("NV_STOP_ON_ERROR");
	CleanString(sStopOnErrorHere);
	MakeUpper(sStopOnErrorHere);
	bool bStopOnError = false;
	if(sStopOnErrorHere == "YES")
		bStopOnError = true;
	
	int i=1;

	// We write the session ID in the NvSessionId parameter of the first part
	if(Session != NULL)
		OutputMessage.AddHeader("Nv_Session_Id", Session->m_sIdent, 1);

	// We write also the server version (new from v 4.7.010)
	OutputMessage.AddHeader("Nv_Server_Version", GetMain()->m_sVersion, 1);

	// The content length for first part is always 0
	OutputMessage.AddHeader("Content-Length", "0", 1);
	
	bool StopIt = false;


	// Now we can process each command
	for(i=1; i<(int)(InputMessage.vDatas.size()); i++)
	{
		// Cleanup the actual command
		if(!bIsCommandAttachConsole)
			Command.Clear();	// Not if the command is to attach a console and with login
		Command.m_sMachineName = SourceMachineName;
		Command.m_sMachineUser = SourceMachineUser;
		Command.m_sClientAddress = IpAddress;
		Command.bClientUnicode = bClientUnicode;
		Command.m_sHost = InputMessage.GetHeader("X-FORWARDED-HOST");
		if(!Command.m_sHost.length())
			Command.m_sHost = InputMessage.GetHeader("HOST");

		// We first write a new part
		NvData *Data = NVNEW NvData;
		OutputMessage.vDatas.push_back(Data);
		provistr = "\r\n--";
		provistr += HTTP_SEND_BOUNDARY;
		OutputMessage.vDatas[i+1]->sInfo = provistr;

		// Add the session ID to the result
		if(Session != NULL)
			OutputMessage.AddHeader("Nv_Session_Id", Session->m_sIdent, i+1);
		else
			OutputMessage.AddHeader("Nv_Session_Id", "", i+1);

		string CommandHere = InputMessage.GetHeader("NV_CMD", i);
		string ClassHere = "";
		string ServiceHere = "";
		if(CommandHere.length())
		{
			string::size_type index = CommandHere.rfind(':');
			if(index != string::npos)
			{
				ClassHere = CommandHere.substr(0, index);
				CommandHere = CommandHere.substr(index+1);
				index = ClassHere.rfind(':');
				if(index != string::npos)
				{
					ServiceHere = ClassHere.substr(0, index);
					ClassHere = ClassHere.substr(index+1);
				}
			}
		}
		else
		{
			ServiceHere = InputMessage.GetHeader("NV_SERVICE", i);
			ClassHere = InputMessage.GetHeader("NV_CLASS", i);
			CommandHere = InputMessage.GetHeader("NV_COMMAND", i);
		}
		CleanString(ServiceHere, true);
		MakeUpper(ServiceHere);
		CleanString(ClassHere, true);
		MakeUpper(ClassHere);
		CleanString(CommandHere, true);
		MakeUpper(CommandHere);
		if((!ServiceHere.length()) && (!ClassHere.length()) && (!CommandHere.length()))
		{
			ServiceHere = "SYSTEM";
			ClassHere = "MISC";
			CommandHere = "NOP";
		}
		if(!ServiceHere.length())
			ServiceHere = "SYSTEM";
		if(!ClassHere.length())
			ClassHere = ServiceHere;

		OutputMessage.AddHeader("Nv_Service", ServiceHere, i+1);
		OutputMessage.AddHeader("Nv_Class", ClassHere, i+1);
		OutputMessage.AddHeader("Nv_Command", CommandHere, i+1);

		// Get the input container name from command
		string InContainerHere = InputMessage.GetHeader("NV_CONTAINER", i);
		string OutContainerHere = InContainerHere;

		// Get the input container name from command
		if(InputMessage.HeaderExist("NV_IN_CONTAINER", i))
			InContainerHere = InputMessage.GetHeader("NV_IN_CONTAINER", i);
		MakeLower(InContainerHere);
		CleanString(InContainerHere);

		// Get the output container name from command
		if(InputMessage.HeaderExist("NV_OUT_CONTAINER", i))
			OutContainerHere = InputMessage.GetHeader("NV_OUT_CONTAINER", i);
		MakeLower(OutContainerHere);
		CleanString(OutContainerHere);

		if(!InContainerHere.length())
			InContainerHere = NV_CONTAINER_DEFAULT_NAME;
		if(!OutContainerHere.length())
			OutContainerHere = NV_CONTAINER_DEFAULT_NAME;

		// Get the language from command
		string LanguageHere = InputMessage.GetHeader("NV_LANGUAGE", i);
		MakeUpper(LanguageHere);
		CleanString(LanguageHere);
		if(!LanguageHere.length())
		{
			if(Session != NULL)
				LanguageHere = Session->m_sLanguage;
			else
				LanguageHere = "ENGLISH";
		}

		// Get the close session from command
		string CloseSessionCommand = InputMessage.GetHeader("NV_CLOSE_SESSION", i);
		MakeUpper(CloseSessionCommand);
		CleanString(CloseSessionCommand);
		bool bCloseSessionFromCommand = false;
		if(CloseSessionCommand == "YES")
			bCloseSessionFromCommand = true;;

		// Add the content type to the result
		OutputMessage.AddHeader("Content-Type", "nirva/result", i+1);

		// Set command error to default value for this part
		Command.Error.m_sErrorService = ServiceHere;
		Command.Error.m_sErrorClass = ClassHere;
		SetCommandError(i+1);


		// Now we can prepare the command
		Command.m_sService = ServiceHere;
		Command.m_sClass = ClassHere;
		Command.m_sCommand = CommandHere;
		Command.m_sInContainer = InContainerHere;
		Command.m_sOutContainer = OutContainerHere;
//		if(!Command.m_sOutContainer.length())
//			Command.m_sOutContainer = Command.m_sInContainer;
		Command.m_sLanguage = LanguageHere;


		if(bGotLogin && i==1)
		{
			// If we just got login, we get parameters of first command from mParameters because we already got them
			// And they may be modified in the session_open procedure using the COMMAND:SET_PARAMETER or REMOVE_PARAMETER commands
			map<string,string>::iterator pos;
			for(pos = InputMessage.mParameters.begin(); pos != InputMessage.mParameters.end(); ++pos)
			{
				string ParameterNameHere = pos->first;
				CleanString(ParameterNameHere);
				string ParameterValueHere = pos->second;
	//			CleanString(ParameterValueHere);
				string Provistr = ParameterNameHere;
				MakeUpper(Provistr);
				bool ReservedKeyWord = false;
				if(Provistr == "NV_SERVICE")
					ReservedKeyWord = true;
				if(Provistr == "NV_CLASS")
					ReservedKeyWord = true;
				if(Provistr == "NV_COMMAND")
					ReservedKeyWord = true;
				if(Provistr == "NV_CMD")
					ReservedKeyWord = true;
				if(Provistr == "NV_IN_CONTAINER")
					ReservedKeyWord = true;
				if(Provistr == "NV_OUT_CONTAINER")
					ReservedKeyWord = true;
				if(Provistr == "CONTENT-TYPE")
					ReservedKeyWord = true;
				if(Provistr == "NV_LANGUAGE")
					ReservedKeyWord = true;
				if(Provistr == "NV_LOCK_SESSION")
					ReservedKeyWord = true;
				if(Provistr == "NV_CLOSE_SESSION")
					ReservedKeyWord = true;
	//			if(Provistr == "NV_NO_ERROR")
	//				ReservedKeyWord = true;

				if(ReservedKeyWord)
					continue;

				if(!ParameterNameHere.length())
					continue;

				Command.Parameters.SetValue(ParameterNameHere, ParameterValueHere);
			}
		}
		else
		{
			map<string,string>::iterator pos;
			for(pos = InputMessage.vDatas[i]->mHeaders.begin(); pos != InputMessage.vDatas[i]->mHeaders.end(); ++pos)
			{
				string ParameterNameHere = pos->first;
				CleanString(ParameterNameHere);
				string ParameterValueHere = pos->second;
	//			CleanString(ParameterValueHere);
				string Provistr = ParameterNameHere;
				MakeUpper(Provistr);
				bool ReservedKeyWord = false;
				if(Provistr == "NV_SERVICE")
					ReservedKeyWord = true;
				if(Provistr == "NV_CLASS")
					ReservedKeyWord = true;
				if(Provistr == "NV_COMMAND")
					ReservedKeyWord = true;
				if(Provistr == "NV_CMD")
					ReservedKeyWord = true;
				if(Provistr == "NV_IN_CONTAINER")
					ReservedKeyWord = true;
				if(Provistr == "NV_OUT_CONTAINER")
					ReservedKeyWord = true;
				if(Provistr == "CONTENT-TYPE")
					ReservedKeyWord = true;
				if(Provistr == "NV_LANGUAGE")
					ReservedKeyWord = true;
				if(Provistr == "NV_LOCK_SESSION")
					ReservedKeyWord = true;
				if(Provistr == "NV_CLOSE_SESSION")
					ReservedKeyWord = true;
	//			if(Provistr == "NV_NO_ERROR")
	//				ReservedKeyWord = true;

				if(ReservedKeyWord)
					continue;

				if(!ParameterNameHere.length())
					continue;

				Command.Parameters.SetValue(ParameterNameHere, ParameterValueHere);
			}
		}

		// now, we manage input data
		// In fact, the input data should exist only for SendObject command

		// The ownership of input data is transfered from http part to the command
		// Except if this is attachment with login
		if(!bIsCommandAttachConsole)
		{
			Command.DataSize = InputMessage.vDatas[i]->DataSize;
			Command.Data = InputMessage.vDatas[i]->Data;
			Command.DataFile = InputMessage.vDatas[i]->DataFile;
			Command.bDeleteDataFile = InputMessage.vDatas[i]->bDeleteDataFile;
		}
		InputMessage.vDatas[i]->DataSize = 0;
		InputMessage.vDatas[i]->Data = NULL;
		InputMessage.vDatas[i]->DataFile = "";
		InputMessage.vDatas[i]->bDeleteDataFile = false;

		bool SessionRetVal = true;

		// Now process the command
		if((Session != NULL) && !StopIt)
		{
			Command.bLockSession = LockSession;
			try
			{
				// Do not process if this is attachment with login because it's ever done
				if(!bIsCommandAttachConsole)
					SessionRetVal = Session->Command(&Command);
			}
			catch(...)
			{
				GetMain()->IncNumExceptions();
				string Message = Command.m_sService;
				Message += ":";
				Message += Command.m_sClass;
				Message += ":";
				Message += Command.m_sCommand;
				Command.Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_EXCEPTION, Message);
				GetMain()->Logs.Write("Exceptions", "Exception in command(Client)", Message.c_str(), NV_LOG_ERROR);
				SessionRetVal = false;
			}
			if(SessionRetVal && (Command.Parameters.KeyExist("NV_SPECIAL_OBJECT_TYPE")))
			{
				// This parameter is added by the GetObject function
				
				// Object name
				OutputMessage.AddHeader("NAME", Command.Parameters.GetValue("NAME"), i+1);
				// Object type
				OutputMessage.AddHeader("TYPE", Command.Parameters.GetValue("NV_SPECIAL_OBJECT_TYPE"), i+1);

			}
		}
		if(Command.Error.m_iErrorCode && (Session != NULL))
		{
			// In case of error check security to erase the info field
				if(!Session->CheckPermission(NULL, "SYSTEM", "DISPLAY_ERROR_INFO"))
					Command.Error.m_sErrorInfo = "";

		}
		if((Session != NULL) && (Command.m_sService == "SYSTEM") && (Command.m_sClass == "SESSION") && (Command.m_sCommand == "CLOSE"))
			bCloseSessionFromCommand = true;
		// Get the close session from command
		// if an internal request has been made
		CloseSessionCommand = Command.Parameters.GetValue("NV_CLOSE_SESSION");
		if(CloseSessionCommand == "YES")
			bCloseSessionFromCommand = true;;
		if(CloseSessionCommand == "NO")
			bCloseSessionFromCommand = false;;
		if((Session != NULL) && bCloseSessionFromCommand)
		{
			// The order is to close the session
			// This is forbiden with the nolock mode
			if(LockSession)
			{
				GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
				GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());

				// The session ID returned must be blank because the session is killed
				OutputMessage.AddHeader("Nv_Session_Id", "", 1);

				// The session is not more used so it should be set to NULL;
				Session = NULL;
				SetSession(Session);

				// Set this flag to tell that an explicit close has been requested
				ExplicitClose = true;
			}
			else
			{
				// We free the session
				GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

				Session = NULL;
				SetSession(Session);
			}
		}

		// now, we manage output data
		// In fact, the output data should exist only for GetObject command

		// The ownership of ouput data is transfered from command to http part
		OutputMessage.vDatas[i+1]->DataSize = Command.DataSize;
		OutputMessage.vDatas[i+1]->Data = Command.Data;
		OutputMessage.vDatas[i+1]->DataFile = Command.DataFile;
		OutputMessage.vDatas[i+1]->bDeleteDataFile = Command.bDeleteDataFile;
		OutputMessage.AddHeader("Content-Length", "0", i+1);
		unsigned NvInt64 RealDataSize = Command.DataSize;
		if(Command.DataFile.length())
			RealDataSize = GetFileLength(Command.DataFile.c_str());
		if(RealDataSize)
		{
			char ProviBuf[50];
			sprintf(ProviBuf, "%llu", RealDataSize);
			OutputMessage.AddHeader("Content-Length", ProviBuf, i+1);
		}
		Command.DataSize = 0;
		Command.Data = NULL;
		Command.DataFile = "";
		Command.bDeleteDataFile = false;

		// Set command error for this part
		if(!Command.Error.m_iErrorCode)
		{
			if((Session == NULL) && !(ExplicitClose))
			{
				if(LockSession)
					Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CLOSED, SessionIdHere);
				else
					Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_CLOSE, SessionIdHere);
			}
			if(StopIt)
				Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_STOPPED, SessionIdHere);
		}

		// Get the no error flag
		string NoError = InputMessage.GetHeader("NV_NO_ERROR", i);
		MakeUpper(NoError);
		CleanString(NoError);
		bool bNoError = false;
		if(NoError == "YES")
			bNoError = true;;

		if(bNoError)
			Command.Error.SetError();
		SetCommandError(i+1);


		// We must stop if requested to stop on error
		if(bStopOnError && (!SessionRetVal))
			StopIt = true;
	}
	
	// Cleanup the actual command
	Command.Clear();
	Command.m_sMachineName = SourceMachineName;
	Command.m_sMachineUser = SourceMachineUser;
	Command.m_sClientAddress = IpAddress;

	// Write the end of the message
	Data = NVNEW NvData;
	provistr = "\r\n--";
	provistr += HTTP_SEND_BOUNDARY;		// last boundary
	provistr += "--";
	Data->sInfo = provistr;
	OutputMessage.vDatas.push_back(Data);

	if(Session != NULL)
	{
		GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
		// Close the session if requested
		string CloseSession = InputMessage.GetHeader("SESSION");
		CleanString(CloseSession);
		MakeLower(CloseSession);
		if(CloseSession == "close")
		{
			GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());

			// The session ID returned must be blank because the session is killed
			OutputMessage.AddHeader("Nv_Session_Id", "", 1);
		}
		Session = NULL;
	}
	SetSession(Session);

	if(bDoConversion && (GetMain()->IsUnicode() != bClientUnicode))
	{
		// We must convert all output data
		int i;
		for(i=0; i<(int)(OutputMessage.vDatas.size()); i++)
		{
			NvData *Data = OutputMessage.vDatas[i];
			std::map<std::string, std::string> mNewHeaders;	
			map<string,string>::iterator pos;
			for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
			{
				string ParameterNameHere = pos->first;
				string ParameterValueHere = pos->second;
				EncodeUrlString(ParameterNameHere, bClientUnicode, false);
				EncodeUrlString(ParameterValueHere, bClientUnicode, false);
				mNewHeaders[ParameterNameHere] = ParameterValueHere;
			}
			Data->mHeaders.clear();
			Data->mHeaders = mNewHeaders;
		}
	}
}



void NvHttpClient::Test()
{
	// get current time and date
	time_t CurrentTime;
	time(&CurrentTime);
	struct tm tCurrentTime;
	char sTime[200];
	if(GetMain()->GetLocalTime())
	{
		localtime_r(&CurrentTime, &tCurrentTime);
//		strftime(sTime, sizeof(sTime), "%a, %d %b %Y %H:%M:%S", &tCurrentTime);
		MyGetTime(sTime, &tCurrentTime, false);
	}
	else
	{
		gmtime_r(&CurrentTime, &tCurrentTime);
//		strftime(sTime, sizeof(sTime), "%a, %d %b %Y %H:%M:%S GMT", &tCurrentTime);
		MyGetTime(sTime, &tCurrentTime, true);
	}

	
	// and write a simple HTML page
	char buffer[200];
	strcpy(buffer, "<html>\r\n");
	strcat(buffer, "<body>\r\n");
	strcat(buffer, "<h1>NIRVA Server");
	strcat(buffer, " : ");
	strcat(buffer, sTime);
	strcat(buffer, "</h1>\r\n");
	strcat(buffer, "</body>\r\n");
	strcat(buffer, "</html>\r\n");

	OutputMessage.vDatas[0]->Data = (char *)malloc(strlen(buffer)+1);
	OutputMessage.vDatas[0]->DataSize = (unsigned long)(strlen(buffer)+1);
	strcpy(OutputMessage.vDatas[0]->Data, buffer);
}


void NvHttpClient::Config()
{
	SetHttpError(HTTP_ERROR_NOTIMPLEMENTED, "Function not yet implemented");
}


void NvHttpClient::Doc()
{
	SetHttpError(HTTP_ERROR_NOTIMPLEMENTED, "Function not yet implemented");
}


void NvHttpClient::Log()
{
	SetHttpError(HTTP_ERROR_NOTIMPLEMENTED, "Function not yet implemented");
}


void NvHttpClient::BrowserFunction(int iInputType)
{
	RemoveSessionOnCleanup = false;

	// Cleanup the actual command
	Command.Clear();
	Command.m_sClientAddress = IpAddress;
	Command.m_sHost = InputMessage.GetHeader("X-FORWARDED-HOST");
	if(!Command.m_sHost.length())
		Command.m_sHost = InputMessage.GetHeader("HOST");
	
	// By default we have no session id
	OutputMessage.AddHeader("Nv_Session_Id", "", 1);

	if(iInputType == NV_HTTP_WSDL)
	{
		InputMessage.mParameters.clear();
	}

	if(iInputType == NV_HTTP_WEBS)
		Command.bWebs = true;

	// Test if the command comes from a client
	// This is possible here with the REQUEST XML_COMMAND client command
	bool bFromClient = false;
	string sNvSourceHere = InputMessage.GetHeader("NV_SOURCE");
	if(sNvSourceHere == HTTP_CLIENT_STRING)
		bFromClient = true;

	string HttpFileInput = "";
	if(!bFromClient && (iInputType == NV_HTTP) && (InputMessage.iMethod == HTTP_METHOD_POST) && (!InputMessage.bMultipart))
	{
		string sInputType = InputMessage.GetParameter("NV_INPUT_TYPE");
		CleanString(sInputType);
		MakeUpper(sInputType);
		if(sInputType.find("FILE") == 0)
		{
			if(sInputType.find("FILE:") == 0)
				HttpFileInput = sInputType.substr(5);
			if(!HttpFileInput.length())
				HttpFileInput = "INPUT_FILE";
		}
	}

	string ContentType = "";
	if(InputMessage.vDatas[0]->mHeaders.find("CONTENT-TYPE") != InputMessage.vDatas[0]->mHeaders.end())
	{
		ContentType = InputMessage.vDatas[0]->mHeaders["CONTENT-TYPE"];
		if((iInputType == NV_HTTP) && !HttpFileInput.length())
		{
			if(ContentType.find("soap") != string::npos)
				iInputType = NV_HTTP_SOAP;
			else if(ContentType.find("xml") != string::npos)
				iInputType = NV_HTTP_XML;
		}
	}
	if(InputMessage.bMultipart)
		iInputType = NV_HTTP;

	int iXmlCoding = XML_MODEL_NORMAL;
	if(iInputType == NV_HTTP_SOAP)
		iXmlCoding = XML_MODEL_SIMPLE;			// Default for XML, SOAP and WEBS
	if(iInputType == NV_HTTP_XML)
		iXmlCoding = XML_MODEL_SIMPLE;			// Default for XML, SOAP and WEBS
	string NvXmlModel = InputMessage.GetParameter("NV_XML_MODEL");
	CleanString(NvXmlModel, true);
	MakeUpper(NvXmlModel);
	string sNvSimpleCoding = InputMessage.GetParameter("NV_XML_SIMPLE");
	CleanString(sNvSimpleCoding, true);
	MakeUpper(sNvSimpleCoding);
	if(NvXmlModel.length())
	{
		if(NvXmlModel == "SIMPLE")
			iXmlCoding = XML_MODEL_SIMPLE;
		if(NvXmlModel == "NORMAL")
			iXmlCoding = XML_MODEL_NORMAL;
		if(NvXmlModel == "COMPACT")
			iXmlCoding = XML_MODEL_COMPACT;
		if(NvXmlModel == "TINY")
			iXmlCoding = XML_MODEL_TINY;
	}
	else
	{
		// Check NV_XML_SIMPLE for ascending compatibility
		if(sNvSimpleCoding == "YES")
			iXmlCoding = XML_MODEL_SIMPLE;
		if(sNvSimpleCoding == "NO")
			iXmlCoding = XML_MODEL_NORMAL;
	}
	if(iInputType == NV_HTTP_WEBS)
		iXmlCoding = XML_MODEL_SIMPLE;			// Always use the simple coding for web services

	// if we have SOAP input, we must extract the body part that
	// becomes the real XML data input
	NvObjBinary SoapCommandHeader;
	if((iInputType == NV_HTTP_SOAP) || (iInputType == NV_HTTP_WEBS))
	{
		if(!GetSoapCommandHeader(&SoapCommandHeader))
		{
			Command.Error.SetError("SYSTEM", "SOAP", NV_SOAP_ERROR_PARSING_MESSAGE);
			SetHttpErrorBrowserCommand(iInputType, iXmlCoding);
			return;
		}
		// Get SoapBody should extract the Soap message body
		if(!GetSoapBody())
		{
			Command.Error.SetError("SYSTEM", "SOAP", NV_SOAP_ERROR_PARSING_MESSAGE);
			SetHttpErrorBrowserCommand(iInputType, iXmlCoding);
			return;
		}
	}

	// If we have input XML or SOAP, we get the eventual command parameters from it
	// They override the command parameters found on the URL
	bool bParseXmlSuccess = true;
	if(iInputType != NV_HTTP)
	{
		// For XML input, we must populate the input container with the XML data
		// The ownership of input data is transfered from http part to the command
		Command.DataSize = InputMessage.vDatas[0]->DataSize;
		Command.Data = InputMessage.vDatas[0]->Data;
		Command.DataFile = InputMessage.vDatas[0]->DataFile;
		Command.bDeleteDataFile = InputMessage.vDatas[0]->bDeleteDataFile;
		InputMessage.vDatas[0]->DataSize = 0;
		InputMessage.vDatas[0]->Data = NULL;
		InputMessage.vDatas[0]->DataFile = "";
		InputMessage.vDatas[0]->bDeleteDataFile = false;


		Command.iXmlInputType = XML_IN_XML;
		if(iInputType == NV_HTTP_SOAP)
			Command.iXmlInputType = XML_IN_SOAP;
		if(iInputType == NV_HTTP_WEBS)
			Command.iXmlInputType = XML_IN_SOAP;
		Command.iXmlCoding = iXmlCoding;

		// We must first do the eventual XSLT transformation of the input
		// For that, we need at least the application name and the
		// NV_XML_XSL_IN parameter in the URL
		string XslIn = InputMessage.GetParameter("NV_XML_XSL_IN");
		CleanString(XslIn);
		if(iInputType == NV_HTTP_SOAP)
			XslIn = "";		// This parameter has no meaning in the SOAP connector
		if(iInputType == NV_HTTP_WEBS)
			XslIn = "";		// This parameter has no meaning in the WEBS connector
		if(XslIn.length())
		{
			// We must get the application name from URL parameter or URL path
			string AppliNameHere = "";
			string::size_type index = XslIn.find(":");
			if(index != string::npos)
			{
				AppliNameHere = XslIn.substr(0, index);
				CleanString(AppliNameHere);
				XslIn = XslIn.substr(index+1);
				CleanString(XslIn);
			}

			if(!AppliNameHere.length())
			{
				if(InputMessage.ParameterExist("NV_APPLICATION"))
					AppliNameHere = InputMessage.GetParameter("NV_APPLICATION");
				else
				{
					// If the NV_APPLICATION parameter is not given, we try to get it from
					// the URL
					AppliNameHere = InputMessage.sPath;
					MakeUpper(AppliNameHere);
					string::size_type index = AppliNameHere.find("NV_APP_");
					if(index != string::npos)
					{
						AppliNameHere = AppliNameHere.substr(index+7);
						if(((index = AppliNameHere.find('/')) != string::npos) || ((index = AppliNameHere.find('\\')) != string::npos))
							AppliNameHere = AppliNameHere.substr(0, index);
					}
					else
						AppliNameHere = "";
				}
				CleanString(AppliNameHere);
			}
			MakeUpper(AppliNameHere);
			if(!AppliNameHere.length())
				AppliNameHere = "NVDEF";

			// Ok we have now the application name

			NvSession TempSession;
			bParseXmlSuccess = TempSession.XsltInputTransform(&Command, XslIn, AppliNameHere);
		}
		if(bParseXmlSuccess)
			bParseXmlSuccess = ParseXmlCommand();
		// The ParseXmlCommand may automatically select the XML model
		iXmlCoding = Command.iXmlCoding;
	}

	if((bParseXmlSuccess) && (SoapCommandHeader.GetSize()))
	{
		NvXml XmlParser;
		NvObjIndStringList Parameters;
		bParseXmlSuccess = XmlParser.GetSoapHeaderCommandParam(&Parameters, SoapCommandHeader.m_cdata, SoapCommandHeader.m_isize);
		map<string,string>::iterator pos;
		for(pos = Parameters.m_mList.begin(); pos != Parameters.m_mList.end(); ++pos)
			InputMessage.mParameters[pos->first] = pos->second;
	}

	if((iInputType != NV_HTTP_SOAP) && (iInputType != NV_HTTP_WEBS))
	{
		if(InputMessage.ParameterExist("NV_XML_NSREF"))
		{
			string NsPrefix = InputMessage.GetParameter("NV_XML_NSPREFIX");
			CleanString(NsPrefix, true);
			string NsRef = InputMessage.GetParameter("NV_XML_NSREF");
			CleanString(NsRef, true);
			if(NsRef.length())
			{
				Command.m_sXmlNameSpace = NsPrefix;
				if(NsPrefix.length())
					Command.m_sXmlNameSpace += "=";
				Command.m_sXmlNameSpace += NsRef;
			}
			else
				Command.m_sXmlNameSpace = "";
		}
	}

	// We search that again because the parameter main come from the XML data itself
	NvXmlModel = InputMessage.GetParameter("NV_XML_MODEL");
	CleanString(NvXmlModel, true);
	MakeUpper(NvXmlModel);
	sNvSimpleCoding = InputMessage.GetParameter("NV_XML_SIMPLE");
	CleanString(sNvSimpleCoding, true);
	MakeUpper(sNvSimpleCoding);
	if(NvXmlModel.length())
	{
		if(NvXmlModel == "SIMPLE")
			iXmlCoding = XML_MODEL_SIMPLE;
		if(NvXmlModel == "NORMAL")
			iXmlCoding = XML_MODEL_NORMAL;
		if(NvXmlModel == "COMPACT")
			iXmlCoding = XML_MODEL_COMPACT;
		if(NvXmlModel == "TINY")
			iXmlCoding = XML_MODEL_TINY;
	
		if(iInputType == NV_HTTP_WEBS)
			iXmlCoding = XML_MODEL_SIMPLE;			// Always use the simple coding for web services
		Command.iXmlCoding = iXmlCoding;
	}
	else
	{
		if(sNvSimpleCoding.length())
		{
			if(sNvSimpleCoding == "YES")
				iXmlCoding = XML_MODEL_SIMPLE;
			if(sNvSimpleCoding == "NO")
				iXmlCoding = XML_MODEL_NORMAL;
		
			if(iInputType == NV_HTTP_WEBS)
				iXmlCoding = XML_MODEL_SIMPLE;			// Always use the simple coding for web services

			Command.iXmlCoding = iXmlCoding;
		}
	}

	// Get the language one more time in cased the NV_LANGUAGE parameter was transmitted
	// in a form or in the SOAP or XML input
	string LanguageHere = InputMessage.GetParameter("NV_LANGUAGE");
	MakeUpper(LanguageHere);
	CleanString(LanguageHere);
	if(!LanguageHere.length())
		LanguageHere = "ENGLISH";
	Command.m_sLanguage = LanguageHere;

	string OutputType = InputMessage.GetParameter("NV_XML_OUTPUT");
	CleanString(OutputType);
	MakeUpper(OutputType);
	int iOutputType = NV_HTTP_HTML;	// Default is html output
	if(iInputType == NV_HTTP_XML)
		iOutputType = NV_HTTP_XML;		// Except if the input was xml
	if(OutputType == "HTML")
		iOutputType = NV_HTTP_HTML;
	if(OutputType == "XML")
		iOutputType = NV_HTTP_XML;
	if(iInputType == NV_HTTP_SOAP)		// If input is SOAP, output also
		iOutputType = NV_HTTP_SOAP;
	if(iInputType == NV_HTTP_WEBS)		// If input is SOAP WEBS, output is SOAP
		iOutputType = NV_HTTP_WEBS;

	if(!bParseXmlSuccess)
	{
		if(Command.Error.m_iErrorCode == 0)
			Command.Error.SetError("SYSTEM", "XML", NV_XML_ERROR_PARSING_COMMAND);
		SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
		return;
	}

	string Charset = ContentType;
	MakeLower(Charset);
	string::size_type index;
	if((index = Charset.find("charset")) == string::npos)
		Charset = "";
	else
		Charset = Charset.substr(index+7);
	if((index = Charset.find("=")) != string::npos)
		Charset = Charset.substr(index+1);
	if((index = Charset.find(";")) != string::npos)
		Charset = Charset.substr(0, index);
	CleanString(Charset, true);
	// remove double quotes
	while((index = Charset.find('\"')) != string::npos)
		Charset.erase(index, 1);
	// remove quotes
	while((index = Charset.find('\'')) != string::npos)
		Charset.erase(index, 1);
	MakeUpper(Charset);
	Command.m_sInputCharset = Charset;

	if(iInputType == NV_HTTP && InputMessage.iMethod == HTTP_METHOD_POST && !HttpFileInput.length())
		GetInputFormValues(ContentType, Charset);
/*	{
		// Try to get the encoding of the input form
		bool bFormUnicode = true;
		bool bAutoDetectEncoding = GetMain()->GetAutoDetectEncoding();
		if(Charset == "ISO-8859-1")
		{
			bFormUnicode = false;
			bAutoDetectEncoding = false;
		}
		if(Charset == "UTF-8")
		{
			bFormUnicode = true;
			bAutoDetectEncoding = false;
		}
		string EncodingFromParam = InputMessage.GetParameter("NV_FORM_ENCODING");
		CleanString(EncodingFromParam);
		MakeUpper(EncodingFromParam);
		if(EncodingFromParam == "UTF-8")
		{
			bFormUnicode = true;
			bAutoDetectEncoding = false;
		}
		if(EncodingFromParam == "ISO-8859-1")
		{
			bFormUnicode = false;
			bAutoDetectEncoding = false;
		}
		if(EncodingFromParam == "AUTO")
		{
			bAutoDetectEncoding = true;
		}

		// We must get information sent from the browser
		// Especially the parameters comming from a HTML form
		string::size_type index;
		if(InputMessage.bMultipart)
		{
			int i;
			for(i=1; i<(int)(InputMessage.vDatas.size()); i++)
			{
				string ContentDisposition = InputMessage.GetHeader("CONTENT-DISPOSITION", i);
				if(!ContentDisposition.length())
					continue;
				if(ContentDisposition.find("form-data") == string::npos)
					continue;
				string ParameterName = "";
				string ParameterValue = "";
				while((index = ContentDisposition.find(';')) != string::npos)
				{
					ContentDisposition.erase(0, index+1);
					CleanString(ContentDisposition);
					index = ContentDisposition.find("name=");
					if(index != 0)
						continue;
					ContentDisposition.erase(0, index+5);
					index = ContentDisposition.find(';');
					if(index != string::npos)
						ContentDisposition = ContentDisposition.substr(0, index);
					CleanString(ContentDisposition);
					if(ContentDisposition.length() && (ContentDisposition[0] == '\"'))
						ContentDisposition.erase(0, 1);
					if(ContentDisposition.length() && (ContentDisposition[ContentDisposition.length()-1] == '\"'))
						ContentDisposition.erase(ContentDisposition.size()-1, 1);
					CleanString(ContentDisposition);
					if(ContentDisposition.length())
						ParameterName = ContentDisposition;
					break;
				}
				if(ParameterName.length())
				{
					ContentDisposition = InputMessage.GetHeader("CONTENT-DISPOSITION", i);
					if(ContentDisposition.find("filename=") != string::npos)
						continue;	// This a file content, we will manage it later
					unsigned long bodysize = InputMessage.vDatas[i]->DataSize;
					if(InputMessage.vDatas[i]->DataFile.length())
						bodysize = (unsigned long)(GetFileLength(InputMessage.vDatas[i]->DataFile.c_str()));
					if(bodysize)
					{
						char *PostData = (char *)malloc(bodysize+1);
						if(InputMessage.vDatas[i]->DataFile.length())
						{
							// We must read the file
							FILE *File;
							File = Nvfopen(InputMessage.vDatas[i]->DataFile.c_str(), "rb");
							if(File != NULL)
							{
								fread(PostData, 1, bodysize, File);
								fclose(File);
							}
						}
						else
							memcpy(PostData, InputMessage.vDatas[i]->Data, bodysize);
						PostData[bodysize] = '\0';
						ParameterValue = PostData;
						free(PostData);
//						CleanString(ParameterValue);
						// We don't need the input data any more
						InputMessage.vDatas[i]->FreeData();
					}
					DecodeUrlString(ParameterName, bFormUnicode, false, bAutoDetectEncoding);
					DecodeUrlString(ParameterValue, bFormUnicode, false, bAutoDetectEncoding);
					MakeUpper(ParameterName);	// the parameter name is case insensitive
					InputMessage.mParameters[ParameterName] = ParameterValue;
				}
			}
		}
		else
		{
			unsigned long bodysize = InputMessage.vDatas[0]->DataSize;
			if(InputMessage.vDatas[0]->DataFile.length())
				bodysize = (unsigned long)(GetFileLength(InputMessage.vDatas[0]->DataFile.c_str()));
			if(bodysize)
			{
				bool bUrlEncoded = false;
				if(ContentType.find("application/x-www-form-urlencoded") != string::npos)
					bUrlEncoded = true;
				if(bUrlEncoded)
				{
					char *PostData = (char *)malloc(bodysize+1);
					if(InputMessage.vDatas[0]->DataFile.length())
					{
						// We must read the file
						FILE *File;
						File = Nvfopen(InputMessage.vDatas[0]->DataFile.c_str(), "rb");
						if(File != NULL)
						{
							fread(PostData, 1, bodysize, File);
							fclose(File);
						}
					}
					else
						memcpy(PostData, InputMessage.vDatas[0]->Data, bodysize);
					PostData[bodysize] = '\0';
					string PostParam = PostData;
					free(PostData);
					if(PostParam.length() && (PostParam[0] != '&'))
					{
						string Provistr = "&";
						Provistr += PostParam;
						PostParam = Provistr;
					}
					while((index = PostParam.rfind('&')) != string::npos)
					{
						string sParameter = PostParam.substr(index+1);
						PostParam.erase(index);
						if((index = sParameter.find('=')) != string::npos)
						{
							string sValue = sParameter.substr(index+1);
							sParameter.erase(index);
							if(sParameter.size())
							{
								DecodeUrlString(sParameter, bFormUnicode, true, bAutoDetectEncoding);
								DecodeUrlString(sValue, bFormUnicode, true, bAutoDetectEncoding);
								MakeUpper(sParameter);	// the parameter name is case insensitive
								InputMessage.mParameters[sParameter] = sValue;
							}
						}
					}
				}
				// We don't need the input data any more
				InputMessage.vDatas[0]->FreeData();
			}
		}
	}
*/
	// Get the language from command (useful here for error messages)
	LanguageHere = InputMessage.GetParameter("NV_LANGUAGE");
	MakeUpper(LanguageHere);
	CleanString(LanguageHere);
	if(!LanguageHere.length())
		LanguageHere = "ENGLISH";
	Command.m_sLanguage = LanguageHere;

	string SoapAction = "";
	string WebsApplication = "";
	string WebsService = "";
	string WebsOperation = "";
	if((iInputType == NV_HTTP_WEBS))
	{
		// The web service operation comes from SOAPACTION header
		WebsOperation = InputMessage.GetHeader("SOAPACTION");
		string::size_type index;
		// remove double quotes
		while((index = WebsOperation.find('\"')) != string::npos)
			WebsOperation.erase(index, 1);
		// eliminate single quotes
		while((index = WebsOperation.find('\'')) != string::npos)
			WebsOperation.erase(index, 1);

		// The application and service name comes from URL
		WebsService = InputMessage.sPath;
		CleanString(WebsService, true);

		// replace backslashes with slashes
		while((index = WebsService.find('\\')) != string::npos)
			WebsService.replace(index, 1, 1, '/');
		// eliminate first slash
		if(WebsService.size() && WebsService[0] == '/')
			WebsService.erase(0, 1);
		// eliminate last slash
		if(WebsService.size() && WebsService[WebsService.size()-1] == '/')
			WebsService.erase(WebsService.size()-1, 1);
	
		index = WebsService.rfind('/');
		if(index != string::npos)
		{
			WebsApplication = WebsService.substr(0, index);
			WebsService = WebsService.substr(index+1);
			index = WebsApplication.rfind('/');
			if(index != string::npos)
				WebsApplication = WebsApplication.substr(index+1);
		}
		MakeUpper(WebsApplication);
		index = WebsApplication.find("NV_APP_");
		if(index == 0)
			WebsApplication = WebsApplication.substr(7);

		string WebServiceName = InputMessage.GetParameter("WEBSERVICE");
		CleanString(WebServiceName);
		MakeLower(WebServiceName);
		if(!WebServiceName.length())
			WebServiceName = WebsService;
		if(!WebServiceName.length())
			WebServiceName = "nvdef";
		InputMessage.mParameters["WEBSERVICE"] = WebServiceName;
		bool bHttpsOnly = false;
		NvWebs* WebService = GetMain()->Webservices.GetWebservice(WebServiceName.c_str());
		if(WebService != NULL)
		{
			bHttpsOnly = WebService->GetHttpsOnly();
			GetMain()->Webservices.FreeWebservice(WebServiceName.c_str());
		}
		if(bHttpsOnly && !(sSocket.bSslConnection))
		{
			Command.Error.SetError("SYSTEM", "WEBSERVICE", NV_WEBSERVICE_NOT_AVAILABLE, WebServiceName);
			SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
			return;
		}
	}

	// Get the session ID from the parameters 
	string SessionIdHere = InputMessage.GetParameter("NV_SESSION_ID");
	CleanString(SessionIdHere);
	if(!SessionIdHere.length())
	{
		// Check if session id is in a cookie
		string Cookie = InputMessage.GetHeader("COOKIE");
		string::size_type index = Cookie.find("NvSessionId=");
		if(index != string::npos)
			SessionIdHere = Cookie.substr(index + 12);
		index = SessionIdHere.find(';');
		if(index != string::npos)
			SessionIdHere = SessionIdHere.substr(0, index);
		CleanString(SessionIdHere);
	}
	if(SessionIdHere.length() && (SessionIdHere[0] == '.'))
		SessionIdHere = "";
	string SessionName = InputMessage.GetParameter("NV_SESSION_NAME");
	CleanString(SessionName);
	MakeLower(SessionName);

	// Get the no error flag
	string NoError = InputMessage.GetParameter("NV_NO_ERROR");
	MakeUpper(NoError);
	CleanString(NoError);
	bool bNoError = false;
	if(NoError == "YES")
		bNoError = true;
	
	string CommandHere = InputMessage.GetParameter("NV_CMD");
	string ClassHere = "";
	string ServiceHere = "";
	if(CommandHere.length())
	{
		string::size_type index = CommandHere.rfind(':');
		if(index != string::npos)
		{
			ClassHere = CommandHere.substr(0, index);
			CommandHere = CommandHere.substr(index+1);
			index = ClassHere.rfind(':');
			if(index != string::npos)
			{
				ServiceHere = ClassHere.substr(0, index);
				ClassHere = ClassHere.substr(index+1);
			}
		}
	}
	else
	{
		ServiceHere = InputMessage.GetParameter("NV_SERVICE");
		ClassHere = InputMessage.GetParameter("NV_CLASS");
		CommandHere = InputMessage.GetParameter("NV_COMMAND");
	}
	CleanString(ServiceHere, true);
	MakeUpper(ServiceHere);
	CleanString(ClassHere, true);
	MakeUpper(ClassHere);
	CleanString(CommandHere, true);
	MakeUpper(CommandHere);
	if((!ServiceHere.length()) && (!ClassHere.length()) && (!CommandHere.length()))
	{
		ServiceHere = "SYSTEM";
		ClassHere = "MISC";
		CommandHere = "NOP";
	}
	if(!ServiceHere.length())
		ServiceHere = "SYSTEM";
	if(!ClassHere.length())
	{
		if(iInputType == NV_HTTP)
		{
			if(ServiceHere == "SYSTEM")
				ClassHere = "MISC";
			else	
				ClassHere = ServiceHere;
		}
		else
			ClassHere = "MISC";
	}
	if(!CommandHere.length())
		CommandHere = "NOP";

	// Set http headers for the command
	map<string,string>::iterator pos;
	for(pos = InputMessage.vDatas[0]->mHeaders.begin(); pos != InputMessage.vDatas[0]->mHeaders.end(); ++pos)
	{
		string HeaderNameHere = pos->first;
		CleanString(HeaderNameHere);
		string  HeaderValueHere = pos->second;
//		CleanString(HeaderValueHere);

		if(!HeaderNameHere.length())
			continue;

		Command.HttpInHeaders.SetValue(HeaderNameHere, HeaderValueHere);
	}

	// And try to get the session
	NvSession *Session = GetMain()->Sessions.GetSession(SessionIdHere.c_str());
	if((Session == NULL) && (!SessionIdHere.length()) && (SessionName.length()))
	{
		// From browser, we can only access named sessions starting with test
		string::size_type index = SessionName.find("test");
		if(index == 0)
			Session = GetMain()->Sessions.GetSessionByName(SessionName.c_str());
	}
	SetSession(Session);


	// We must manage the case where the session is not null but the browser send an authorisation request for posting POST data
	if(Session != NULL && iInputType == NV_HTTP && InputMessage.iMethod == HTTP_METHOD_POST && InputMessage.GetHeader("AUTHORIZATION").length())
	{
		string SSOUser ="";
		string SSODomain ="";
		string AppliNameForSSO = "";
		if(Session->Application != NULL)
			AppliNameForSSO = Session->Application->GetName();
		string SSOErrInfo ="";
		int SSORetVal = SSOAuthenticate(AppliNameForSSO, Session->GetUserName(), SSOUser, SSODomain, false, SSOErrInfo);	
		if(SSORetVal <= 0)
		{
			// SSO failed
			// We prepare error information

			string Info = "user=";
			
			if(Session->GetUserName().length())
				Info += Session->GetUserName();
			else
				Info += "(Default user)";
			Info += " - application=";
			if(AppliNameForSSO.length())
				Info += AppliNameForSSO;
			else
				Info += "(Default application)";

			if(SSORetVal == 0 && SSOErrInfo.length())
			{
				Info += " - ";
				Info += SSOErrInfo;
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_AUTHENTICATION_ERROR, Info);
			}
			if(SSORetVal == -4)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_APPLICATION, Info);

			if(SSORetVal != -200)
				SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);

			// We free the session
			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

			if(SSORetVal == -200)
				SetHttpError(HTTP_ERROR_UNAUTHORIZED, "", false);

			return;
		}
	}

	bool SessionCreated = false;
	if(Session == NULL)
	{
		// First test if requested to create a new session if the actual one is expired
		string NewIfExpired = InputMessage.GetParameter("NV_NEW_IF_EXPIRED");
		CleanString(NewIfExpired);
		MakeUpper(NewIfExpired);
		if(NewIfExpired == "YES")
		{
			InputMessage.mParameters.erase("NV_SESSION_ID");
			SessionIdHere = "";
		}
		if(SessionIdHere.length())
		{
			// The session has probably expired
			if((ServiceHere == "SYSTEM") && (ClassHere == "SESSION") && (CommandHere == "CLOSE") && (iInputType == NV_HTTP))
			{
				if(InputMessage.ParameterExist("LOGOUT_PAGE"))
				{
					string LogoutPage = InputMessage.GetParameter("LOGOUT_PAGE");
					GetLogoutPage(LogoutPage);
					SetCookieSession(NULL);
					return;
				}
			}
			Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_FIND);
			SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
			SetCookieSession(NULL);
			return;
		}

		// The session doesn't exist, we check for a redirection
		// Get application name
		string AppliNameHere = WebsApplication;
		if(InputMessage.ParameterExist("NV_APPLICATION"))
			AppliNameHere = InputMessage.GetParameter("NV_APPLICATION");
		if(!AppliNameHere.length())
			AppliNameHere = WebsApplication;
		else
		{
			// If the NV_APPLICATION paramater is not given, we try to get it from
			// the URL
			AppliNameHere = InputMessage.sPath;
			MakeUpper(AppliNameHere);
			string::size_type index = AppliNameHere.find("NV_APP_");
			if(index != string::npos)
			{
				AppliNameHere = AppliNameHere.substr(index+7);
				if(((index = AppliNameHere.find('/')) != string::npos) || ((index = AppliNameHere.find('\\')) != string::npos))
					AppliNameHere = AppliNameHere.substr(0, index);
			}
			else
				AppliNameHere = "";
		}
		CleanString(AppliNameHere);
		MakeUpper(AppliNameHere);
		if(!AppliNameHere.length())
			AppliNameHere = "NVDEF";

		if(DoRedirect(AppliNameHere.c_str()))
			return;

		// The requested session is blank (or not given)
		// So we try to create it
		if(GetMain()->CheckOverTime())
		{
			Command.Error.SetError("SYSTEM", "SYSTEM", NV_SYSTEM_OVER_TIME);
			SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
		}
		else
		{
			Session = GetMain()->Sessions.GetSession(NULL);
			if((Session != NULL) && (SessionName.length()))
			{
				// From browser, we accept only named sessions starting with test
				string::size_type index = SessionName.find("test");
				if(index == 0)
					Session->m_sName = SessionName;
			}
			SetSession(Session);
			if(Session == NULL)
			{
				int MaxSessions = GetMain()->GetMaxSessions();
				if(MaxSessions != -1)
				{
					if(GetMain()->Sessions.GetNumSessions() >= MaxSessions)
					{
						string AppliNameHere = WebsApplication;
						if(InputMessage.ParameterExist("NV_APPLICATION"))
							AppliNameHere = InputMessage.GetParameter("NV_APPLICATION");
						if(!AppliNameHere.length())
							AppliNameHere = WebsApplication;
						else
						{
							// If the NV_APPLICATION paramater is not given, we try to get it from
							// the URL
							AppliNameHere = InputMessage.sPath;
							MakeUpper(AppliNameHere);
							string::size_type index = AppliNameHere.find("NV_APP_");
							if(index != string::npos)
							{
								AppliNameHere = AppliNameHere.substr(index+7);
								if(((index = AppliNameHere.find('/')) != string::npos) || ((index = AppliNameHere.find('\\')) != string::npos))
									AppliNameHere = AppliNameHere.substr(0, index);
							}
							else
								AppliNameHere = "";
						}
						CleanString(AppliNameHere);
						MakeUpper(AppliNameHere);
						if(!AppliNameHere.length())
							AppliNameHere = "NVDEF";
						string UserHere = InputMessage.GetParameter("NV_USER");
						CleanString(UserHere);
						if(!UserHere.length())
							UserHere = "nvdef";
						MakeLower(UserHere);
						if((AppliNameHere == "NVDEF") && (UserHere == "nvadmin"))
						{
							Session = GetMain()->Sessions.GetSession(NULL, false);
							SetSession(Session);
							if(Session != NULL)
							{
								// This session will only be authorized to list and remove sessions
								Session->m_bOverMaxSession = true;
	//							SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
							}
						}
						else
						{
							Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_TOO_MANY);
							SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
						}
					}
					else
					{
						Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTCREATE);
						SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
					}
				}
				else
				{
					Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTCREATE);
					SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
				}
			}
		}
		if(Session == NULL)
		{
			SetCookieSession(NULL);
			return;
		}

		SessionCreated = true;

		// The session creation was OK
		// Now, we must login to it
		 AppliNameHere = WebsApplication;
		if(InputMessage.ParameterExist("NV_APPLICATION"))
			AppliNameHere = InputMessage.GetParameter("NV_APPLICATION");
		else
		{
			// If the NV_APPLICATION parameter is not given, we try to get it from
			// the URL
			AppliNameHere = InputMessage.sPath;
			MakeUpper(AppliNameHere);
			string::size_type index = AppliNameHere.find("NV_APP_");
			if(index != string::npos)
			{
				AppliNameHere = AppliNameHere.substr(index+7);
				if(((index = AppliNameHere.find('/')) != string::npos) || ((index = AppliNameHere.find('\\')) != string::npos))
					AppliNameHere = AppliNameHere.substr(0, index);
			}
			else
				AppliNameHere = "";
		}
		CleanString(AppliNameHere);
		if(!AppliNameHere.length())
			AppliNameHere = WebsApplication;
		string UserHere = InputMessage.GetParameter("NV_USER");
		CleanString(UserHere);
		string PasswordHere = InputMessage.GetParameter("NV_PASSWORD");
		CleanString(PasswordHere);
		bool bWithNewPassword = false;
		if(InputMessage.ParameterExist("NV_NEW_PASSWORD"))
			bWithNewPassword = true;
		string NewPasswordHere = InputMessage.GetParameter("NV_NEW_PASSWORD");
		CleanString(NewPasswordHere);
		string NewPasswordHereConfirm = InputMessage.GetParameter("NV_NEW_PASSWORD_CONFIRM");
		CleanString(NewPasswordHereConfirm);
		bool bWithNewPasswordConfirm = false;
		if(bWithNewPassword)
		{
			if(InputMessage.ParameterExist("NV_NEW_PASSWORD_CONFIRM"))
				bWithNewPasswordConfirm = true;
		}
		string Hash = AppliNameHere + UserHere + PasswordHere;
		Hash = MD5_string(Hash);
		MakeUpper(Hash);
		string sSessionTimeOut = InputMessage.GetParameter("NV_TIMEOUT");
		CleanString(sSessionTimeOut);
		int iSessionTimeOut = atoi(sSessionTimeOut.c_str());
		if(iSessionTimeOut < 0)
			iSessionTimeOut = 0;
		string OpenProc = InputMessage.GetParameter("NV_SESSION_OPEN");
		CleanString(OpenProc);
		if(!OpenProc.length())
			OpenProc = "session_open";
		if(OpenProc == "NV_SESSION_OPEN_NONE")
			OpenProc = "";
		Session->m_sOpenProc = OpenProc;
		string CloseProc = InputMessage.GetParameter("NV_SESSION_CLOSE");
		CleanString(CloseProc);
		if(!CloseProc.length())
			CloseProc = "session_close";
		if(CloseProc == "NV_SESSION_CLOSE_NONE")
			CloseProc = "";
		Session->m_sCloseProc = CloseProc;
		if(!UserHere.length())
			UserHere = "nvdef";

		string SSOUser ="";
		string SSODomain ="";
		string SSOErrInfo ="";
		int SSORetVal = SSOAuthenticate(AppliNameHere, UserHere, SSOUser, SSODomain, false, SSOErrInfo);	
		if(SSORetVal <= 0)
		{
			// SSO failed
			// We prepare error information
			// We need user name in session for error message
			// In case of SSO, the UserName is eventually the SSOUser name
			NvApplication *Application = GetMain()->Applications.GetApplication(AppliNameHere.c_str());
			if(Application != NULL)
			{
				if(Application->GetNirvaIsSsoUser())
					UserHere = SSOUser;
				GetMain()->Applications.FreeApplication(AppliNameHere.c_str());
			}

			Session->SetUserName(UserHere.c_str());

			string Info = "user=";
			
			if(UserHere.length())
				Info += UserHere;
			else
				Info += "(Default user)";
			Info += " - application=";
			if(AppliNameHere.length())
				Info += AppliNameHere;
			else
				Info += "(Default application)";

			if(SSORetVal == 0 && SSOErrInfo.length())
			{
				Info += " - ";
				Info += SSOErrInfo;
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_AUTHENTICATION_ERROR, Info);
			}
			if(SSORetVal == -4)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_APPLICATION, Info);

			if(SSORetVal != -200)
				SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);

			Session->SetUserName("");
			// We free the session
			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

			// And remove it
			GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());

			Session = NULL;
			SetSession(Session);

			if(SSORetVal == -200)
				SetHttpError(HTTP_ERROR_UNAUTHORIZED, "", false);

			SetCookieSession(NULL);
			return;
		}
		else
		{
			// SSO stuff was ok so we can login
			bool bSSO = false;
			if(SSORetVal == 2)
				bSSO = true;

			// If in sso and there is a post request from web, we must close the http connection after sending
			// the result. This is to force browser 
			//if(bSSO && iInputType == NV_HTTP && InputMessage.iMethod == HTTP_METHOD_POST)
			//	OutputMessage.AddHeader("Connection", "close");

			// Attach the session to an nvd console if there is such request
			int AttachType = ATTACH_REQUEST_TYPE_WEB;
			if(iInputType == NV_HTTP_XML)
				AttachType = ATTACH_REQUEST_TYPE_XML;
			else if(iInputType == NV_HTTP_WEBS)
				AttachType = ATTACH_REQUEST_TYPE_WEBS;
			else if(iInputType == NV_HTTP_SOAP)
				AttachType = ATTACH_REQUEST_TYPE_SOAP;
			GetMain()->AttachRequestList.Get(Session, AttachType, AppliNameHere);
			int LoginRetVal = Session->Login(UserHere.c_str(), PasswordHere.c_str(), AppliNameHere.c_str(), SESSION_TYPE_CLIENT, NewPasswordHere.c_str(), NewPasswordHereConfirm.c_str(), bWithNewPassword, bWithNewPasswordConfirm, iSessionTimeOut, Hash.c_str(), LanguageHere.c_str(), &InputMessage.mParameters, &Command, bSSO, SSOUser.c_str(), SSODomain.c_str());
			if(LoginRetVal != 1)
			{
				// Login failed
				// We prepare error information

				// We need user name in session for error message
				// In case of SSO, the UserName is eventually the SSOUser name
				if(bSSO)
				{
					NvApplication *Application = GetMain()->Applications.GetApplication(AppliNameHere.c_str());
					if(Application != NULL)
					{
						if(Application->GetNirvaIsSsoUser())
							UserHere = SSOUser;
						GetMain()->Applications.FreeApplication(AppliNameHere.c_str());
					}
				}

				Session->SetUserName(UserHere.c_str());

				string Info = "user=";
				
				if(UserHere.length())
					Info += UserHere;
				else
					Info += "(Default user)";
				Info += " - application=";
				if(AppliNameHere.length())
					Info += AppliNameHere;
				else
					Info += "(Default application)";


				if(LoginRetVal < 0)
				{
					if(LoginRetVal == -1)
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_USER, Info);
					if(LoginRetVal == -2)
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CANNOT_GET_CONTEXT, Info);
					if(LoginRetVal == -3)
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD, Info);
					if(LoginRetVal == -4)
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_APPLICATION, Info);
					if(LoginRetVal == -5)
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CHANGE_PASSWORD, Info);
					if(LoginRetVal == -6)
						Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD_SYNTAX, Info);
				}
				else
					Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTLOGIN, Info);
				
				SetHttpErrorBrowserCommand(iOutputType, iXmlCoding, true);

				Session->SetUserName("");
				// We free the session
				GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

				// And remove it
				GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());

				Session = NULL;
				SetSession(Session);
				SetCookieSession(NULL);

				return;
			}
		}
//		string UserHereForCheck = UserHere;
//		CleanString(UserHereForCheck, true);
//		MakeLower(UserHereForCheck);
//		if(Session != NULL)
//			if(UserHereForCheck == "nvadmin")
//				Session->m_bAdministrator = true; // Administrator user
	}

	
	if(iInputType == NV_HTTP_WEBS)
	{
		// For web services, the command is always WEBS EXECUTE with a parameter named OPERATION that
		// comes from the SOAPAction header but can be overriden
		ServiceHere = "SYSTEM";
		ClassHere = "WEBSERVICE";
		CommandHere = "EXECUTE";
		string Operation = InputMessage.GetParameter("OPERATION");
		CleanString(Operation);
		MakeLower(Operation);
		if(!Operation.length())
			Operation = WebsOperation;
		if(!Operation.length())
			Operation = "nvdef";
		InputMessage.mParameters["OPERATION"] = Operation;
		// The web service name is inside the WEBSERVICE parameter (set before session open)
	}

	// Get the input container name from command
	string InContainerHere = InputMessage.GetParameter("NV_CONTAINER");
	string OutContainerHere = InContainerHere;

	// Get the input container name from command
	if(InputMessage.ParameterExist("NV_IN_CONTAINER"))
		InContainerHere = InputMessage.GetParameter("NV_IN_CONTAINER");
	MakeLower(InContainerHere);
	CleanString(InContainerHere);

	// Get the output container name from command
	if(InputMessage.ParameterExist("NV_OUT_CONTAINER"))
		OutContainerHere = InputMessage.GetParameter("NV_OUT_CONTAINER");
	MakeLower(OutContainerHere);
	CleanString(OutContainerHere);

	if(!InContainerHere.length())
		InContainerHere = NV_CONTAINER_DEFAULT_NAME;
	if(!OutContainerHere.length())
		OutContainerHere = NV_CONTAINER_DEFAULT_NAME;
	
//	if(iInputType == NV_HTTP_WEBS)
//	{
//		// For web services, the default input container is in
//		// and the default output is out
//		if(!InContainerHere.length())
//			InContainerHere = "in";
//		if(!OutContainerHere.length())
//			OutContainerHere = "out";
//	}


	// Get the language from command again because now we may have the session
	// that gives the default language
	LanguageHere = InputMessage.GetParameter("NV_LANGUAGE");
	MakeUpper(LanguageHere);
	CleanString(LanguageHere);
	if(!LanguageHere.length())
	{
		if(Session != NULL)
			LanguageHere = Session->m_sLanguage;
		else
			LanguageHere = "ENGLISH";
	}

	// Get the close session from command
	string CloseSessionCommand = InputMessage.GetParameter("NV_CLOSE_SESSION");
	MakeUpper(CloseSessionCommand);
	CleanString(CloseSessionCommand);
	bool bCloseSessionFromCommand = false;
	if((iInputType == NV_HTTP_XML) || (iInputType == NV_HTTP_SOAP) || (iInputType == NV_HTTP_WEBS))
		if(SessionCreated)
			bCloseSessionFromCommand = true;	// default is yes for XML or SOAP connector except if a session ID has been given
	if(CloseSessionCommand == "YES")
		bCloseSessionFromCommand = true;;
	if(CloseSessionCommand == "NO")
		bCloseSessionFromCommand = false;;


	// Now we can prepare the command
	Command.m_iCommandSource = NV_SOURCE_BROWSER;
	if(bFromClient)
		Command.m_iCommandSource = NV_SOURCE_CLIENT;
	Command.m_sService = ServiceHere;
	Command.m_sClass = ClassHere;
	Command.m_sCommand = CommandHere;
	Command.m_sInContainer = InContainerHere;
	Command.m_sOutContainer = OutContainerHere;
//	if(!Command.m_sOutContainer.length())
//		Command.m_sOutContainer = Command.m_sInContainer;
	Command.m_sHost = InputMessage.GetHeader("X-FORWARDED-HOST");
	if(!Command.m_sHost.length())
		Command.m_sHost = InputMessage.GetHeader("HOST");
	Command.m_sLanguage = LanguageHere;



	for(pos = InputMessage.mParameters.begin(); pos != InputMessage.mParameters.end(); ++pos)
	{
		string ParameterNameHere = pos->first;
		CleanString(ParameterNameHere);
		string ParameterValueHere = pos->second;
//		CleanString(ParameterValueHere);
		string Provistr = ParameterNameHere;
		MakeUpper(Provistr);
		bool ReservedKeyWord = false;
		if(Provistr == "NV_SERVICE")
			ReservedKeyWord = true;
		if(Provistr == "NV_CLASS")
			ReservedKeyWord = true;
		if(Provistr == "NV_COMMAND")
			ReservedKeyWord = true;
		if(Provistr == "NV_CMD")
			ReservedKeyWord = true;
		if(Provistr == "NV_IN_CONTAINER")
			ReservedKeyWord = true;
		if(Provistr == "NV_OUT_CONTAINER")
			ReservedKeyWord = true;
		if(Provistr == "NV_LANGUAGE")
			ReservedKeyWord = true;
		if(Provistr == "NV_CLOSE_SESSION")
			ReservedKeyWord = true;
		if(Provistr == "NV_NO_ERROR")
			ReservedKeyWord = true;

		if(ReservedKeyWord)
			continue;

		if(!ParameterNameHere.length())
			continue;

		Command.Parameters.SetValue(ParameterNameHere, ParameterValueHere);
	}

	// Manage HTTP POST
	if((iInputType == NV_HTTP) && (InputMessage.iMethod == HTTP_METHOD_POST))
	{
		if(HttpFileInput.length())
		{
			// We put all the message content into an object file
			NvContainer *InContainer = Session->GetContainer(Command.m_sInContainer.c_str(), true, &Command);
			NvObjFile *Object = NULL;
			if(InContainer != NULL)
			{
				InContainer->RemoveObject(HttpFileInput.c_str());
				InContainer->CreateObject(HttpFileInput.c_str(), NV_OBJTYPE_FILE);
				Object = (NvObjFile *)(InContainer->GetObject(HttpFileInput.c_str()));
			}
			if(Object != NULL)
			{
				// These files cannot be persistent
				int iPersistTimeFileInCommand = 0;

				string FileExtensionInCommand = Command.Parameters.GetValue("EXTENSION");
				CleanString(FileExtensionInCommand);
				if(!FileExtensionInCommand.length())
					FileExtensionInCommand = "tmp";

				string FilePrefixInCommand = Command.Parameters.GetValue("PREFIX");
				CleanString(FilePrefixInCommand);
				if(!FilePrefixInCommand.length())
					FilePrefixInCommand = Session->Application->GetAppNameForFiles();
				string FileSuffixInCommand = Command.Parameters.GetValue("SUFFIX");
				CleanString(FileSuffixInCommand);
				string TargetDir = Session->Application->GetWorkDir();

				if(!Object->SetAutoFileName(TargetDir, FileExtensionInCommand, FilePrefixInCommand, FileSuffixInCommand))
					InContainer->RemoveObject(HttpFileInput.c_str());
				else
					Object->SetPersistTime(iPersistTimeFileInCommand);

				unsigned NvInt64 bodysize = (unsigned NvInt64)(InputMessage.vDatas[0]->DataSize);
				if(InputMessage.vDatas[0]->DataFile.length())
					bodysize = GetFileLength(InputMessage.vDatas[0]->DataFile.c_str());
				if(bodysize)
				{
					if(InputMessage.vDatas[0]->DataFile.length())
					{
						// We populate the file object directly from the data file in data object
						string ObjectFileName = Object->GetFilePathName();
						CopyFileEx(InputMessage.vDatas[0]->DataFile.c_str(), ObjectFileName.c_str());
					}
					else
						Object->Populate(InputMessage.vDatas[0]->Data, InputMessage.vDatas[0]->DataSize);
				}
			}
			// We don't need the input data any more
			InputMessage.vDatas[0]->FreeData();
			
			// We free the container
			if(InContainer != NULL)
				Session->FreeContainer(InContainer, Command.m_sInContainer.c_str(), true, &Command);
		}
		else if(InputMessage.bMultipart)
		{
			// We must create the eventual object if we have a POST command from a form with files
			NvContainer *InContainer = Session->GetContainer(Command.m_sInContainer.c_str(), true, &Command);
			string::size_type index;
			int i;
			for(i=1; i<(int)(InputMessage.vDatas.size()); i++)
			{
				string ContentDisposition = InputMessage.GetHeader("CONTENT-DISPOSITION", i);
				if(!ContentDisposition.length())
					continue;
				if(ContentDisposition.find("form-data") == string::npos)
					continue;
				string ParameterName = "";
				while((index = ContentDisposition.find(';')) != string::npos)
				{
					ContentDisposition.erase(0, index+1);
					CleanString(ContentDisposition);
					index = ContentDisposition.find("name=");
					if(index != 0)
						continue;
					ContentDisposition.erase(0, index+5);
					index = ContentDisposition.find(';');
					if(index != string::npos)
						ContentDisposition = ContentDisposition.substr(0, index);
					CleanString(ContentDisposition);
					if(ContentDisposition.length() && (ContentDisposition[0] == '\"'))
						ContentDisposition.erase(0, 1);
					if(ContentDisposition.length() && (ContentDisposition[ContentDisposition.length()-1] == '\"'))
						ContentDisposition.erase(ContentDisposition.size()-1, 1);
					CleanString(ContentDisposition);
					if(ContentDisposition.length())
						ParameterName = ContentDisposition;
					break;
				}
				if(ParameterName.length())
				{
					ContentDisposition = InputMessage.GetHeader("CONTENT-DISPOSITION", i);
					if((index = ContentDisposition.find("filename=")) == string::npos)
						continue;	// This is not a file content
				
					// Try to get the encoding of the input form (PM Modif 2012/09/03 following bug on filename encoding)
					bool bFormUnicode = true;
					bool bAutoDetectEncoding = GetMain()->GetAutoDetectEncoding();
					if(Charset == "ISO-8859-1")
					{
						bFormUnicode = false;
						bAutoDetectEncoding = false;
					}
					if(Charset == "UTF-8")
					{
						bFormUnicode = true;
						bAutoDetectEncoding = false;
					}
					string EncodingFromParam = InputMessage.GetParameter("NV_FORM_ENCODING");
					CleanString(EncodingFromParam);
					MakeUpper(EncodingFromParam);
					if(EncodingFromParam == "UTF-8")
					{
						bFormUnicode = true;
						bAutoDetectEncoding = false;
					}
					if(EncodingFromParam == "ISO-8859-1")
					{
						bFormUnicode = false;
						bAutoDetectEncoding = false;
					}
					if(EncodingFromParam == "AUTO")
					{
						bAutoDetectEncoding = true;
					}
					string Origin = ContentDisposition.substr(index+9);
					index = Origin.find(';');
					if(index != string::npos)
						Origin = Origin.substr(0, index);
					CleanString(Origin);
					// remove double quotes
					while((index = Origin.find('\"')) != string::npos)
						Origin.erase(index, 1);
					// eliminate single quotes
					while((index = Origin.find('\'')) != string::npos)
						Origin.erase(index, 1);
					// keep only file name if complete path is given
					if((index = Origin.rfind('\\')) != string::npos)
						Origin = Origin.substr(index+1);
					if((index = Origin.rfind('/')) != string::npos)
						Origin = Origin.substr(index+1);
					DecodeUrlString(ParameterName, bFormUnicode, false, bAutoDetectEncoding);
					DecodeUrlString(Origin, bFormUnicode, false, bAutoDetectEncoding);
					MakeLower(ParameterName);	// the parameter name is case insensitive
					NvObjFile *Object = NULL;
					if(InContainer != NULL)
					{
						InContainer->RemoveObject(ParameterName.c_str());
						InContainer->CreateObject(ParameterName.c_str(), NV_OBJTYPE_FILE);

						// Check the persist parameter
						string PersistTimeFileInCommand = Command.Parameters.GetValue("PERSIST");
						CleanString(PersistTimeFileInCommand);
						MakeUpper(PersistTimeFileInCommand);
						int iPersistTimeFileInCommand = 0;	// The file is not persistant by default on the server
						if(PersistTimeFileInCommand.length())
							iPersistTimeFileInCommand = atoi(PersistTimeFileInCommand.c_str());
						if(iPersistTimeFileInCommand < 0)
							iPersistTimeFileInCommand = -1;


						string FileExtensionInCommand = Command.Parameters.GetValue("EXTENSION");
						CleanString(FileExtensionInCommand);
						if(!FileExtensionInCommand.length())
						{
							// We try to get it from the origin
							if(Origin.length())
							{
								index = Origin.rfind('.');
								if(index != string::npos)
									FileExtensionInCommand = Origin.substr(index+1);
							}
							else
							{
								if(iPersistTimeFileInCommand == -1)
									FileExtensionInCommand = "obj";
								else
									FileExtensionInCommand = "tmp";
							}
						}
						string FilePrefixInCommand = Command.Parameters.GetValue("PREFIX");
						CleanString(FilePrefixInCommand);
						if(!FilePrefixInCommand.length())
							FilePrefixInCommand = Session->Application->GetAppNameForFiles();
						string FileSuffixInCommand = Command.Parameters.GetValue("SUFFIX");
						CleanString(FileSuffixInCommand);
						string TargetDir = Session->Application->GetWorkDir();
						if(iPersistTimeFileInCommand == -1)
						{
							// If the file is persistant, it's created in the application File directory
							TargetDir = Session->Application->GetFileDir();
						}

						Object = (NvObjFile *)(InContainer->GetObject(ParameterName.c_str()));
						if(Object != NULL)
						{

							if(!Object->SetAutoFileName(TargetDir, FileExtensionInCommand, FilePrefixInCommand, FileSuffixInCommand))
								InContainer->RemoveObject(ParameterName.c_str());
							else
								Object->SetPersistTime(iPersistTimeFileInCommand);

							Object->SetOrigin(Origin);
						}
					}

					unsigned NvInt64 bodysize = (unsigned NvInt64)(InputMessage.vDatas[i]->DataSize);
					if(InputMessage.vDatas[i]->DataFile.length())
						bodysize = GetFileLength(InputMessage.vDatas[i]->DataFile.c_str());
					if(bodysize)
					{
						if(Object != NULL)
						{
							if(InputMessage.vDatas[i]->DataFile.length())
							{
								// We populate the file object directly from the data file in data object
								string ObjectFileName = Object->GetFilePathName();
								CopyFileEx(InputMessage.vDatas[i]->DataFile.c_str(), ObjectFileName.c_str());
							}
							else
								Object->Populate(InputMessage.vDatas[i]->Data, InputMessage.vDatas[i]->DataSize);
						}
					}
					// We don't need the input data any more
					InputMessage.vDatas[i]->FreeData();
				}
			}
			Session->FreeContainer(InContainer, Command.m_sInContainer.c_str(), true, &Command);
		}
	}


	if((iInputType == NV_HTTP_XML) || (iInputType == NV_HTTP_SOAP) || (iInputType == NV_HTTP_WEBS))
	{
		Command.iXmlInputType = XML_IN_XML;
		if(iInputType == NV_HTTP_SOAP)
			Command.iXmlInputType = XML_IN_SOAP;
		if(iInputType == NV_HTTP_WEBS)
			Command.iXmlInputType = XML_IN_SOAP;
		Command.iXmlCoding = iXmlCoding;

		if(!Charset.length())
			Charset="UTF-8";
		Command.m_sInputCharset = Charset;
	}

	// We execute the command
	bool SessionRetVal = true;
	try
	{
		SessionRetVal = Session->Command(&Command);
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		string Message = Command.m_sService;
		Message += ":";
		Message += Command.m_sClass;
		Message += ":";
		Message += Command.m_sCommand;
		Command.Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_EXCEPTION, Message);
		GetMain()->Logs.Write("Exceptions", "Exception in command(Web)", Message.c_str(), NV_LOG_ERROR);
		SessionRetVal = false;
	}

	if(SessionRetVal)
	{
		// now, we manage output data
		// This should be something readable from a browser

		// Set http output headers

		for(pos = Command.HttpOutHeaders.m_mList.begin(); pos != Command.HttpOutHeaders.m_mList.end(); ++pos)
		{
			string HeaderNameHere = pos->first;
			CleanString(HeaderNameHere);
			string  HeaderValueHere = pos->second;
//			CleanString(HeaderValueHere);

			if(!HeaderNameHere.length())
				continue;

			OutputMessage.AddHeader(HeaderNameHere, HeaderValueHere);
		}
		
		OutputMessage.AddHeader("Content-Length", "0");

		// The ownership of ouput data is transfered from command to http part
		OutputMessage.vDatas[0]->DataSize = Command.DataSize;
		OutputMessage.vDatas[0]->Data = Command.Data;
		OutputMessage.vDatas[0]->DataFile = Command.DataFile;
		OutputMessage.vDatas[0]->bDeleteDataFile = Command.bDeleteDataFile;
		OutputMessage.AddHeader("Content-Length", "0");
		if(Command.DataSize)
		{
			char ProviBuf[50];
			sprintf(ProviBuf, "%ld", (unsigned long)(Command.DataSize));
			OutputMessage.AddHeader("Content-Length", ProviBuf);
		}
		Command.DataSize = 0;
		Command.Data = NULL;
		Command.DataFile = "";
		Command.bDeleteDataFile = false;
	}
	
	if(bNoError)
		Command.Error.SetError();

	if(Command.Error.m_iErrorCode && (Session != NULL))
	{
		// In case of error check security to erase the info field
			if(!Session->CheckPermission(NULL, "SYSTEM", "DISPLAY_ERROR_INFO"))
				Command.Error.m_sErrorInfo = "";

	}

	
	if((SessionRetVal) && (Command.m_sService == "SYSTEM") && (Command.m_sClass == "SESSION") && (Command.m_sCommand == "CLOSE"))
	{
		// The order is to close the session
		if(Session != NULL)
			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		Session = NULL;
		if(InputMessage.ParameterExist("LOGOUT_PAGE"))
		{
			string LogoutPage = InputMessage.GetParameter("LOGOUT_PAGE");
			GetLogoutPage(LogoutPage);
			SetSession(Session);
			SetCookieSession(NULL);
			return;
		}
	}
	else
	{
		// Get the close session from command
		// if an internal request has been made
		string CloseSessionCommand = Command.Parameters.GetValue("NV_CLOSE_SESSION");
		if(CloseSessionCommand == "YES")
			bCloseSessionFromCommand = true;;
		if(CloseSessionCommand == "NO")
			bCloseSessionFromCommand = false;;
		if(bCloseSessionFromCommand)
		{
			// We also close the session if explicitly requested as a command parameter
			if(Session != NULL)
				GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
			GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
			Session = NULL;
		}
	}
//	else if(SessionCreated && Command.Error.m_iErrorCode)
//	{
//		// We also close the session if an error occured and the session has just been created
//		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
//		Session = NULL;
//	}

	if(Session != NULL)
		OutputMessage.AddHeader("Nv_Session_Id", Session->m_sIdent);
	SetSession(Session);
	SetCookieSession(Session);
	// The error information is generated at the end in order for it to eventually not contain the session ID
	// if closed
	if(Command.Error.m_iErrorCode)
	{
		SetHttpErrorBrowserCommand(iOutputType, iXmlCoding);
//		SetHttpError(HTTP_ERROR_INTERNAL, Info);
	}
	
	if(Session != NULL)
		GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
	Session = NULL;
	SetSession(Session);
	// Cleanup the actual command
	Command.Clear();

}


void NvHttpClient::GetInputFormValues(string ContentType, string Charset)
{
	// Try to get the encoding of the input form
	bool bFormUnicode = true;
	bool bAutoDetectEncoding = GetMain()->GetAutoDetectEncoding();
	if(Charset == "ISO-8859-1")
	{
		bFormUnicode = false;
		bAutoDetectEncoding = false;
	}
	if(Charset == "UTF-8")
	{
		bFormUnicode = true;
		bAutoDetectEncoding = false;
	}
	string EncodingFromParam = InputMessage.GetParameter("NV_FORM_ENCODING");
	CleanString(EncodingFromParam);
	MakeUpper(EncodingFromParam);
	if(EncodingFromParam == "UTF-8")
	{
		bFormUnicode = true;
		bAutoDetectEncoding = false;
	}
	if(EncodingFromParam == "ISO-8859-1")
	{
		bFormUnicode = false;
		bAutoDetectEncoding = false;
	}
	if(EncodingFromParam == "AUTO")
	{
		bAutoDetectEncoding = true;
	}

	// We must get information sent from the browser
	// Especially the parameters comming from a HTML form
	string::size_type index;
	if(InputMessage.bMultipart)
	{
		int i;
		for(i=1; i<(int)(InputMessage.vDatas.size()); i++)
		{
			string ContentDisposition = InputMessage.GetHeader("CONTENT-DISPOSITION", i);
			if(!ContentDisposition.length())
				continue;
			if(ContentDisposition.find("form-data") == string::npos)
				continue;
			string ParameterName = "";
			string ParameterValue = "";
			while((index = ContentDisposition.find(';')) != string::npos)
			{
				ContentDisposition.erase(0, index+1);
				CleanString(ContentDisposition);
				index = ContentDisposition.find("name=");
				if(index != 0)
					continue;
				ContentDisposition.erase(0, index+5);
				index = ContentDisposition.find(';');
				if(index != string::npos)
					ContentDisposition = ContentDisposition.substr(0, index);
				CleanString(ContentDisposition);
				if(ContentDisposition.length() && (ContentDisposition[0] == '\"'))
					ContentDisposition.erase(0, 1);
				if(ContentDisposition.length() && (ContentDisposition[ContentDisposition.length()-1] == '\"'))
					ContentDisposition.erase(ContentDisposition.size()-1, 1);
				CleanString(ContentDisposition);
				if(ContentDisposition.length())
					ParameterName = ContentDisposition;
				break;
			}
			if(ParameterName.length())
			{
				ContentDisposition = InputMessage.GetHeader("CONTENT-DISPOSITION", i);
				if(ContentDisposition.find("filename=") != string::npos)
					continue;	// This a file content, we will manage it later
				unsigned long bodysize = InputMessage.vDatas[i]->DataSize;
				if(InputMessage.vDatas[i]->DataFile.length())
					bodysize = (unsigned long)(GetFileLength(InputMessage.vDatas[i]->DataFile.c_str()));
				if(bodysize)
				{
					char *PostData = (char *)malloc(bodysize+1);
					if(InputMessage.vDatas[i]->DataFile.length())
					{
						// We must read the file
						FILE *File;
						File = Nvfopen(InputMessage.vDatas[i]->DataFile.c_str(), "rb");
						if(File != NULL)
						{
							fread(PostData, 1, bodysize, File);
							fclose(File);
						}
					}
					else
						memcpy(PostData, InputMessage.vDatas[i]->Data, bodysize);
					PostData[bodysize] = '\0';
					ParameterValue = PostData;
					free(PostData);
//						CleanString(ParameterValue);
					// We don't need the input data any more
					InputMessage.vDatas[i]->FreeData();
				}
				DecodeUrlString(ParameterName, bFormUnicode, false, bAutoDetectEncoding);
				DecodeUrlString(ParameterValue, bFormUnicode, false, bAutoDetectEncoding);
				MakeUpper(ParameterName);	// the parameter name is case insensitive
				InputMessage.mParameters[ParameterName] = ParameterValue;
			}
		}
	}
	else
	{
		unsigned long bodysize = InputMessage.vDatas[0]->DataSize;
		if(InputMessage.vDatas[0]->DataFile.length())
			bodysize = (unsigned long)(GetFileLength(InputMessage.vDatas[0]->DataFile.c_str()));
		if(bodysize)
		{
			bool bUrlEncoded = false;
			if(ContentType.find("application/x-www-form-urlencoded") != string::npos)
				bUrlEncoded = true;
			if(bUrlEncoded)
			{
				char *PostData = (char *)malloc(bodysize+1);
				if(InputMessage.vDatas[0]->DataFile.length())
				{
					// We must read the file
					FILE *File;
					File = Nvfopen(InputMessage.vDatas[0]->DataFile.c_str(), "rb");
					if(File != NULL)
					{
						fread(PostData, 1, bodysize, File);
						fclose(File);
					}
				}
				else
					memcpy(PostData, InputMessage.vDatas[0]->Data, bodysize);
				PostData[bodysize] = '\0';
				string PostParam = PostData;
				free(PostData);
				if(PostParam.length() && (PostParam[0] != '&'))
				{
					string Provistr = "&";
					Provistr += PostParam;
					PostParam = Provistr;
				}
				while((index = PostParam.rfind('&')) != string::npos)
				{
					string sParameter = PostParam.substr(index+1);
					PostParam.erase(index);
					if((index = sParameter.find('=')) != string::npos)
					{
						string sValue = sParameter.substr(index+1);
						sParameter.erase(index);
						if(sParameter.size())
						{
							DecodeUrlString(sParameter, bFormUnicode, true, bAutoDetectEncoding);
							DecodeUrlString(sValue, bFormUnicode, true, bAutoDetectEncoding);
							MakeUpper(sParameter);	// the parameter name is case insensitive
							InputMessage.mParameters[sParameter] = sValue;
						}
					}
				}
			}
			// We don't need the input data any more
			InputMessage.vDatas[0]->FreeData();
		}
	}
}

void NvHttpClient::SetCookieSession(NvSession* Session)
{
	string CookieValue = "NvSessionId=; expires=Thu, 01-Jan-1970 00:00:00 GMT";
	if((Session != NULL) && Session->m_bUseSessionCookie)
	{
		// Set Session cookie
		CookieValue = "NvSessionId=";
		CookieValue += Session->m_sIdent;
		if(Session->m_iSessionCookieTimeOut > 0)
		{
			char Buf[256];
			time_t tTime = time(NULL) + (time_t)(Session->m_iSessionCookieTimeOut);
			struct tm tCurrentTime;
			gmtime_r(&tTime, &tCurrentTime);
			char wday_name[7][4] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
			static const char mon_name[12][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
			sprintf(Buf, "; expires=%s, %02d-%s-%04d %02d:%02d:%02d GMT",
							wday_name[tCurrentTime.tm_wday],
							tCurrentTime.tm_mday,
							mon_name[tCurrentTime.tm_mon],
							tCurrentTime.tm_year + 1900,
							tCurrentTime.tm_hour,
							tCurrentTime.tm_min,
							tCurrentTime.tm_sec); 
			CookieValue += Buf;
		}
	}
	string ParameterValue = OutputMessage.GetHeader("Set-Cookie");
	if(ParameterValue.length())
		ParameterValue += "\n";
	ParameterValue += CookieValue;
	CookieValue = ParameterValue;
	OutputMessage.AddHeader("Set-Cookie", CookieValue);
}


// Do SSO stuff
// return values:
//  1   -> Ok, don't use SSO
//  2   -> Ok, use SSO, SSOUser and SSODomain are set
//  0   -> General error (cannot authenticate)
// -4   -> Application not enabled
// -200 -> in negotiation
int NvHttpClient::SSOAuthenticate(string AppliName, string NirvaUserName, string& SSOUSer, string& SSODomain, bool FromNvc, string& ErrInfo)
{
	SSOUSer = "";
	SSODomain = "";

	FILE *LogSSO = NULL;
	string LogMessage = "";
	if(bDebugSso)
	{
		string LogFileName = GetMain()->GetNirvaDir();
		AddDirSuffix(LogFileName);
		LogFileName += "Bin";
		AddDirSuffix(LogFileName);
		LogFileName += "ssolog.txt";
		LogSSO = Nvfopen(LogFileName.c_str(), "a+b");
	}

	// Get The authorization header
	// We must force authentication if the input request contains an authorization header 
	bool bForceAuthentication = false;
	string sAuthorization = InputMessage.GetHeader("AUTHORIZATION");
	if(sAuthorization.length())
		bForceAuthentication = true;
	if(FromNvc)	// We never force authorisation if from nvc
		bForceAuthentication = false;

	CleanString(NirvaUserName, true);
	MakeLower(NirvaUserName);
	if(LogSSO != NULL)
	{
		LogMessage = "\r\n\r\nStart SSOAuthenticate function \r\nNirva user is ";
		LogMessage += NirvaUserName;
		LogMessage += "\r\nSource=";
		if(FromNvc)
			LogMessage += "Nirva client";
		else
			LogMessage += "Browser or other HTTP client\r\nObject=";
		char buf[20];
		sprintf(buf,"%lx", (unsigned long)this);
		LogMessage += buf;
		LogMessage += "\r\n";
		fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
	}
	if(LogSSO != NULL)
	{
		LogMessage = "Authorization header = " + sAuthorization + "\r\n";
		fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
	}
	
	if(!bForceAuthentication)
	{
		// If the user is nvadmin, we cannot authenticate via SSO
		if(NirvaUserName == "nvadmin")
		{
			SsoInfo.Clear();
			if(LogSSO != NULL)
			{
				LogMessage = "The user is nvadmin: no SSO in this case\r\n";
				fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
			}
			if(LogSSO != NULL)
				fclose(LogSSO);
			return 1;
		}
	}

	bool bAppSSO = false;	// This flag tells if we must SSO from app point of view (whatever the bForceAuthentication flag)
	CleanString(AppliName);
	MakeUpper(AppliName);
	if(!AppliName.c_str())
		AppliName = NV_APPLICATION_DEFAULT_NAME;
	NvApplication *Application = GetMain()->Applications.GetApplication(AppliName.c_str());
	if(Application == NULL)
	{
		SsoInfo.Clear();
		if(LogSSO != NULL)
		{
			LogMessage = "Application doesn't exist\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		if(LogSSO != NULL)
			fclose(LogSSO);
		return -4;
	}

	if(!Application->IsEnable())
	{
		GetMain()->Applications.FreeApplication(AppliName.c_str());
		SsoInfo.Clear();
		if(LogSSO != NULL)
		{
			LogMessage = "Application disabled\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		if(LogSSO != NULL)
			fclose(LogSSO);
		return -4;
	}
	int AuthenticationMode = Application->GetAuthMode();
	GetMain()->Applications.FreeApplication(AppliName.c_str());
	if(AuthenticationMode == NV_AUTHENTICATION_SSO)
	{
		if(LogSSO != NULL)
		{
			LogMessage = "Authentication mode set to SSO\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		bAppSSO = true;
	}

	if(bForceAuthentication)
	{
		AuthenticationMode = NV_AUTHENTICATION_SSO;
		if(LogSSO != NULL)
		{
			LogMessage = "SSO forced\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
	}

	if(AuthenticationMode == NV_AUTHENTICATION_NIRVA)
	{
		SsoInfo.Clear();
		if(LogSSO != NULL)
		{
			LogMessage = "Authentication mode set to Nirva: no SSO\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		if(LogSSO != NULL)
			fclose(LogSSO);
		return 1;
	}

	bool bDoSSO = true;
	if(AuthenticationMode == NV_AUTHENTICATION_NIRVA_SSO)
	{
		if(LogSSO != NULL)
		{
			LogMessage = "Authentication mode set to client choice: ";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		// Client choice
		string sDoSSO = InputMessage.GetHeader("NV_SSO");
		CleanString(sDoSSO);
		MakeUpper(sDoSSO);
		if(sDoSSO != "YES")
			bDoSSO = false;
		else
			bAppSSO = true;
		if(LogSSO != NULL)
		{
			if(bDoSSO)
				LogMessage = "do SSO\r\n";
			else
				LogMessage = "no SSO\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
	}

	if(!bDoSSO)
	{
		SsoInfo.Clear();
		if(LogSSO != NULL)
			fclose(LogSSO);
		return 1;	// Do not authenticate via SSO
	}

	// Now we must authenticate via SSO
	
	// eliminate first and last spaces
	while(sAuthorization.size() && sAuthorization[0] == ' ')
		sAuthorization.erase(0, 1);
	while(sAuthorization.size() && sAuthorization[sAuthorization.size()-1] == ' ')
		sAuthorization.erase(sAuthorization.size()-1, 1);

	string::size_type index = sAuthorization.find(' ');
	string Method = sAuthorization;
	if(index != string::npos)
	{
		Method = sAuthorization.substr(0, index);
		sAuthorization = sAuthorization.substr(index+1);
	}
	else
		sAuthorization = "";
	CleanString(Method);
#ifdef WIN32
	if(SsoInfo.secPackInfo == NULL)
		QuerySecurityPackageInfo((SEC_CHAR *)(Method.c_str()), &(SsoInfo.secPackInfo));
	if(!sAuthorization.length())
	{
		// This is the first request so we must answer with a WWW-Authenticate header set to the correct method
//		string HeaderValue = "Kerberos";
		string HeaderValue = "Negotiate\r\nWWW-Authenticate: Kerberos\r\nWWW-Authenticate: NTLM";
		OutputMessage.AddHeader("WWW-Authenticate", HeaderValue);

		SsoInfo.Clear();
		if(LogSSO != NULL)
		{
			LogMessage = "First request WWW-Autenticate: ";
			LogMessage += HeaderValue;
			LogMessage += "\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		if(LogSSO != NULL)
			fclose(LogSSO);
		return -200;
	}
	SECURITY_STATUS err;
	if(!SsoInfo.bValidhcred)
	{
		if(LogSSO != NULL)
		{
			LogMessage = "Acquire credential handle for the the current user (the one running nirva)\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		err = AcquireCredentialsHandle(NULL,
									(SEC_CHAR *)(Method.c_str()),
									SECPKG_CRED_INBOUND,
									NULL,
									NULL,
									NULL,
									NULL,
									&(SsoInfo.hcred),
									&(SsoInfo.expiry));
		if (err != SEC_E_OK)
		{
			SsoInfo.Clear();
			ErrInfo = "AcquireCredentialsHandle returned an error number ";
			char Buff[20];
			sprintf(Buff, "%x", err);
			ErrInfo += Buff;
			if(LogSSO != NULL)
			{
				LogMessage = ErrInfo;
				LogMessage += "\r\n";
				fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
			}
			if(LogSSO != NULL)
				fclose(LogSSO);
			return 0;
		}
		SsoInfo.bValidhcred = true;
	}

	// prepare to get the client's response
	if(SsoInfo.ib.pvBuffer != NULL)
	{
		free(SsoInfo.ib.pvBuffer);
		SsoInfo.ib.pvBuffer = NULL;
	}
	SsoInfo.ibd.ulVersion = SECBUFFER_VERSION;
	SsoInfo.ibd.cBuffers = 1;
	SsoInfo.ibd.pBuffers = &(SsoInfo.ib); // just one buffer
	SsoInfo.ib.BufferType = SECBUFFER_TOKEN; // preping a token here

	int DataSize = (int)(sAuthorization.length());
	MemBase64Decode(sAuthorization.c_str(), (char **)(&(SsoInfo.ib.pvBuffer)), &DataSize);
	SsoInfo.ib.cbBuffer = (unsigned long)(DataSize);
	SsoInfo.haveInbuffer = true;
	string Hash = InputMessage.GetHeader("NV_AUTHENTICATE_DATA_HASH");
	if(Hash.length())
	{
		string CalculatedHash = MD5_string(sAuthorization);
		if(CalculatedHash != Hash)
		{
			SsoInfo.Clear();
			ErrInfo = "Received hash from nirva client differs from calculated hash";
			if(LogSSO != NULL)
			{
				LogMessage = ErrInfo;
				LogMessage += "\r\n";
				fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
			}
			if(LogSSO != NULL)
				fclose(LogSSO);
			return 0;
		}
	}

	if(SsoInfo.ob.pvBuffer != NULL)
	{
		free(SsoInfo.ob.pvBuffer);
		SsoInfo.ob.pvBuffer = NULL;
	}
	SsoInfo.obd.ulVersion = SECBUFFER_VERSION;
	SsoInfo.obd.cBuffers = 1;
	SsoInfo.obd.pBuffers = &(SsoInfo.ob); // just one buffer
	SsoInfo.ob.BufferType = SECBUFFER_TOKEN; // preping a token here
	SsoInfo.ob.cbBuffer = SsoInfo.secPackInfo->cbMaxToken+1;
	SsoInfo.ob.pvBuffer = malloc(SsoInfo.ob.cbBuffer);

	DWORD ctxAttr;
	if(LogSSO != NULL)
	{
		char buf[1024];
		sprintf(buf, "Parameters for AcceptSecurityContext (before): phCredential=%x phContext=%x pInput=%x phNewContext=%x pOutput=%x\r\n", &(SsoInfo.hcred), SsoInfo.haveContext ? &(SsoInfo.srvCtx): NULL, SsoInfo.haveInbuffer ? &(SsoInfo.ibd): NULL, &(SsoInfo.srvCtx), &(SsoInfo.obd));
		LogMessage = buf;
		fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
	}
	err = AcceptSecurityContext(&(SsoInfo.hcred),
								SsoInfo.haveContext ? &(SsoInfo.srvCtx): NULL,
								SsoInfo.haveInbuffer ? &(SsoInfo.ibd): NULL,
								ASC_REQ_REPLAY_DETECT | ASC_REQ_SEQUENCE_DETECT | ASC_REQ_CONFIDENTIALITY | ASC_REQ_DELEGATE,
								SECURITY_NATIVE_DREP,
								&(SsoInfo.srvCtx),
								&(SsoInfo.obd),
								&ctxAttr,
								NULL);
	if(LogSSO != NULL)
	{
		char buf[1024];
		sprintf(buf, "Parameters for AcceptSecurityContext (after): phCredential=%x phContext=%x pInput=%x phNewContext=%x pOutput=%x\r\n", &(SsoInfo.hcred), SsoInfo.haveContext ? &(SsoInfo.srvCtx): NULL, SsoInfo.haveInbuffer ? &(SsoInfo.ibd): NULL, &(SsoInfo.srvCtx), &(SsoInfo.obd));
		LogMessage = buf;
		fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
	}

	if (err == SEC_I_COMPLETE_AND_CONTINUE || err == SEC_I_COMPLETE_NEEDED)
	{
		CompleteAuthToken(&(SsoInfo.srvCtx), &(SsoInfo.obd));
		if (err == SEC_I_COMPLETE_NEEDED)
			err = SEC_E_OK;
		else if (err == SEC_I_COMPLETE_AND_CONTINUE)
			err = SEC_I_CONTINUE_NEEDED;
	}
	if ((err != SEC_E_OK) && (err != SEC_I_CONTINUE_NEEDED))
	{
		ErrInfo = "Unknown";
		switch(err)
		{
			case SEC_E_INSUFFICIENT_MEMORY:
				ErrInfo = "SEC_E_INSUFFICIENT_MEMORY";
				break;
			case SEC_E_INTERNAL_ERROR:
				ErrInfo = "SEC_E_INTERNAL_ERROR";
				break;
			case SEC_E_INVALID_HANDLE:
				ErrInfo = "SEC_E_INVALID_HANDLE";
				break;
			case SEC_E_INVALID_TOKEN:
				ErrInfo = "SEC_E_INVALID_TOKEN";
				break;
			case SEC_E_LOGON_DENIED:
				ErrInfo = "SEC_E_LOGON_DENIED";
				break;
			case SEC_E_NO_AUTHENTICATING_AUTHORITY:
				ErrInfo = "SEC_E_NO_AUTHENTICATING_AUTHORITY";
				break;
			case SEC_E_NO_CREDENTIALS:
				ErrInfo = "SEC_E_NO_CREDENTIALS";
				break;
			case SEC_E_TARGET_UNKNOWN:
				ErrInfo = "SEC_E_TARGET_UNKNOWN";
				break;
			case SEC_E_UNSUPPORTED_FUNCTION:
				ErrInfo = "SEC_E_UNSUPPORTED_FUNCTION";
				break;
			case SEC_E_WRONG_PRINCIPAL:
				ErrInfo = "SEC_E_WRONG_PRINCIPAL";
				break;
		}

		if(LogSSO != NULL)
		{
			LogMessage = "AcceptSecurityContext returned an error number \r\n";
			char Buff[20];
			sprintf(Buff, "%x", err);
			LogMessage += Buff;
			LogMessage += " - " + ErrInfo + "\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		
		SsoInfo.Clear();
		if(LogSSO != NULL)
			fclose(LogSSO);
		return 0;
	}

	// Prepare to send the output to client (even in case of SEC_E_OK)
	SsoInfo.bValidsrvCtx = true;
	string AuthenticationHeader = Method.c_str();
	AuthenticationHeader += " ";
	char *Buffer = NULL;
	DataSize = (int)(SsoInfo.ob.cbBuffer);
	if(DataSize)
	{
		MemBase64Encode((const char *)(SsoInfo.ob.pvBuffer), &Buffer, &DataSize, false);
		Buffer[DataSize] = '\0';
		string Hash = MD5_string(Buffer);
		if(FromNvc)
			OutputMessage.AddHeader("Nv_authenticate_data_hash", Hash);
		AuthenticationHeader += Buffer;
		free(Buffer);
		OutputMessage.AddHeader("WWW-Authenticate", AuthenticationHeader);
	}
	// by now we have a client context
	SsoInfo.haveContext = true;

	if(SsoInfo.ob.pvBuffer != NULL)
	{
		free(SsoInfo.ob.pvBuffer);
		SsoInfo.ob.pvBuffer = NULL;
	}
	if(err == SEC_I_CONTINUE_NEEDED)
	{
		// Continue negotiation
		if(LogSSO != NULL)
		{
			LogMessage = "Continue negotiation\r\n";
			fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
		}
		if(LogSSO != NULL)
			fclose(LogSSO);
		return -200;
	}

	// The authentication succeeded
	// So we can get user name and domain before return

	HANDLE	token;
	if(QuerySecurityContextToken(&(SsoInfo.srvCtx), &token) == SEC_E_OK)
	{
		
		DWORD		retlen;
		if(GetTokenInformation(token, TokenUser, NULL, 0, &retlen) || (GetLastError() == ERROR_INSUFFICIENT_BUFFER))
		{
			TOKEN_USER *tokenuser = (TOKEN_USER *)(malloc(retlen));
			char		accountname[1024];
			char		domainname[1024];
			DWORD		accountnamesize = sizeof(accountname);
			DWORD		domainnamesize = sizeof(domainname);
			SID_NAME_USE accountnameuse;
			if(GetTokenInformation(token, TokenUser, tokenuser, retlen, &retlen))
			{
				if(LookupAccountSid(NULL, tokenuser->User.Sid, accountname, &accountnamesize,
							domainname, &domainnamesize, &accountnameuse))
				{
					SSOUSer = accountname;
					SSODomain = domainname;
				}
			}
			free(tokenuser);
		}
		CloseHandle(token);
	}

	if(LogSSO != NULL)
	{
		LogMessage = "Sso login was successfull user=" + SSOUSer + " domain=" + SSODomain + "\r\n";
		fwrite(LogMessage.c_str(), 1, LogMessage.length(), LogSSO);
	}

	SsoInfo.Clear();
	if(LogSSO != NULL)
		fclose(LogSSO);
	if(!bAppSSO)
	{
		// The application was not requesting to do SSO but we were forced to do it (ex POST from web browser)
		SSOUSer = "";
		SSODomain = "";
		return 1;
	}
	// If the user is nvadmin, we cannot authenticate via SSO
	if(NirvaUserName == "nvadmin")
		return 1;

	return 2;

#else //WIN32
	// The SSO authentication is not available on unix for now
	return 1;
#endif

}

// Do redirection if configured to do so
// This command is called from BrowserFunction only when the session doesn't exist
bool NvHttpClient::DoRedirect(const char* AppName)
{
	if(AppName == NULL)
		return false;

	if(InputMessage.GetParameter("NV_NO_REDIRECT") == "YES")
		return false;

	string Path = InputMessage.vDatas[0]->sInfo;
	// Search method
	string::size_type index = Path.find(' ');
	if(index == string::npos)
		return false;
	Path = Path.substr(index+1);
	// Search HTTP version
	index = Path.find(' ');
	if(index != string::npos)
		Path = Path.substr(0, index);
	if(!Path.length())
		Path = "/";
	// Path must always start with /
	if(Path[0] != '/')
	{
		string Provistr = "/";
		Provistr += Path;
		Path = Provistr;
	}

	// Get redirections from application
	NvApplication *Application = GetMain()->Applications.GetApplication(AppName);
	if(Application == NULL)
		return false;
	vector<string> Redirections;
	Application->Redirections.GetRedirections(Redirections);
	GetMain()->Applications.FreeApplication(AppName);

	if(!Redirections.size())
		return false;

	string sServer = "";
	bool bRedirectSsl = false;
	Redirections.push_back("NV_LOCAL_SERVER_REDIRECTION");
	while(Redirections.size())
	{
		int EntryIndex = rand();
		EntryIndex = EntryIndex % Redirections.size();
		NvSocket ProviSocket;
		string Entry = Redirections[EntryIndex];
		CleanString(Entry);
		if(!Entry.length())
		{
			Redirections.erase(Redirections.begin()+EntryIndex);
			continue;
		}

		if(Entry == "NV_LOCAL_SERVER_REDIRECTION")
			return false;	// The redirection mecanism choose to stay on this server

		// An entry must start with http:// or https://
		// if not we consider http://
		bRedirectSsl = false;
		if(Entry.find("https://") == 0)
			bRedirectSsl = true;
		else if(Entry.find("http://") == string::npos)
		{
			string Provistr = "http://";
			Provistr += Entry;
			Entry = Provistr;
		}
		string Address = Entry;
		string::size_type index = Address.find("://");
		Address = Address.substr(index+3);
		int Port = 0;
		SplitAddress(Address, Port);
/*		index = Address.find(':');
		if(index != string::npos)
		{
			Port = atoi(Address.substr(index+1).c_str());
			Address = Address.substr(0, index);
			CleanString(Address);
		}*/
		if(Port <= 0)
		{
			Port = GetMain()->GetHttpPort();
			if(sSocket.bSslConnection)
				Port = GetMain()->GetHttpsPort();
		}

		// If the redirection is for ssl and we are not ssl, we skip and the opposite
		if(bRedirectSsl != sSocket.bSslConnection)
		{
			Redirections.erase(Redirections.begin()+EntryIndex);
			continue;
		}

		// try to see if server is not in exclude list
		string ExcludeEntry = "http://";
		if(bRedirectSsl)
			ExcludeEntry = "https://";
		ExcludeEntry += Address;
		char Buff[20];
		sprintf(Buff, ":%d", Port);
		ExcludeEntry += Buff;
		bool bIsExcluded = false;
		NV_LOCKLOCK(&ProtectRedirectionExcludeList);
		if(RedirectionExcludeList.find(ExcludeEntry) != RedirectionExcludeList.end())
		{
			// The server is in the exclude list
			// We check the number of access
			int NumAccess = RedirectionExcludeList[ExcludeEntry];
			if(NumAccess < 100)
			{
				// We requested to acces it less than 100 times so the server stays in exclude list
				// This is to optimize performance so an excluded server can be accessed only after
				// at least 100 requests
				bIsExcluded = true;
				NumAccess++;
				RedirectionExcludeList[ExcludeEntry] = NumAccess;
			}
			else
			{
				// We remove it from the Excluded list
				RedirectionExcludeList.erase(ExcludeEntry);
			}
		}
		NV_UNLOCKLOCK(&ProtectRedirectionExcludeList);
		
		if(bIsExcluded)
		{
			// If excluded, we must not try to connect it
			Redirections.erase(Redirections.begin()+EntryIndex);
			continue;
		}

		ProviSocket.bSslConnection = bRedirectSsl;
		if(ProviSocket.Connect(Address.c_str(), Port, 2))
		{
			// We were connected
			ProviSocket.Close();
			sServer = Address;
			bool bAddPort = true;
			if(bRedirectSsl && (Port == 443))
				bAddPort = false;
			if(!bRedirectSsl && (Port == 80))
				bAddPort = false;
			if(bAddPort)
			{
				sprintf(Buff, ":%d", Port);
				sServer += Buff;
			}
			// We can leave the loop, the redirection should be successfull
			break;
		}
		else
		{
			// We cannot connect the server so wwe add it in the exclude list and we continue loop
			NV_LOCKLOCK(&ProtectRedirectionExcludeList);
			RedirectionExcludeList[ExcludeEntry] = 1;
			NV_UNLOCKLOCK(&ProtectRedirectionExcludeList);
			Redirections.erase(Redirections.begin()+EntryIndex);
		}
	}

	if(!sServer.length())
		return false;

	string Location = "http://";
	if(bRedirectSsl)
		Location = "https://";
	Location += sServer;
	Location += Path;
	Location += "&NV_NO_REDIRECT=YES";

	SetHttpError(HTTP_ERROR_TEMPORARY_REDIRECT, "");
	OutputMessage.AddHeader("Location", Location);
	bRedirected = true;
	return true;
}

void NvHttpClient::GetLogoutPage(string LogoutPage)
{
	if(!LogoutPage.length())
	{
		string AppName = "";
		if(InputMessage.ParameterExist("NV_APPLICATION"))
			AppName = InputMessage.GetParameter("NV_APPLICATION");
		else
		{
			// If the NV_APPLICATION paramter is not given, we try to get it from
			// the URL
			AppName = InputMessage.sPath;
			MakeUpper(AppName);
			string::size_type index = AppName.find("NV_APP_");
			if(index != string::npos)
			{
				AppName = AppName.substr(index+7);
				if(((index = AppName.find('/')) != string::npos) || ((index = AppName.find('\\')) != string::npos))
					AppName = AppName.substr(0, index);
			}
			else
				AppName = "";
		}
		CleanString(AppName);
		MakeUpper(AppName);
		if(AppName.length())
		{
			NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
			if(Application != NULL)
			{
				LogoutPage = Application->GetStartPage();
				GetMain()->Applications.FreeApplication(AppName.c_str());
			}
		}
	}

	string ErrorPage = "";
	// and write a simple HTML page
	ErrorPage = "<html>\r\n";
	ErrorPage += "<head>\n";
	ErrorPage += "<title>Nirva logout page</title>\r\n";
	ErrorPage += "</head>\r\n";
	ErrorPage += "<body>\r\n";
	if(LogoutPage.length())
	{
		ErrorPage += "<SCRIPT language=\"Javascript1.2\">\r\n";
		ErrorPage += "window.location.href=\'";
		ErrorPage += LogoutPage;
		ErrorPage += "\'\r\n";
		ErrorPage += "</SCRIPT>\r\n";
	}
	else
		ErrorPage += "<h1>Nirva server: Logout page not found</h1>";
	ErrorPage += "</body>\r\n";
	ErrorPage += "</html>\r\n";

	string ContentType = "text/html; charset=UTF-8";
	MakeLower(ContentType);
	OutputMessage.AddHeader("Content-Type", ContentType);
	OutputMessage.AddHeader("Cache-Control", "no-cache");
	OutputMessage.AddHeader("Pragma", "no-cache");
	OutputMessage.AddHeader("Expires", "Wed, 26 Feb 1997 08:21:57 GMT");

	OutputMessage.vDatas[0]->FreeData();
	OutputMessage.vDatas[0]->Data = (char *)malloc(ErrorPage.length());
	if(OutputMessage.vDatas[0]->Data != NULL)
	{
		OutputMessage.vDatas[0]->DataSize = (unsigned long)(ErrorPage.length());
		memcpy(OutputMessage.vDatas[0]->Data, ErrorPage.c_str(), ErrorPage.length());
	}
}


void NvHttpClient::GetWsdl()
{
	// Special order to get the WSDL definition file for a given web service
	// The URL is: host/appname/webservice/NVS?WSDL

	// Cleanup the actual command
	Command.Clear();
	Command.m_sClientAddress = IpAddress;

	SetSession(NULL);
	string WebServiceNameHere = InputMessage.sPath;
	
	string::size_type index;
	
	// replace backslashes with slashes
	while((index = WebServiceNameHere.find('\\')) != string::npos)
		WebServiceNameHere.replace(index, 1, 1, '/');
	// eliminate first slash
	if(WebServiceNameHere.size() && WebServiceNameHere[0] == '/')
		WebServiceNameHere.erase(0, 1);
	// eliminate last slash
	if(WebServiceNameHere.size() && WebServiceNameHere[WebServiceNameHere.size()-1] == '/')
		WebServiceNameHere.erase(WebServiceNameHere.size()-1, 1);

	string ApplicationNameHere = "";
	index = WebServiceNameHere.rfind('/');
	if(index != string::npos)
	{
		ApplicationNameHere = WebServiceNameHere.substr(0, index);
		WebServiceNameHere = WebServiceNameHere.substr(index+1);
	}
	MakeUpper(ApplicationNameHere);
	index = ApplicationNameHere.find("NV_APP_");
	if(index == 0)
		ApplicationNameHere = ApplicationNameHere.substr(7);

	// We must create a special session for getting the WSDL
	NvSession Session;
	Session.Init();
	int LoginRetVal = Session.Login(NULL, NULL, ApplicationNameHere.c_str(), SESSION_TYPE_SYSTEM);	// Login with application and super internal user 
	if(LoginRetVal != 1)
	{
		// Login failed
		// We prepare error information(
	
		string Info = "";
		Info += "application=";
		if(ApplicationNameHere.length())
			Info += ApplicationNameHere;
		else
			Info += "(Default application)";


		if(LoginRetVal < 0)
		{
			if(LoginRetVal == 0)
				Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTLOGIN, Info);
			if(LoginRetVal == -1)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_USER, Info);
			if(LoginRetVal == -2)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CANNOT_GET_CONTEXT, Info);
			if(LoginRetVal == -3)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD, Info);
			if(LoginRetVal == -4)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_APPLICATION, Info);
			if(LoginRetVal == -5)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CHANGE_PASSWORD, Info);
			if(LoginRetVal == -6)
				Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD_SYNTAX, Info);
		}
		else
			Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTLOGIN, Info);

		SetHttpErrorBrowserCommand(NV_HTTP_HTML, XML_MODEL_SIMPLE);
		return;
	}

	Session.m_bAdministrator = true;
	
	Session.OriginalCommand = &Command;
	Command.m_pSession = &Session;
	Command.m_iCommandSource = NV_SOURCE_BROWSER;
	Command.m_sService = "SYSTEM";
	Command.m_sClass = "WEBSERVICE";
	Command.m_sCommand = "WSDL";
	Command.m_sInContainer = NV_CONTAINER_DEFAULT_NAME;
	Command.m_sOutContainer = NV_CONTAINER_DEFAULT_NAME;
	Command.m_sHost = InputMessage.GetHeader("X-FORWARDED-HOST");
	if(!Command.m_sHost.length())
		Command.m_sHost = InputMessage.GetHeader("HOST");
	Command.Parameters.SetValue("WEBSERVICE", WebServiceNameHere);

	// We execute the command
	bool SessionRetVal = true;
	try
	{
		SessionRetVal = Session.Command(&Command);
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		string Message = Command.m_sService;
		Message += ":";
		Message += Command.m_sClass;
		Message += ":";
		Message += Command.m_sCommand;
		Command.Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_EXCEPTION, Message);
		GetMain()->Logs.Write("Exceptions", "Exception in command (GetWsdl)", Message.c_str(), NV_LOG_ERROR);
		SessionRetVal = false;
	}

	if(SessionRetVal)
	{
		// now, we manage output data
		// This should be something readable from a browser

		// Set http output headers

		map<string,string>::iterator pos;
		for(pos = Command.HttpOutHeaders.m_mList.begin(); pos != Command.HttpOutHeaders.m_mList.end(); ++pos)
		{
			string HeaderNameHere = pos->first;
			CleanString(HeaderNameHere);
			string  HeaderValueHere = pos->second;
//			CleanString(HeaderValueHere);

			if(!HeaderNameHere.length())
				continue;

			OutputMessage.AddHeader(HeaderNameHere, HeaderValueHere);
		}
		
		OutputMessage.AddHeader("Content-Length", "0");

		// The ownership of ouput data is transfered from command to http part
		OutputMessage.vDatas[0]->DataSize = Command.DataSize;
		OutputMessage.vDatas[0]->Data = Command.Data;
		OutputMessage.vDatas[0]->DataFile = Command.DataFile;
		OutputMessage.vDatas[0]->bDeleteDataFile = Command.bDeleteDataFile;
		OutputMessage.AddHeader("Content-Length", "0");
		if(Command.DataSize)
		{
			char ProviBuf[50];
			sprintf(ProviBuf, "%ld", (unsigned long)(Command.DataSize));
			OutputMessage.AddHeader("Content-Length", ProviBuf);
		}
		Command.DataSize = 0;
		Command.Data = NULL;
		Command.DataFile = "";
		Command.bDeleteDataFile = false;
	}

	Session.Logout();
	SetSession(NULL);

	// The error information is generated at the end in order for it to eventually not contain the session ID
	// if closed
	if(Command.Error.m_iErrorCode)
		SetHttpErrorBrowserCommand(NV_HTTP_HTML, XML_MODEL_SIMPLE);
	
	// Cleanup the actual command
	Command.Clear();

}




void NvHttpClient::ProcessPrivateCommand()
{
	// Special internal command used for synchronisation
	// between different Nirva servers

	// for now, it returns the state of the server
	// The state can be STARTING, STOPPING, RUNNING or STANDBY

	// Cleanup the actual command
	Command.Clear();
	Command.m_sClientAddress = IpAddress;

	SetSession(NULL);
	string CommandToProcess = InputMessage.sPath;
	CleanString(CommandToProcess);
	MakeUpper(CommandToProcess);
	if(CommandToProcess == "GET_STATUS")
	{
		OutputMessage.AddHeader("Content-Length", "0");
		OutputMessage.AddHeader("Nirva-status", GetMain()->GetStatus());
	}
}



bool NvHttpClient::SendFile(string FileName)
{
	CleanString(FileName);
	string sFileName = FileName;
	string AppName = "";
	string DefaultPage = Server->GetDefaultPage();
	bool bIsDocAppRootFile = false;
	string::size_type index;
	string Location = InputMessage.sPath;	// used when request to a directory but the last / char is not given
	if(Location.size() && Location[Location.size()-1] != '/')
		Location += "/";
	if(InputMessage.sService.length())
		Location += InputMessage.sService;
	else
		Location += InputMessage.sCommand;
	Location += "/";
	bool AppFound = false;
	string AbsoluteStartPath = "";
	if(!sFileName.length())
	{
		// If the file name is not given
		// We try to get it from the input message
		sFileName = Server->GetRootDir();
		AbsoluteStartPath = sFileName;
		RemoveDirSuffix(sFileName);
		string RealPath = InputMessage.sPath;
		CleanString(RealPath);

		// remove points
//		while((index = RealPath.find('.')) != string::npos)
//			RealPath.erase(index, 1);

		// concatenate 2 consecutive slashes
		while((index = RealPath.find("//")) != string::npos)
			RealPath.replace(index, 2, "/");

		// concatenate 2 consecutive backslashes
		while((index = RealPath.find("\\\\")) != string::npos)
			RealPath.replace(index, 2, "\\");

		string Rest = "";
		if(InputMessage.sService.length())
			AppName = RealPath.substr(1) + InputMessage.sService;
		else
			AppName = RealPath.substr(1) + InputMessage.sCommand;
		string AppNameForSearch = AppName;
		MakeUpper(AppNameForSearch);
		if(AppNameForSearch.find("NV_") != string::npos)
		{
			index = AppName.find('/');
			if(index != string::npos)
			{
				Rest = AppName.substr(index);
				AppName = AppName.substr(0, index);
			}
			MakeUpper(AppName);
			CleanString(AppName);
			
			// Check for an application or system web file
			index = AppName.find("NV_APP_");
			if(index == 0)
			{
				// Application web file requested
				AppName = AppName.erase(0,7);
				CleanString(AppName);
				NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
				if(Application != NULL)
				{
					sFileName = Application->GetWrootDir();
					AbsoluteStartPath = sFileName;
					RemoveDirSuffix(sFileName);
					sFileName += Rest;
					DefaultPage = Application->GetStartPage();
					GetMain()->Applications.FreeApplication(AppName.c_str());
					AppFound = true;
				}
				index = string::npos;
			}
			else
				index = AppName.find("NV_SRV_");
			if(index == 0)
			{
				// Service web file requested
				AppName = AppName.erase(0,7);
				CleanString(AppName);
				NvService *Service = GetMain()->Services.GetService(AppName.c_str());
				if(Service != NULL)
				{
					sFileName = Service->GetWrootDir();
					AbsoluteStartPath = sFileName;
					RemoveDirSuffix(sFileName);
					sFileName += Rest;
					GetMain()->Services.FreeService(AppName.c_str());
					AppFound = true;
				}
				index = string::npos;
			}
			else
				index = AppName.find("NV_DOC_APP_");
			if(index == 0)
			{
				// Application doc web file requested
				AppName = AppName.erase(0,11);
				CleanString(AppName);
				NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
				if(Application != NULL)
				{
					sFileName = Application->GetDocDir();
					AbsoluteStartPath = sFileName;
					RemoveDirSuffix(sFileName);
					sFileName += Rest;
					GetMain()->Applications.FreeApplication(AppName.c_str());
					AppFound = true;
					bIsDocAppRootFile = true;
				}
				index = string::npos;
			}
			else
				index = AppName.find("NV_DOC_SRV_");
			if(index == 0)
			{
				// Service doc web file requested
				AppName = AppName.erase(0,11);
				CleanString(AppName);
				NvService *Service = GetMain()->Services.GetService(AppName.c_str());
				if(Service != NULL)
				{
					sFileName = Service->GetDocDir();
					AbsoluteStartPath = sFileName;
					RemoveDirSuffix(sFileName);
					sFileName += Rest;
					GetMain()->Services.FreeService(AppName.c_str());
					AppFound = true;
				}
				index = string::npos;
			}
			else
				index = AppName.find("NV_DOC_WEBS_");
			if(index == 0)
			{
				// Webservice doc web file requested
				AppName = AppName.erase(0,12);
				CleanString(AppName);
				NvWebs *Webs = GetMain()->Webservices.GetWebservice(AppName.c_str(), true, true);
				if(Webs != NULL)
				{
					sFileName = Webs->GetDocDir();
					AbsoluteStartPath = sFileName;
					RemoveDirSuffix(sFileName);
					sFileName += Rest;
					GetMain()->Webservices.FreeWebservice(AppName.c_str());
					AppFound = true;
				}
				index = string::npos;
			}
			else
				index = AppName.find("NV_DOC_");
			if(index == 0)
			{
				// system doc web file requested
				sFileName = GetMain()->GetNirvaDir();
				AbsoluteStartPath = sFileName;
				RemoveDirSuffix(sFileName);
				AddDirSuffix(sFileName);
				sFileName += "Docs";
				AddDirSuffix(sFileName);
				sFileName += Rest;
				AppFound = true;
			}
		}


		if(!AppFound)
			sFileName += RealPath;
	}

	// replace \ with /
	while((index = sFileName.find('\\')) != string::npos)
		sFileName.replace(index, 1, 1, '/');
	
	if(!AppFound)
	{
		if(InputMessage.sService.length())
			sFileName += InputMessage.sService;
		else
			sFileName += InputMessage.sCommand;
	}

	if(sFileName.length() && (sFileName[sFileName.length() - 1] == '/'))
	{
		if(bIsDocAppRootFile)
		{
			string OldFileName = sFileName;
			sFileName += "index.html";
			if(!GetNoCaseRealFileName(sFileName))
			{
				sFileName = OldFileName;
				sFileName += AppName;
				sFileName += ".htm";
				if(!GetNoCaseRealFileName(sFileName))
				{
					sFileName = OldFileName;
					sFileName += DefaultPage;
				}
			}
		}
		else
			sFileName += DefaultPage;
	}
	else
	{
		if(DirExist(sFileName.c_str()))
		{
			SetHttpError(HTTP_ERROR_MOVED_PERMANENTLY, "");
			OutputMessage.AddHeader("Location", Location);
			GetMain()->Logs.Write("WebFiles", "Access requested to directory (last slash is missing)", sFileName.c_str(), NV_LOG_ERROR, 1, AppName.c_str(), "system");
			return true;
		}
		else
		{
			if(!GetNoCaseRealFileName(sFileName))
			{
				// if there is no extension
				// we try to get the html or htm file
				string::size_type index;
				index = sFileName.find('.');
				if(index != string::npos)
				{
					GetMain()->Logs.Write("WebFiles", "File not found", sFileName.c_str(), NV_LOG_ERROR, 1, AppName.c_str(), "system");
					return false;
				}
				string sRealFileName = sFileName;
				sRealFileName += ".html";
				if(!GetNoCaseRealFileName(sRealFileName))
				{
					sRealFileName = sFileName;
					sRealFileName += ".htm";
					if(!GetNoCaseRealFileName(sRealFileName))
					{
						GetMain()->Logs.Write("WebFiles", "File not found (tried with .htm or .html)", sFileName.c_str(), NV_LOG_ERROR, 1, AppName.c_str(), "system");
						return false;
					}
				}
				sFileName = sRealFileName;
			}
		}
	}

	if(!FileExist(sFileName.c_str()))
	{
		GetMain()->Logs.Write("WebFiles", "File not found (cannot open)", sFileName.c_str(), NV_LOG_ERROR, 1, AppName.c_str(), "system");
		return false;	// not a directory
	}

	// Now we must check if we went outside the Wroot directory
	string FullPathName = MyGetFullPathName(sFileName.c_str());
	if(!FullPathName.length())
	{
		GetMain()->Logs.Write("WebFiles", "Cannot get full pathname", sFileName.c_str(), NV_LOG_ERROR, 1, AppName.c_str(), "system");
		return false;	// cannot resolve real file name
	}
	// replace \ with / for AbsoluteStartPath
	while((index = AbsoluteStartPath.find('\\')) != string::npos)
		AbsoluteStartPath.replace(index, 1, 1, '/');
	// replace \ with / for FullPathName
	while((index = FullPathName.find('\\')) != string::npos)
		FullPathName.replace(index, 1, 1, '/');
	if(FullPathName.find(AbsoluteStartPath) != 0)
	{
		GetMain()->Logs.Write("WebFiles", "Attempt to access file outside the authorised directory", sFileName.c_str(), NV_LOG_ERROR, 1, AppName.c_str(), "system");
		return false;	// cannot resolve real file name
	}

	string sExtensionInFile = sFileName;
	if((index = sExtensionInFile.rfind('.')) != string::npos)
		sExtensionInFile = sExtensionInFile.substr(index+1);
	else
		sExtensionInFile = "";
	CleanString(sExtensionInFile);
	MakeLower(sExtensionInFile);

	int ActualWebServerLogLevel = GetMain()->Logs.GetLevel("WebFiles");

	// Get last modified time for the file
	string sTime = GetFileLastModifiedTime(sFileName.c_str());
	// Get eventual header If-Modified-Since
	string IfModifiedSince = InputMessage.GetHeader("IF-MODIFIED-SINCE");
	if(IfModifiedSince.length())
	{
//		string ActualTime = sTime;
		if(IfModifiedSince == sTime)
		{
			SetHttpError(HTTP_ERROR_NOT_MODIFIED, "");
			bool bLogIt = true;
			if(ActualWebServerLogLevel == 1)
			{
				if(sExtensionInFile == "gif")
					bLogIt = false;
				if(sExtensionInFile == "jpeg")
					bLogIt = false;
				if(sExtensionInFile == "jpg")
					bLogIt = false;
				if(sExtensionInFile == "css")
					bLogIt = false;
			}
			if(bLogIt)
				GetMain()->Logs.Write("WebFiles", "File not modified", sFileName.c_str(), NV_LOG_INFO, 1, AppName.c_str(), "system");
			return true;
		}
	}

	// and set it as the Last-Modified header
	OutputMessage.AddHeader("Last-Modified", sTime);
	
	// We directly attach the file to be sent
	NvData *Data = OutputMessage.vDatas[0];
	Data->DataFile = sFileName;

//	if(!OutputMessage.SetDataFromFile(sFileName.c_str(), 0))
//		return false;


	// Set mime type regarding extension
	if(sExtensionInFile.length())
	{
		string sMimeType = GetMain()->GetMimeType(sExtensionInFile);
		CleanString(sMimeType);
		if(sMimeType.length())
			OutputMessage.AddHeader("Content-Type", sMimeType);
		else
			OutputMessage.AddHeader("Content-Type", "application/octet-stream");
	}

	if(ActualWebServerLogLevel)
	{
		bool bLogIt = true;
		if(ActualWebServerLogLevel == 1)
		{
			if(sExtensionInFile == "gif")
				bLogIt = false;
			if(sExtensionInFile == "jpeg")
				bLogIt = false;
			if(sExtensionInFile == "jpg")
				bLogIt = false;
			if(sExtensionInFile == "css")
				bLogIt = false;
		}
		if(bLogIt)
		{
			string Message = sFileName;
			if((index = Message.rfind('/')) != string::npos)
				Message = Message.substr(index+1);
			if((index = Message.rfind('\\')) != string::npos)
				Message = Message.substr(index+1);

			GetMain()->Logs.Write("WebFiles", Message.c_str(), sFileName.c_str(), NV_LOG_INFO, 1, AppName.c_str(), "system");
		}
	}

	return true;
}


bool NvHttpClient::GetStaticView()
{
	string ContentType = "";
	if(InputMessage.vDatas[0]->mHeaders.find("CONTENT-TYPE") != InputMessage.vDatas[0]->mHeaders.end())
		ContentType = InputMessage.vDatas[0]->mHeaders["CONTENT-TYPE"];

	string Charset = ContentType;
	MakeLower(Charset);
	string::size_type index;
	if((index = Charset.find("charset")) == string::npos)
		Charset = "";
	else
		Charset = Charset.substr(index+7);
	if((index = Charset.find("=")) != string::npos)
		Charset = Charset.substr(index+1);
	if((index = Charset.find(";")) != string::npos)
		Charset = Charset.substr(0, index);
	CleanString(Charset, true);
	// remove double quotes
	while((index = Charset.find('\"')) != string::npos)
		Charset.erase(index, 1);
	// remove quotes
	while((index = Charset.find('\'')) != string::npos)
		Charset.erase(index, 1);
	MakeUpper(Charset);

	GetInputFormValues(ContentType, Charset);
	string ViewName = InputMessage.GetParameter("NV_HREND_PAGE");

	string AppName = "";
	if(InputMessage.ParameterExist("NV_APPLICATION"))
		AppName = InputMessage.GetParameter("NV_APPLICATION");
	else
	{
		// If the NV_APPLICATION parameter is not given, we try to get it from
		// the URL
		AppName = InputMessage.sPath;
		MakeUpper(AppName);
		string::size_type index = AppName.find("NV_APP_");
		if(index != string::npos)
		{
			AppName = AppName.substr(index+7);
			if(((index = AppName.find('/')) != string::npos) || ((index = AppName.find('\\')) != string::npos))
				AppName = AppName.substr(0, index);
		}
		else
			AppName = "";
	}
	CleanString(AppName);
	MakeUpper(AppName);
	if(!AppName.length())
		AppName = "NVDEF";

	NvApplication *Application = GetMain()->Applications.GetApplication(AppName.c_str());
	if((Application == NULL) || !(Application->IsEnable()) || !(Application->GetUseFramework()))
		return false;

	NvData *Data = OutputMessage.vDatas[0];
	string SessionIdHere = InputMessage.GetParameter("NV_SESSION_ID");
	CleanString(SessionIdHere);
	if(SessionIdHere.length() && (SessionIdHere[0] == '.'))
		SessionIdHere = "";
	NvSession *SessionHere = GetMain()->Sessions.GetSession(SessionIdHere.c_str());

	// Get language
	string LanguageHere = InputMessage.GetParameter("NV_LANGUAGE");
	MakeUpper(LanguageHere);
	CleanString(LanguageHere);
	if(!LanguageHere.length())
		if(SessionHere != NULL)
			LanguageHere = SessionHere->m_sLanguage;

	if(SessionHere != NULL)
		GetMain()->Sessions.FreeSession(SessionIdHere.c_str());


	bool RetVal = false;
	try
	{

		NvCommand CommandToExecute;
		CommandToExecute.m_sLanguage = LanguageHere;
		RetVal = Application->NidgetFramework.GetView(ViewName.c_str(), SessionIdHere.c_str(), &CommandToExecute, &(Data->Data), &(Data->DataSize), GetMain()->GetDebug());
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Exception in get static view", ViewName.c_str(), NV_LOG_ERROR);
		RetVal = false;
	}

	GetMain()->Applications.FreeApplication(AppName.c_str());
	
	return RetVal;
}



bool NvHttpClient::ParseXmlCommand(NvObjBinary *XmlDataSource)
{
	
	string::size_type index;
	int inputSize = Command.DataSize;
	bool bFileInput = false;
	if(Command.DataFile.length())
	{
		inputSize = (int)(GetFileLength(Command.DataFile.c_str()));
		bFileInput = true;
	}
	
	char *inputData = Command.Data;
	if(XmlDataSource != NULL)	// Test if the source data is message or XmlDataSource
	{
		inputSize = XmlDataSource->GetSize();
		inputData = XmlDataSource->m_cdata;
		bFileInput = false;
	}


	// We read a maximum of 1024 characters for retrieving the <NVCOMMAND> tag
	// Except if the input comes from XmlDataSource (at this time we take all (SOAP headers probably))
	int MaxSizeToRead = inputSize;
	if(XmlDataSource != NULL)
		if(MaxSizeToRead > 1024)
			MaxSizeToRead = 1024;

	if(MaxSizeToRead < 5)
		return true;	// nothing to do

	NvObjBinary XmlData;
	if(bFileInput)
	{
		if(!XmlData.Load(Command.DataFile.c_str(), 0, MaxSizeToRead))
			return false;
	}
	else
	{
		if(!XmlData.Populate(inputData, MaxSizeToRead))
			return false;
	}
	
	// We must first search for the root tag

	bool bCommandTagFound = false;
	bool bRootFound = false;
	string Root = "NIRVA";
	string CompleteRoot = "NIRVA";
	string CommandTag = "NVCOMMAND";
	string Line;
	string NameSpace = "";
	string StartFlow = "";

	while(XmlData.ReadLine(Line) > 0)
	{
		if(!bRootFound)
		{
			if(!Line.length())
				continue;
			if(Line.length() < 2)
				continue;
			index = Line.find("<?");
			if(index != string::npos)
			{
				StartFlow = Line.substr(index);
				while((index = Line.find(">")) == string::npos)
				{
					if(XmlData.ReadLine(Line) == 0)
						return false;	// Error
					StartFlow += Line;
				}
				index = StartFlow.find(">");
				if(index == string::npos)
					return false;	// Error
				if(index == StartFlow.length() - 1)
					Line = "";
				else
				{
					Line = StartFlow.substr(index+1);
					StartFlow = StartFlow.substr(0, index+1);
				}
				StartFlow += "\n";
			}

			index = Line.find("<");
			if(index != string::npos)
			{
				if(index < Line.length())
				{
					// Probably first tag
					Root = Line.substr(index+1);
					index = Root.find('>');
					if(index != string::npos)
						Root = Root.substr(0, index);
					else
					{
						while(1)
						{
							if(XmlData.ReadLine(Line) == 0)
								return false;
							index = Root.find('>');
							if(index == string::npos)
							{
								Root += Line;
								continue;
							}
							Root += Line.substr(0, index);
							Line = Line.substr(index+1);
							break;
						}
					}
					if(Root.length())
					{
						index = Root.rfind('/');
						if(index == Root.length()-1)
							Root = Root.substr(0, index);
					}
					CompleteRoot = Root;
					index = Root.find(' ');
					if(index != string::npos)
						Root = Root.substr(0, index);
					index = Root.find('\t');
					if(index != string::npos)
						Root = Root.substr(0, index);
					index = Root.find('\n');
					if(index != string::npos)
						Root = Root.substr(0, index);
					if(Root.length())
					{

						// We are checking now for a namespace qualified root tag
						index = Root.find(':');
						if(index != string::npos)
						{
							NameSpace = Root.substr(0, index+1);
							Root = Root.substr(index+1);
						}
						CleanString(NameSpace, true);
						CleanString(Root, true);
						if(Root.length())
							bRootFound = true;
						else
							NameSpace = "";
					}
				}
			}
		}

		if(!bRootFound)
			continue;
		
		CommandTag = "NVCOMMAND>";
		index = Line.find(CommandTag);
		if(index == string::npos)
		{
			CommandTag = "nvcommand>";
			index = Line.find(CommandTag);
			if(index == string::npos)
				continue;
		}

		// We suspect a NVCOMMAND tag
		// We must search for the start tag
		string PreTag = Line.substr(0, index);
		string::size_type index2 = PreTag.rfind('<');
		if(index2 == string::npos)
			continue;

		if(index2 == index-1)
		{
			// Real NVCOMMAND tag with no namespace
			bCommandTagFound = true;
			break;
		}

/*		// We are checking now for a namespace qualified NVCOMMAND tag
		if(PreTag.length() <= 1)
			continue;

		PreTag = PreTag.substr(index2+1);
		CleanString(PreTag);
		
		// We should not have spaces
		if(PreTag.find(' ') != string::npos)
			continue;

		if(PreTag[PreTag.length()-1] != ':')
			continue;
		
		NameSpace = PreTag;*/
		bCommandTagFound = true;
		break;
	}
	
	Command.m_sXmlNameSpace = NameSpace;

	XmlData.Clear();	// Not used for now

	if(!bCommandTagFound)
	{
		string DataFlow = "<?xml version=\"1.0\" ?>\n<";
		DataFlow += CompleteRoot;
		DataFlow += ">\n";
		DataFlow += "</";
		if(NameSpace.length())
			DataFlow += NameSpace;
		DataFlow += Root;
		DataFlow += ">\n";
		XmlData.Append(DataFlow.c_str(), (int)(DataFlow.length()));
	}
	else
	{
		NvObjBinary JustData;
		StartFlow += "<";
		StartFlow += NameSpace;
		StartFlow += Root;
		if(NameSpace.length())
		{
			StartFlow += " xmlns:";
			StartFlow += NameSpace;
			if(StartFlow.size() && StartFlow[StartFlow.size()-1] == ':')
				StartFlow.erase(StartFlow.size()-1, 1);
			StartFlow += "=\"http://zut.com\"";
		}
		StartFlow += ">\n";
		XmlData.Append(StartFlow.c_str(), (int)(StartFlow.length()));

		// Now we must search for the end tag </NVCOMMAND>
		string StartTagToSearch = "<";
		StartTagToSearch += NameSpace;
		StartTagToSearch += CommandTag;

		string EndTagToSearch = "</";
		EndTagToSearch += NameSpace;
		EndTagToSearch += CommandTag;
		string EndFlow = EndTagToSearch;
		EndFlow += "\n</";
		EndFlow += NameSpace;
		EndFlow += Root;
		EndFlow += ">\n";

		int SizeToRead = 0;
		int Offset = 0;
		if(bFileInput)
		{
			FILE* File =  Nvfopen(Command.DataFile.c_str(), "rb");
			if(File == NULL)
				return false;

			char *LineBuffer = (char*)malloc(1024);
			if(LineBuffer == NULL)
			{
				fclose(File);
				return false;
			}
			int NumBytesRead = 0;
			int TotalBytesRead = 0;

			// First search for the start tag
			while(NumBytesRead = freadline(LineBuffer, 1023, File))
			{
				TotalBytesRead += NumBytesRead;
				Line = LineBuffer;
				index = Line.find(StartTagToSearch);
				if(index != string::npos)
				{
					Offset = TotalBytesRead-NumBytesRead+(int)index;
					TotalBytesRead = NumBytesRead-(int)index;
					Line = Line.substr(index);
					break;
				}
			}

			// We have found it, we must verify that the end tag is not on the same line
			index = Line.find(EndTagToSearch);
			if(index != string::npos)
			{
				SizeToRead = (int)index;
			}
			else
			{
				while(NumBytesRead = freadline(LineBuffer, 1023, File))
				{
					TotalBytesRead += NumBytesRead;
					Line = LineBuffer;
					index = Line.find(EndTagToSearch);
					if(index != string::npos)
					{
						SizeToRead = TotalBytesRead-NumBytesRead+(int)index;
						break;
					}
				}
			}
			fclose(File);
			free(LineBuffer);
			if(SizeToRead)
			{
				if(!JustData.Load(Command.DataFile.c_str(), Offset, SizeToRead))
					return false;
			}
		}
		else
		{
			JustData.m_cdata = inputData;
			JustData.m_isize = inputSize;
			JustData.m_iposition = 0;

			int NumBytesRead = 0;
			int TotalBytesRead = 0;
			// First search for the start tag
			while(NumBytesRead = JustData.ReadLine(Line))
			{
				TotalBytesRead += NumBytesRead;
				index = Line.find(StartTagToSearch);
				if(index != string::npos)
				{
					Offset = TotalBytesRead-NumBytesRead+(int)index;
					TotalBytesRead = NumBytesRead-(int)index;
					Line = Line.substr(index);
					break;
				}
			}

			// We have found it, we must verify that the end tag is not on the same line
			index = Line.find(EndTagToSearch);
			if(index != string::npos)
			{
				SizeToRead = (int)index;
			}
			else
			{
				while(NumBytesRead = JustData.ReadLine(Line))
				{
					TotalBytesRead += NumBytesRead;
					index = Line.find(EndTagToSearch);
					if(index != string::npos)
					{
						SizeToRead = TotalBytesRead-NumBytesRead+(int)index;
						break;
					}
				}
			}
			JustData.m_cdata = NULL;
			JustData.m_isize = 0;
			JustData.Clear();
			if(SizeToRead)
				if(!JustData.Populate(inputData+Offset, SizeToRead))
					return false;
		}

		if(!SizeToRead)		// Nothing found
			return true;

		XmlData.AppendFrom(&JustData);
		XmlData.Append(EndFlow.c_str(), (int)(EndFlow.length()));
	}


	int iInputXmlCoding = XML_MODEL_SIMPLE;
	if(bRootFound && Root == "NIRVA")
		iInputXmlCoding = XML_MODEL_NORMAL;
	if(iInputXmlCoding != Command.iXmlCoding)
	{
		// We must verify if we apply automatic selection of XML output model or not
		string NvXmlModel = InputMessage.GetParameter("NV_XML_MODEL");
		CleanString(NvXmlModel, true);
		MakeUpper(NvXmlModel);
		if((NvXmlModel != "NORMAL") && (NvXmlModel != "SIMPLE") && (NvXmlModel != "COMPACT") && (NvXmlModel != "TINY"))
		{
			string sSimpleModel = Command.Parameters.GetValue("NV_XML_SIMPLE");
			CleanString(sSimpleModel, true);
			MakeUpper(sSimpleModel);
			if((sSimpleModel != "NO") && (sSimpleModel != "YES"))
				Command.iXmlCoding = iInputXmlCoding;	// We apply automatic selection
		}
	}


	// The XML flow is now in the XmlData object
	NvXml XmlParser;
	NvObjIndStringList Parameters;
	string NameSpaceInXml;
	if(!XmlParser.GetCommandParam(&Parameters, XmlData.m_cdata, XmlData.m_isize, NameSpaceInXml))
		return false;

	Command.m_sXmlNameSpace = NameSpaceInXml;

	map<string,string>::iterator pos;
	for(pos = Parameters.m_mList.begin(); pos != Parameters.m_mList.end(); ++pos)
		InputMessage.mParameters[pos->first] = pos->second;

	return true;
}

// This function reads the SOAP body from the input data and replaces the input data by this
// body content.
bool NvHttpClient::GetSoapBody()
{
	int inputsize = InputMessage.vDatas[0]->DataSize;
	bool bFileInput = false;
	if(InputMessage.vDatas[0]->DataFile.length())
	{
		inputsize = (int)(GetFileLength(InputMessage.vDatas[0]->DataFile.c_str()));
		bFileInput = true;
	}
	// We read a maximum of 1024 characters for retrieving the <Envelope> tag
	int MaxSizeToRead = inputsize;
	if(MaxSizeToRead > 1024)
		MaxSizeToRead = 1024;

	if(MaxSizeToRead < 10)
		return true;	// nothing to do

	NvObjBinary XmlData;
	if(bFileInput)
	{
		if(!XmlData.Load(InputMessage.vDatas[0]->DataFile.c_str(), 0, MaxSizeToRead-1))
			return false;
	}
	else
	{
		if(!XmlData.Populate(InputMessage.vDatas[0]->Data, MaxSizeToRead-1))
			return false;
	}
	
	int index = 0;
	int Offset = 0;
	int indexstarttag = -1;
	string NameSpace = "";
	char *Data = XmlData.m_cdata;
	int DataSize = XmlData.m_isize;
	while(1)
	{
		index = findstring(Data, DataSize-Offset, "Envelope");
		if(index == -1)
			return false;	// not a SOAP message

		// We have found the Envelope string, we must verify that its a real XML tag
		// and get the eventual namespace
		
		index = Offset+index;	// Real index
		Offset = index+1;
		if(index == 0)
			continue;	// The data starts directly with Envelope, this is not SOAP
		if((Data[index-1] != ':') && (Data[index-1] != '<'))
			continue;	// This is not valid tag
		int i=0;
		for(i=index-1; i>=0; i--)
		{
			if(Data[i] == ' ')	// No space authorized
				break;
			if(Data[i] == '<')
			{
				indexstarttag = i;
				break;
			}
		}
		if(indexstarttag != -1)
			break;
	}

	if(indexstarttag == -1)
		return false;

	// indexstarttag is the < character for the Envelope tag
	// index points to the start of the Envelope string
	Data[index] = '\0';
	if(Data[index-1] == ':')
	{
		Data[index-1] = '\0';
		NameSpace = Data+indexstarttag+1;
	}

	Data[indexstarttag] = '\0';
	string BeforeEnvelope = Data;

	XmlData.Clear();	// Not used for now

	// We have a SOAP message, we must now find for the SOAP body and extract it
	string StartTagToSearch = "<";
	if(NameSpace.length())
	{
		StartTagToSearch += NameSpace;
		StartTagToSearch += ":";
	}
	StartTagToSearch += "Body";

	string EndTagToSearch = "</";
	if(NameSpace.length())
	{
		EndTagToSearch += NameSpace;
		EndTagToSearch += ":";
	}
	EndTagToSearch += "Body";

	int SizeToRead = 0;
	int OffsetStart = -1;
	if(bFileInput)
	{
		FILE* File =  Nvfopen(InputMessage.vDatas[0]->DataFile.c_str(), "rb");
		if(File == NULL)
			return false;
		

		// Search the start body tag
		int Offset = findstring(File, StartTagToSearch.c_str());
		if(Offset == -1)
		{
			fclose(File);
			return false;
		}

		// Search the end of the start body tag
		Offset = findstring(File, ">");
		if(Offset == -1)
		{
			fclose(File);
			return false;
		}
		OffsetStart = Offset+1;

		Offset = findstring(File, EndTagToSearch.c_str());
		if(Offset == -1)
		{
			fclose(File);
			return false;
		}

		SizeToRead = Offset - OffsetStart;

		// Write the new file content with only NIRVA content
		string CopyFile = InputMessage.vDatas[0]->DataFile;
		CopyFile += ".nvcpsoap";

		FILE* FileCopy =  Nvfopen(CopyFile.c_str(), "wb");
		if(FileCopy == NULL)
		{
			RemoveFile(CopyFile.c_str());
			fclose(File);
			return false;
		}

		if(BeforeEnvelope.length())
		{
			if(fwrite(BeforeEnvelope.c_str(), 1, BeforeEnvelope.length(), FileCopy) != BeforeEnvelope.length())
			{
				fclose(FileCopy);
				fclose(File);
				RemoveFile(CopyFile.c_str());
				return false;
			}
		}
		if(!AppendFileEx(File, FileCopy, OffsetStart, SizeToRead))
		{
			fclose(FileCopy);
			fclose(File);
			RemoveFile(CopyFile.c_str());
			return false;
		}

		fclose(FileCopy);
		fclose(File);
		if(!CopyFileEx(CopyFile.c_str(), InputMessage.vDatas[0]->DataFile.c_str()))
		{
			fclose(FileCopy);
			fclose(File);
			RemoveFile(CopyFile.c_str());
			return false;
		}
		RemoveFile(CopyFile.c_str());

	}
	else
	{
		// Search the start body tag
		const char *Data = (const char*)(InputMessage.vDatas[0]->Data);
		int DataSize = InputMessage.vDatas[0]->DataSize;
		int Offset = findstring(Data, DataSize, StartTagToSearch.c_str());
		if(Offset == -1)
			return false;

		// Search the end of the start body tag
		int Offset2 = findstring(Data+Offset, DataSize-Offset, ">");
		if(Offset2 == -1)
			return false;
		OffsetStart = Offset+Offset2+1;
		if(OffsetStart >= DataSize)
			return false;

		Offset = findstring(Data+OffsetStart, DataSize-OffsetStart, EndTagToSearch.c_str());
		if(Offset == -1)
			return false;

		SizeToRead = Offset;
		if(SizeToRead)
		{
			int NewSize = SizeToRead + (int)(BeforeEnvelope.length());
			char *NewData = (char*)malloc(NewSize);
			if(NewData == NULL)
				return false;
			if(BeforeEnvelope.length())
				memcpy(NewData, BeforeEnvelope.c_str(), BeforeEnvelope.length());
			memcpy(NewData+BeforeEnvelope.length(), InputMessage.vDatas[0]->Data + OffsetStart, SizeToRead);
			free(InputMessage.vDatas[0]->Data);
			InputMessage.vDatas[0]->Data = NewData;
			InputMessage.vDatas[0]->DataSize = NewSize;
		}
		else
		{
			free(InputMessage.vDatas[0]->Data);
			InputMessage.vDatas[0]->Data = NULL;
			InputMessage.vDatas[0]->DataSize = 0;
		}
	}

	return true;
}


bool NvHttpClient::GetSoapCommandHeader(NvObjBinary *SoapCommandHeader)
{
	SoapCommandHeader->Clear();
	int inputsize = InputMessage.vDatas[0]->DataSize;
	bool bFileInput = false;
	if(InputMessage.vDatas[0]->DataFile.length())
	{
		inputsize = (int)GetFileLength(InputMessage.vDatas[0]->DataFile.c_str());
		bFileInput = true;
	}
	// We read a maximum of 1024 characters for retrieving the <Envelope> tag
	int MaxSizeToRead = inputsize;
	if(MaxSizeToRead > 1024)
		MaxSizeToRead = 1024;

	if(MaxSizeToRead < 10)
		return true;	// nothing to do

	NvObjBinary XmlData;
	if(bFileInput)
	{
		if(!XmlData.Load(InputMessage.vDatas[0]->DataFile.c_str(), 0, MaxSizeToRead-1))
			return false;
	}
	else
	{
		if(!XmlData.Populate(InputMessage.vDatas[0]->Data, MaxSizeToRead-1))
			return false;
	}
	
	int index = 0;
	int Offset = 0;
	int indexstarttag = -1;
	string NameSpace = "";
	char *Data = XmlData.m_cdata;
	int DataSize = XmlData.m_isize;
	while(1)
	{
		index = findstring(Data, DataSize-Offset, "Envelope");
		if(index == -1)
			return false;	// not a SOAP message

		// We have found the Envelope string, we must verify that its a real XML tag
		// and get the eventual namespace
		
		index = Offset+index;	// Real index
		Offset = index+1;
		if(index == 0)
			continue;	// The data starts directly with Envelope, this is not SOAP
		if((Data[index-1] != ':') && (Data[index-1] != '<'))
			continue;	// This is not valid tag
		int i=0;
		for(i=index-1; i>=0; i--)
		{
			if(Data[i] == ' ')	// No space authorized
				break;
			if(Data[i] == '<')
			{
				indexstarttag = i;
				break;
			}
		}
		if(indexstarttag != -1)
			break;
	}

	if(indexstarttag == -1)
		return false;

	// indexstarttag is the < character for the Envelope tag
	// index points to the start of the Envelope string
	Data[index] = '\0';
	if(Data[index-1] == ':')
	{
		Data[index-1] = '\0';
		NameSpace = Data+indexstarttag+1;
	}

	Data[indexstarttag] = '\0';
	string BeforeEnvelope = Data;

	string SpecialStartRootTag = "<ZutdeZut";
	if(NameSpace.length())
	{
		SpecialStartRootTag += " xmlns:";
		SpecialStartRootTag += NameSpace;
		SpecialStartRootTag += "=\"http://schemas.xmlsoap.org/soap/envelope/\"";
	}
	SpecialStartRootTag += ">\n";
	string SpecialEndRootTag = "</ZutdeZut>\n";

	XmlData.Clear();	// Not used for now

	// We have a SOAP message, we must now find for the SOAP body and extract it
	string StartTagToSearch = "<";
	if(NameSpace.length())
	{
		StartTagToSearch += NameSpace;
		StartTagToSearch += ":";
	}
	StartTagToSearch += "Header";

	string EndTagToSearch = "</";
	if(NameSpace.length())
	{
		EndTagToSearch += NameSpace;
		EndTagToSearch += ":";
	}
	EndTagToSearch += "Header";

	int SizeToRead = 0;
	int OffsetStart = -1;
	if(bFileInput)
	{
		FILE* File =  Nvfopen(InputMessage.vDatas[0]->DataFile.c_str(), "rb");
		if(File == NULL)
			return false;
		

		// Search the start header tag
		int Offset = findstring(File, StartTagToSearch.c_str());
		if(Offset == -1)
		{
			fclose(File);
			return true;	// No header
		}

		// Search the end of the start header tag
		Offset = findstring(File, ">");
		if(Offset == -1)
		{
			fclose(File);
			return false;	// Error
		}
		OffsetStart = Offset+1;

		// Verify if the header is not empty
		fseek(File, -1, SEEK_CUR);
		char c = 0;
		fread(&c, 1, 1, File);
		if(c != '/')
		{
			Offset = findstring(File, EndTagToSearch.c_str());
			if(Offset == -1)
			{
				fclose(File);
				return false;
			}

			SizeToRead = Offset - OffsetStart;
			fclose(File);
			if(!SizeToRead)
				return true;	// Should never go here
	

			if(BeforeEnvelope.length())
				SoapCommandHeader->Append(BeforeEnvelope);
			SoapCommandHeader->Append(SpecialStartRootTag);

			NvObjBinary TempObject;
			TempObject.Load(InputMessage.vDatas[0]->DataFile.c_str(), OffsetStart, SizeToRead);
			SoapCommandHeader->AppendFrom(&TempObject);
			SoapCommandHeader->Append(SpecialEndRootTag);
		}
		else
		{
			// The header is empty
			fclose(File);
			return true;
		}
	}
	else
	{
		// Search the start header tag
		const char *Data = (const char*)(InputMessage.vDatas[0]->Data);
		int DataSize = InputMessage.vDatas[0]->DataSize;
		int Offset = findstring(Data, DataSize, StartTagToSearch.c_str());
		if(Offset == -1)
			return true;	// No header

		// Search the end of the start header tag
		int Offset2 = findstring(Data+Offset, DataSize-Offset, ">");
		if(Offset2 == -1)
			return false;
		OffsetStart = Offset+Offset2+1;
		if(OffsetStart >= DataSize)
			return false;

		if(Data[Offset+Offset2-1] != '/')
		{
			// read only if this is not an empty header
			Offset = findstring(Data+OffsetStart, DataSize-OffsetStart, EndTagToSearch.c_str());
			if(Offset == -1)
				return false;

			SizeToRead = Offset;
			if(SizeToRead)
			{
				if(BeforeEnvelope.length())
					SoapCommandHeader->Append(BeforeEnvelope);
				SoapCommandHeader->Append(SpecialStartRootTag);
				SoapCommandHeader->Append(InputMessage.vDatas[0]->Data + OffsetStart, SizeToRead);
				SoapCommandHeader->Append(SpecialEndRootTag);
			}
		}
	}

	return true;
}

// Following are classes used for connection to other NIRVA servers


/////////////////////////////////////////////////////////////////////////////
// NvHttpConnectionList

NvHttpConnectionList::NvHttpConnectionList()
{
}

NvHttpConnectionList::~NvHttpConnectionList()
{
	Clear();
}




void NvHttpConnectionList::Clear()
{
	int i;
	Lock();
	for(i=0; i<(int)(m_vConnections.size()); i++)
	{
		NvHttpConnection *Connection = (NvHttpConnection *)m_vConnections[i];
		if(Connection != NULL)
			delete Connection;
	}
	m_vConnections.clear();
	Unlock();
}


NvHttpConnection* NvHttpConnectionList::GetConnection(const char *Address, int Port, int ConnectionTimeOut, bool UseSSL, int ProxyMode, const char *FinalAddress, int FinalPort, bool Reuse)
{
	// We stop updating list for each connection request for performance reasons
	//	Update();

	string AddressHere = "127.0.0.1";
	if(Address != NULL)
		AddressHere = Address;
	CleanString(AddressHere);
	if(!AddressHere.length())
		AddressHere = "127.0.0.1";

	NvHttpConnection *FoundConnection = NULL;
	
	Lock();
	int i;
	if(Reuse)
	{
		for(i=0; i<(int)(m_vConnections.size()); i++)
		{
			NvHttpConnection *Connection = (NvHttpConnection *)m_vConnections[i];
			if(Connection == NULL)
				continue;
			if((AddressHere == Connection->m_sServer) && (Port == Connection->m_iPort) && (UseSSL == Connection->b_UseSSL) && (ProxyMode == Connection->m_iProxyMode))
			{
				if(Connection->IsInUse())
					continue;
				if((ProxyMode == PROXY_FINAL_SSL))
				{
					if((Connection->m_sFinalAddress != FinalAddress) || (Connection->m_iFinalPort != FinalPort))
						continue;
				}
				FoundConnection = Connection;
				break;
			}
		}
	}

	if(FoundConnection == NULL)
	{
		NvHttpConnection *Connection = new NvHttpConnection;
		if(Connection == NULL)
		{
			Unlock();
			return NULL;
		}
		Connection->m_sServer = AddressHere;
		Connection->m_sFinalAddress = FinalAddress;
		Connection->m_iPort = Port;
		Connection->m_iFinalPort = FinalPort;
		Connection->m_iConnectionTimeOut = ConnectionTimeOut;
		Connection->b_UseSSL = UseSSL;
		Connection->m_iProxyMode = ProxyMode;
		Connection->sSocket.bSslConnection = UseSSL;
		m_vConnections.push_back((void *)Connection);
		FoundConnection = Connection;
	}

	FoundConnection->IncInuse();

	// We must unlock before trying to connect otherwise we block all other threads
	Unlock();

//	if(!FoundConnection->Connect())
//	{
//		FoundConnection->DecInUse();
//		return NULL;
//	}
	return FoundConnection;
}


void NvHttpConnectionList::FreeConnection(NvHttpConnection* Connection)
{
	Lock();
	if(Connection != NULL)
		Connection->DecInUse();
	Unlock();
}


void NvHttpConnectionList::RemoveConnection(NvHttpConnection* Connection)
{
	int i;
	Lock();
	if(Connection != NULL)
		delete Connection;
	for(i=0; i<(int)(m_vConnections.size()); i++)
	{
		NvHttpConnection *ConnectionInList = (NvHttpConnection *)m_vConnections[i];
		if(ConnectionInList == Connection)
			break;
	}
	if(i != m_vConnections.size())
		m_vConnections.erase(m_vConnections.begin()+i);

	Unlock();
}


void NvHttpConnectionList::Update()
{
	Lock();

	int i;
	vector<void *> m_vOldConnections;
	m_vOldConnections = m_vConnections;
	m_vConnections.clear();

	for(i=0; i<(int)(m_vOldConnections.size()); i++)
	{
		NvHttpConnection *Connection = (NvHttpConnection *)m_vOldConnections[i];
		if(Connection == NULL)
			continue;
		if((!(Connection->IsInUse())) && (Connection->IsTimeOut()))
		{
			delete Connection;
			continue;
		}
		m_vConnections.push_back((void *)Connection);
	}
	m_vOldConnections.clear();
	Unlock();
}




/////////////////////////////////////////////////////////////////////////////
// NvHttpConnection

NvHttpConnection::NvHttpConnection()
{
	InMessage.sSocket = &sSocket;
	OutMessage.sSocket = &sSocket;
	InUse = 0;
	iTimeOut = NV_CONNECTION_DEFAULT_TIMEOUT;
	tLastUseTime = time(NULL);
	m_sServer = "127.0.0.1";
	m_sFinalAddress = "";
	m_iPort = 1081;
	b_UseSSL = false;
	b_ProxyHttps = false;
	m_iProxyMode = NO_PROXY;
	m_iFinalPort = 1081;
}

NvHttpConnection::~NvHttpConnection()
{
	Lock();
	sSocket.Close();
	AuthInfo.Clear();
	ProxyAuthInfo.Clear();
	Unlock();
}




void NvHttpConnection::IncInuse()
{
	ProtectInUse.Lock();
	InUse++;
	tLastUseTime = time(NULL);
	ProtectInUse.Unlock();
}

void NvHttpConnection::DecInUse()
{
	ProtectInUse.Lock();
	if(InUse)
	{
		InUse--;
		tLastUseTime = time(NULL);
	}
	ProtectInUse.Unlock();
}

bool NvHttpConnection::IsInUse()
{
	bool RetVal = false;
	ProtectInUse.Lock();
	if(InUse)
		RetVal = true;
	ProtectInUse.Unlock();

	return RetVal;
}


int NvHttpConnection::GetTimeFromLastUse()
{
	time_t tActualTime = time(NULL);
	ProtectInUse.Lock();
	int RetVal = (int)(tActualTime - tLastUseTime);
	ProtectInUse.Unlock();
	return RetVal;
}


bool NvHttpConnection::IsTimeOut()
{
	// A 0 time out means no time out
	// This should not occur because a connection should always have a time out
	if(iTimeOut == 0)
		return false;
	bool RetVal = false;
	time_t tActualTime = time(NULL);
	ProtectInUse.Lock();
	int NumSeconds = (int)(tActualTime - tLastUseTime);
	if(NumSeconds > iTimeOut)
		RetVal = true;
	ProtectInUse.Unlock();
	return RetVal;
}
	

// Send the orders to the server
int NvHttpConnection::SendOrder(bool UseDataFile, int DataFileMinSize, const char *WorkDir)
{
	if(b_ProxyHttps)
		return SendOrderProxyHttps(UseDataFile, DataFileMinSize, WorkDir);
	
	bool bProxyAuthRequest = false;
	bool bAuthRequest = false;

	Lock();
	while(1)
	{
		// we clear the input message
		InMessage.Clear();

		// We add the authentication headers if there are some for both proxy and host
		SetAuthHeaders(&OutMessage, true);
		SetAuthHeaders(&OutMessage, false);

		// We send the output to client

		int Result = OutMessage.Write();

		if(Result != 1)
		{
	//		if(sSocket.WaitWrite() == 1)
	//		{
				// We are still connected so there is another error
	//			Unlock();
	//			return false;
	//		}
			// We are probably not still connected so we must reconnect
			if(sSocket.Connect(m_sServer.c_str(), m_iPort, m_iConnectionTimeOut) != 1)
			{
				// The connection failed
				Unlock();
				sSocket.Close();
				return 0;
			}

			// Let's try again after the reconnection
			Result = OutMessage.Write();
		}
		// We retry again after a short sleep
		if(Result != 1)
		{
			SLEEP(1000);
	//		if(sSocket.WaitWrite() == 1)
	//		{
				// We are still connected so there is another error
	//			Unlock();
	//			return false;
	//		}
			// We are probably not still connected so we must reconnect (the connection time out is then 5 seconds)
			if(sSocket.Connect(m_sServer.c_str(), m_iPort, 5) != 1)
			{
				// The connection failed
				Unlock();
				sSocket.Close();
				return 0;
			}

			// Let's try again after the reconnection
			Result = OutMessage.Write();
			if(Result != 1)
			{
				Unlock();
				sSocket.Close();
				return 0;
			}
		}

		Result = InMessage.Read(0, UseDataFile, DataFileMinSize, WorkDir);	// We wait infinitly
		if(Result < 0)	// error
		{
			// We are maybe not still connected so we must reconnect
			if(sSocket.Connect(m_sServer.c_str(), m_iPort, m_iConnectionTimeOut) != 1)
			{
				// The connection failed
				Unlock();
				sSocket.Close();
				return false;
			}

			// Let's try again the complete after the reconnection
			Result = OutMessage.Write();
			if(Result != 1)
			{
				Unlock();
				sSocket.Close();
				return false;
			}
			Result = InMessage.Read(0, UseDataFile, DataFileMinSize, WorkDir);	// We wait infinitly
			if(Result < 0)	// error
			{
				Unlock();
				sSocket.Close();
				return 0;
			}
			else
			{
				string sInfo = "";
				if(InMessage.vDatas.size())
					sInfo = InMessage.vDatas[0]->sInfo;
				MakeUpper(sInfo);
				if(sInfo.find("100 CONTINUE") != string::npos)
				{
					InMessage.Clear();
					Result = InMessage.Read(0, UseDataFile, DataFileMinSize, WorkDir);	// We wait infinitly
					if(Result < 0)	// error
					{
						Unlock();
						sSocket.Close();
						return 0;
					}
				}
			}
		}
		else
		{
			string sInfo = "";
			if(InMessage.vDatas.size())
				sInfo = InMessage.vDatas[0]->sInfo;
			MakeUpper(sInfo);
			if(sInfo.find("100 CONTINUE") != string::npos)
			{
				InMessage.Clear();
				Result = InMessage.Read(0, UseDataFile, DataFileMinSize, WorkDir);	// We wait infinitly
				if(Result < 0)	// error
				{
					Unlock();
					sSocket.Close();
					return 0;
				}
			}
		}
	
		// Check if there is an authentication request from proxy
		if(CheckAuthRequest(&InMessage, true))
		{
			// If we ever checked that, so this is an authentication error
			if(bProxyAuthRequest)
			{
				Unlock();
				sSocket.Close();
				return -2;
			}

			// We get the authentication headers for proxy
			GetAuthRequestHeaders(&InMessage, true);
			// And retry;
			bProxyAuthRequest = true;
			continue;
		}
		
		// Get the correct answer authentication headers for proxy
		GetAuthHeaders(&InMessage, true);

		// Check if there is an authentication request from host
		if(CheckAuthRequest(&InMessage, false))
		{
			// If we ever checked that, so this is an authentication error
			if(bAuthRequest)
			{
				Unlock();
				sSocket.Close();
				return -1;
			}

			// We get the authentication headers for host
			GetAuthRequestHeaders(&InMessage, false);
			// And retry;
			bAuthRequest = true;
			continue;
		}
		
		// Get the correct answer authentication headers for host
		GetAuthHeaders(&InMessage, false);
		break;
	}

	Unlock();
	return true;
}



// Send the orders to the server when using https via proxy
// In that case, we must tunnel the proxy via an HTTP CONNECT order
int NvHttpConnection::SendOrderProxyHttps(bool UseDataFile, int DataFileMinSize, const char *WorkDir)
{
	
	bool bProxyAuthRequest = false;
	bool bAuthRequest = false;

	NvHttpProtocol TmpInMessage;
	NvHttpProtocol TmpOutMessage;

	Lock();
	sSocket.bSslConnection = true;

	while(1)
	{
		// we clear the input message
		InMessage.Clear();

		// We add the authentication headers if there are some for host
		SetAuthHeaders(&OutMessage, false);

		// We send the output to client

		int Result = OutMessage.Write();

		while(Result != 1)
		{
	//		if(sSocket.WaitWrite() == 1)
	//		{
				// We are still connected so there is another error
	//			Unlock();
	//			return false;
	//		}
			// We are probably not still connected so we must reconnect
			sSocket.Close();
			sSocket.bSslConnection = false;
			if(sSocket.Connect(m_sServer.c_str(), m_iPort, m_iConnectionTimeOut) != 1)
			{
				// The connection failed
				Unlock();
				sSocket.Close();
				sSocket.bSslConnection = true;
				return 0;
			}
			//sSocket.bSslConnection = false;

			TmpInMessage.Clear();
			TmpOutMessage.Clear();
			TmpInMessage.sSocket = &sSocket;
			TmpOutMessage.sSocket = &sSocket;
			NvData *Data = NVNEW NvData;
			TmpOutMessage.vDatas.push_back(Data);
			// We add the authentication headers if there are some for proxy
			SetAuthHeaders(&TmpOutMessage, true);
			Data->sInfo = "CONNECT ";
			Data->sInfo += OutMessage.GetHeader("Host"); 
			Data->sInfo += " HTTP/1.1"; 
			TmpOutMessage.AddHeader("Host", OutMessage.GetHeader("Host"));
			TmpOutMessage.AddHeader("Content-Length", "0");
			TmpOutMessage.AddHeader("User-Agent", "Nirva");
			TmpOutMessage.AddHeader("Connection", "Keep-Alive");
			TmpOutMessage.AddHeader("Proxy-Connection", "Keep-Alive");

			// Let's try again after the reconnection
			Result = TmpOutMessage.Write();
			if(Result != 1)
			{
				Unlock();
				sSocket.Close();
				sSocket.bSslConnection = true;
				return 0;
			}

			// So we get connected to the proxy
			Result = TmpInMessage.Read();	// We wait infinitly
			if(Result < 0)	// error
			{
				// We are maybe not still connected so we must reconnect
				if(sSocket.Connect(m_sServer.c_str(), m_iPort, m_iConnectionTimeOut) != 1)
				{
					// The connection failed
					Unlock();
					sSocket.Close();
					sSocket.bSslConnection = true;
					return 0;
				}

				// Let's try again the complete after the reconnection
				Result = TmpOutMessage.Write();
				if(Result != 1)
				{
					Unlock();
					sSocket.Close();
					sSocket.bSslConnection = true;
					return 0;
				}
				Result = TmpInMessage.Read();	// We wait infinitly
				if(Result < 0)	// error
				{
					Unlock();
					sSocket.Close();
					sSocket.bSslConnection = true;
					return 0;
				}
				else
				{
					string sInfo = "";
					if(TmpInMessage.vDatas.size())
						sInfo = TmpInMessage.vDatas[0]->sInfo;
					MakeUpper(sInfo);
					if(sInfo.find("100 CONTINUE") != string::npos)
					{
						TmpInMessage.Clear();
						Result = TmpInMessage.Read();	// We wait infinitly
						if(Result < 0)	// error
						{
							Unlock();
							sSocket.Close();
							sSocket.bSslConnection = true;
							return 0;
						}
					}
				}
			}
			else
			{
				string sInfo = "";
				if(TmpInMessage.vDatas.size())
					sInfo = TmpInMessage.vDatas[0]->sInfo;
				MakeUpper(sInfo);
				if(sInfo.find("100 CONTINUE") != string::npos)
				{
					TmpInMessage.Clear();
					Result = TmpInMessage.Read();	// We wait infinitly
					if(Result < 0)	// error
					{
						Unlock();
						sSocket.Close();
						sSocket.bSslConnection = true;
						return 0;
					}
				}
			}
		
			// Check if there is an authentication request from proxy
			if(CheckAuthRequest(&TmpInMessage, true))
			{
				// If we ever checked that, so this is an authentication error
				if(bProxyAuthRequest)
				{
					Unlock();
					sSocket.Close();
					sSocket.bSslConnection = true;
					return -2;
				}

				// We get the authentication headers for proxy
				GetAuthRequestHeaders(&TmpInMessage, true);
				// And retry;
				bProxyAuthRequest = true;
				Result = -1;
				sSocket.bSslConnection = true;
				continue;
			}
			
			// Get the correct answer authentication headers for proxy
			GetAuthHeaders(&TmpInMessage, true);

			// Check if we got a 200 Connection established answer
			string sInfo = "";
			if(TmpInMessage.vDatas.size())
				sInfo = TmpInMessage.vDatas[0]->sInfo;
			MakeUpper(sInfo);
			if(sInfo.find("200") == string::npos)
			{
				Unlock();
				sSocket.Close();
				sSocket.bSslConnection = true;
				return 0;
			}
			// The proxy tunnel has been established
			// So we can now work in SSL
			sSocket.bSslConnection = true;
			if(!sSocket.SslInit())
			{
				Unlock();
				sSocket.Close();
				return 0;
			}
			TmpInMessage.Clear();
			TmpOutMessage.Clear();
		
			Result = OutMessage.Write();
			if(Result != 1)
			{
				Unlock();
				sSocket.Close();
				return 0;
			}
			break;
		}

		Result = InMessage.Read(0, UseDataFile, DataFileMinSize, WorkDir);	// We wait infinitly
		if(Result < 0)	// error
		{
			Unlock();
			sSocket.Close();
			return 0;
		}
		else
		{
			string sInfo = "";
			if(InMessage.vDatas.size())
				sInfo = InMessage.vDatas[0]->sInfo;
			MakeUpper(sInfo);
			if(sInfo.find("100 CONTINUE") != string::npos)
			{
				InMessage.Clear();
				Result = InMessage.Read(0, UseDataFile, DataFileMinSize, WorkDir);	// We wait infinitly
				if(Result < 0)	// error
				{
					Unlock();
					sSocket.Close();
					return 0;
				}
			}
		}

		// Check if there is an authentication request from host
		if(CheckAuthRequest(&InMessage, false))
		{
			// If we ever checked that, so this is an authentication error
			if(bAuthRequest)
			{
				Unlock();
				sSocket.Close();
				return -1;
			}

			// We get the authentication headers for host
			GetAuthRequestHeaders(&InMessage, false);
			// And retry;
			bAuthRequest = true;
			continue;
		}
		
		// Get the correct answer authentication headers for host
		GetAuthHeaders(&InMessage, false);
		break;
	}

	Unlock();
	return true;
}


// Connects the server if not connected
// Not used actually
bool NvHttpConnection::Connect()
{

	Lock();

	if(sSocket.WaitWrite() != 1)
	{
		// We are not still connected so we must reconnect
		if(sSocket.Connect(m_sServer.c_str(), m_iPort, m_iConnectionTimeOut) != 1)
		{
			Unlock();
			return false;
		}
	}
	Unlock();
	
	return true;
}




// Check if the server or proxy requests an authentication
bool NvHttpConnection::CheckAuthRequest(NvHttpProtocol* Message, bool Proxy)
{
	string CommandHere = Message->sCommand;
	MakeUpper(CommandHere);
	CleanString(CommandHere);
	string StringToSearch = "401";
	if(Proxy)
		StringToSearch = "407";

	if(CommandHere.find(StringToSearch) == 0)
		return true;

	return false;
}

// Get the headers information from authentication request response
void NvHttpConnection::GetAuthRequestHeaders(NvHttpProtocol* Message, bool Proxy)
{
	NvAuthInfo *AuthInfoHere = &AuthInfo;
	string HttpHeaderName = "WWW-AUTHENTICATE";
	if(Proxy)
	{
		HttpHeaderName = "PROXY-AUTHENTICATE";
		AuthInfoHere = &ProxyAuthInfo;
	}
	
	bool bBasic = true;

	string HttpHeader = Message->GetHeader(HttpHeaderName);

	// eliminate first and last spaces
	while(HttpHeader.size() && HttpHeader[0] == ' ')
		HttpHeader.erase(0, 1);
	while(HttpHeader.size() && HttpHeader[HttpHeader.size()-1] == ' ')
		HttpHeader.erase(HttpHeader.size()-1, 1);

	string::size_type index = HttpHeader.find(' ');
	string Method = "";
	if(index != string::npos)
	{
		Method = HttpHeader.substr(0, index);
		HttpHeader = HttpHeader.substr(index+1);
	}
	CleanString(Method);
	MakeUpper(Method);
	if(Method == "DIGEST")
		bBasic = false;

	// Parse the header data now
	map<string, string> Headers;
	while(HttpHeader.length())
	{
		string KeyValue = "";
		index = HttpHeader.find(',');
		if(index == string::npos)
		{
			KeyValue = HttpHeader;
			HttpHeader = "";
		}
		else
		{
			KeyValue = HttpHeader.substr(0, index);
			HttpHeader = HttpHeader.substr(index+1);
		}
		string KeyHere = "";
		string ValueHere = "";
		index = KeyValue.find('=');
		if(index == string::npos)
		{
			KeyHere = KeyValue;
			ValueHere = "";
		}
		else
		{
			KeyHere = KeyValue.substr(0, index);
			ValueHere = KeyValue.substr(index+1);
		}
		if(KeyHere.length())
		{
			CleanString(KeyHere);
			MakeUpper(KeyHere);
			// eliminate first and last spaces
			while(ValueHere.size() && ValueHere[0] == ' ')
				ValueHere.erase(0, 1);
			while(ValueHere.size() && ValueHere[ValueHere.size()-1] == ' ')
				ValueHere.erase(ValueHere.size()-1, 1);
			// eliminate first and last double quotes
			if(ValueHere.size() && ValueHere[0] == '\"')
				ValueHere.erase(0, 1);
			if(ValueHere.size() && ValueHere[ValueHere.size()-1] == '\"')
				ValueHere.erase(ValueHere.size()-1, 1);
			Headers[KeyHere] = ValueHere;
		}
	}
	
	if(bBasic)
	{
		// basic

		AuthInfoHere->m_iDigest = HTTP_AUTHENTICATION_BASIC;
		AuthInfoHere->m_sRealm = "";
		if(Headers.find("REALM") != Headers.end())
			AuthInfoHere->m_sRealm = Headers["REALM"];
		AuthInfoHere->m_sDomain = "";
		AuthInfoHere->m_sNonce = "";
		AuthInfoHere->m_sOpaque = "";
		AuthInfoHere->m_bStale = false;
		AuthInfoHere->m_sAlgorithm = "MD5";
		AuthInfoHere->m_sQop = "";
		AuthInfoHere->m_iNc = 1;
		AuthInfoHere->m_sCnonce = "";
	}
	else
	{
		// digest
		AuthInfoHere->m_iDigest = HTTP_AUTHENTICATION_DIGEST;
		AuthInfoHere->m_sRealm = "";
		if(Headers.find("REALM") != Headers.end())
			AuthInfoHere->m_sRealm = Headers["REALM"];
		AuthInfoHere->m_sDomain = "";
		if(Headers.find("DOMAIN") != Headers.end())
			AuthInfoHere->m_sDomain = Headers["DOMAIN"];
		AuthInfoHere->m_sNonce = "";
		if(Headers.find("NONCE") != Headers.end())
			AuthInfoHere->m_sNonce = Headers["NONCE"];
		AuthInfoHere->m_sOpaque = "";
		if(Headers.find("OPAQUE") != Headers.end())
			AuthInfoHere->m_sOpaque = Headers["OPAQUE"];
		AuthInfoHere->m_bStale = false;
		if(Headers.find("STALE") != Headers.end())
		{
			string sStale = Headers["STALE"];
			CleanString(sStale);
			MakeUpper(sStale);
			if(sStale == "TRUE")
			{
				AuthInfoHere->m_bStale = true;
				AuthInfoHere->m_iNc = 1;
			}
		}
		AuthInfoHere->m_sAlgorithm = "MD5";
		if(Headers.find("ALGORITHM") != Headers.end())
			AuthInfoHere->m_sAlgorithm = Headers["ALGORITHM"];
		CleanString(AuthInfoHere->m_sAlgorithm);
		MakeUpper(AuthInfoHere->m_sAlgorithm);
		AuthInfoHere->m_sQop = "";
		if(Headers.find("QOP") != Headers.end())
			AuthInfoHere->m_sQop = Headers["QOP"];
		CleanString(AuthInfoHere->m_sQop);
		if(AuthInfoHere->m_sQop.length())
			AuthInfoHere->m_sQop = "auth";
		AuthInfoHere->m_sCnonce = "";
	}
}

// Get the headers information from correct server response
void NvHttpConnection::GetAuthHeaders(NvHttpProtocol* Message, bool Proxy)
{
	NvAuthInfo *AuthInfoHere = &AuthInfo;
	if(Proxy)
	{
		AuthInfoHere = &ProxyAuthInfo;
	}

	// Nothing to do here for Basic authentication
	if(AuthInfoHere->m_iDigest != HTTP_AUTHENTICATION_DIGEST)
		return;

	string HttpHeaderName = "AUTHENTICATION-INFO";
	if(Proxy)
	{
		HttpHeaderName = "PROXY-AUTHENTICATION-INFO";
		AuthInfoHere = &ProxyAuthInfo;
	}
	
	string HttpHeader = Message->GetHeader(HttpHeaderName);
	if(!HttpHeader.length())
		return;

	// eliminate first and last spaces
	while(HttpHeader.size() && HttpHeader[0] == ' ')
		HttpHeader.erase(0, 1);
	while(HttpHeader.size() && HttpHeader[HttpHeader.size()-1] == ' ')
		HttpHeader.erase(HttpHeader.size()-1, 1);

	// Parse the header data now
	map<string, string> Headers;
	while(HttpHeader.length())
	{
		string KeyValue = "";
		string::size_type index = HttpHeader.find(',');
		if(index == string::npos)
		{
			KeyValue = HttpHeader;
			HttpHeader = "";
		}
		else
		{
			KeyValue = HttpHeader.substr(0, index);
			HttpHeader = HttpHeader.substr(index+1);
		}
		string KeyHere = "";
		string ValueHere = "";
		index = KeyValue.find('=');
		if(index == string::npos)
		{
			KeyHere = KeyValue;
			ValueHere = "";
		}
		else
		{
			KeyHere = KeyValue.substr(0, index);
			ValueHere = KeyValue.substr(index+1);
		}
		if(KeyHere.length())
		{
			CleanString(KeyHere);
			MakeUpper(KeyHere);
			// eliminate first and last spaces
			while(ValueHere.size() && ValueHere[0] == ' ')
				ValueHere.erase(0, 1);
			while(ValueHere.size() && ValueHere[ValueHere.size()-1] == ' ')
				ValueHere.erase(ValueHere.size()-1, 1);
			// eliminate first and last double quotes
			if(ValueHere.size() && ValueHere[0] == '\"')
				ValueHere.erase(0, 1);
			if(ValueHere.size() && ValueHere[ValueHere.size()-1] == '\"')
				ValueHere.erase(ValueHere.size()-1, 1);
			Headers[KeyHere] = ValueHere;
		}
	}
	
	if(Headers.find("NEXTNONCE") != Headers.end())
	{
		AuthInfoHere->m_sNonce = Headers["NEXTNONCE"];
		AuthInfoHere->m_iNc = 1;
	}
}


// Set the response authentication headers
void NvHttpConnection::SetAuthHeaders(NvHttpProtocol* Message, bool Proxy)
{
	NvAuthInfo *AuthInfoHere = &AuthInfo;
	if(Proxy)
	{
		AuthInfoHere = &ProxyAuthInfo;
	}
	if(!AuthInfoHere->m_sUser.length())
		return;
	if(AuthInfoHere->m_iDigest == HTTP_AUTHENTICATION_NONE)
		return;

	if(AuthInfoHere->m_iDigest == HTTP_AUTHENTICATION_DIGEST)
	{
		// Process for digest authentication
		if(!AuthInfoHere->m_sNonce.length())
			return;
		if(!AuthInfoHere->m_iNc)
			AuthInfoHere->m_iNc = 1;

		// Create a new cnonce if required
		if(!AuthInfoHere->m_sCnonce.length())
		{
			time_t actualtime = time(NULL);
			char Buf[20];
			sprintf(Buf, "%ld", actualtime);
			AuthInfoHere->m_sCnonce = Buf;
			AuthInfoHere->m_sCnonce = MD5_string(AuthInfoHere->m_sCnonce);
			AuthInfoHere->m_sCnonce = AuthInfoHere->m_sCnonce.substr(0,8);
		}
		
		// if the algorithm is "MD5" or unspecified (which then defaults to MD5):
		// A1 = unq(username-value) ":" unq(realm-value) ":" passwd
		// if the algorithm is "MD5-sess" then:
		// A1 = H( unq(username-value) ":" unq(realm-value) ":" passwd )
		//		":" unq(nonce-value) ":" unq(cnonce-value)

		string hA1;
		hA1 = AuthInfoHere->m_sUser;
		hA1 += ":";
		hA1 += AuthInfoHere->m_sRealm;
		hA1 += ":";
		hA1 += AuthInfoHere->m_sPassword;
		hA1 = MD5_string(hA1);
		if(AuthInfoHere->m_sAlgorithm == "MD5-SESS")
		{
			hA1 += ":";
			hA1 += AuthInfoHere->m_sNonce;
			hA1 += ":";
			hA1 += AuthInfoHere->m_sCnonce;
			hA1 = MD5_string(hA1);
		}

		// If the "qop" directive's value is "auth" or is unspecified, then A2 is:
		// A2       = Method ":" digest-uri-value
		// If the "qop" value is "auth-int", then A2 is:
		// A2       = Method ":" digest-uri-value ":" H(entity-body)
		// (The "Method" value is the HTTP request method as specified in section
		// 5.1.1 of RFC 2616)
		// For now only the auth is accepted by Nirva

		// First get the method and URI from request
		string Method = Message->vDatas[0]->sInfo;
		string Uri = "";
		string::size_type index = Method.find(' ');
		if(index != string::npos)
		{
			Uri = Method.substr(index+1);
			Method = Method.substr(0, index);
		}
		index = Uri.find(' ');
		if(index != string::npos)
			Uri = Uri.substr(0, index);

		string hA2 = Method;
		hA2 += ":";
		hA2 += Uri;
		hA2 = MD5_string(hA2);

		string Response = hA1;
		Response += ":";
		Response += AuthInfoHere->m_sNonce;
		Response += ":";
		if(AuthInfoHere->m_sQop.length())
		{
			char Buf[20];
			sprintf(Buf, "%08x", AuthInfoHere->m_iNc);
			Response += Buf;
			Response += ":";
			Response += AuthInfoHere->m_sCnonce;
			Response += ":";
			Response += AuthInfoHere->m_sQop;
		}
		Response += ":";
		Response += hA2;
		Response = MD5_string(Response);

		string Authentication = "username=\"";
		Authentication += AuthInfoHere->m_sUser;
		Authentication += "\" ,realm=\"";
		Authentication += AuthInfoHere->m_sRealm;
		Authentication += "\" ,nonce=\"";
		Authentication += AuthInfoHere->m_sNonce;
		Authentication += "\" ,uri=\"";
		Authentication += Uri;
		if(AuthInfoHere->m_sQop.length())
		{
			Authentication += "\" ,cnonce=\"";
			Authentication += AuthInfoHere->m_sCnonce;
			char Buf[20];
			sprintf(Buf, "%08x", AuthInfoHere->m_iNc);
			Authentication += "\" ,nc=";
			Authentication += Buf;
			Authentication += " ,qop=\"";
			Authentication += AuthInfoHere->m_sQop;
		}
		Authentication += "\" ,response=\"";
		Authentication += Response;
		Authentication += "\"";
	
		if(AuthInfoHere->m_sQop.length())
			AuthInfoHere->m_iNc++;

		if(AuthInfoHere->m_sOpaque.length())
		{
			Authentication += " ,opaque=\"";
			Authentication += AuthInfoHere->m_sOpaque;
			Authentication += "\"";
		}

		if(AuthInfoHere->m_sAlgorithm.length())
		{
			Authentication += " ,algorithm=\"";
			Authentication += AuthInfoHere->m_sAlgorithm;
			Authentication += "\"";
		}
	
		string AuthenticationHeader = "Digest ";
		AuthenticationHeader += Authentication;

		if(Proxy)
			Message->AddHeader("Proxy-Authorization", AuthenticationHeader);
		else
			Message->AddHeader("Authorization", AuthenticationHeader);
	}
	if(AuthInfoHere->m_iDigest == HTTP_AUTHENTICATION_BASIC)
	{
		// Process for basic authentication
		string Authentication = AuthInfoHere->m_sUser;
		Authentication += ":";
		Authentication += AuthInfoHere->m_sPassword;
		char *Buffer = NULL;
		int DataSize = (int)(Authentication.length());
		MemBase64Encode(Authentication.c_str(), &Buffer, &DataSize, false);
		Authentication = Buffer;
		free(Buffer);
		CleanString(Authentication);
		string AuthenticationHeader = "Basic ";
		AuthenticationHeader += Authentication;

		if(Proxy)
			Message->AddHeader("Proxy-Authorization", AuthenticationHeader);
		else
			Message->AddHeader("Authorization", AuthenticationHeader);
	}
}




/////////////////////////////////////////////////////////////////////////////
// NvAuthInfo

NvAuthInfo::NvAuthInfo()
{
	Clear();
}

NvAuthInfo::~NvAuthInfo()
{
}

void NvAuthInfo::Clear()
{
	m_iDigest = HTTP_AUTHENTICATION_NONE;
	m_sUser = "";
	m_sPassword = "";
	m_sRealm = "";
	m_sDomain = "";
	m_sNonce = "";
	m_sOpaque = "";
	m_bStale = false;
	m_sAlgorithm = "MD5";
	m_sQop = "";
	m_iNc = 1;
	m_sCnonce = "";
}





/////////////////////////////////////////////////////////////////////////////
// NvSsoInfo

NvSsoInfo::NvSsoInfo()
{
#ifdef WIN32
	haveInbuffer = false;
	haveContext = false;
	secPackInfo = NULL;
	ib.pvBuffer = NULL;
	ob.pvBuffer = NULL;
	bValidhcred = false;
	bValidsrvCtx = false;
#endif
}

NvSsoInfo::~NvSsoInfo()
{
	Clear();
}

void NvSsoInfo::Clear()
{
#ifdef WIN32
	haveInbuffer = false;
	haveContext = false;
	if(secPackInfo != NULL)
		FreeContextBuffer(secPackInfo);
	if(bValidsrvCtx)
		DeleteSecurityContext(&srvCtx);
	if(bValidhcred)
		FreeCredentialHandle(&hcred);
	bValidhcred = false;
	bValidsrvCtx = false;
	secPackInfo = NULL;
	if(ib.pvBuffer != NULL)
	{
		free(ib.pvBuffer);
		ib.pvBuffer = NULL;
	}
	if(ob.pvBuffer != NULL)
	{
		free(ob.pvBuffer);
		ob.pvBuffer = NULL;
	}
#endif
}


